#!/usr/bin/sudo /usr/bin/perl

#############################################################################
# omail.pl - Provides a web interface to user maildir spools                #
# Copyright (C) 2001 Olivier Mller <om@omnis.ch>                           #
# Copyright (C) 2000 Ernie Miller  (original GPL project: Neomail)          #
# See the CREDITS file for project contributors.                            #
#                                                                           #
# $Id: omail-prefs.pl,v 1.54 2001/08/26 12:35:27 swix Exp $                 #
# $Source: /cvsroot/omail/webmail/omail-prefs.pl,v $                        #
#                                                                           #
# This program is free software; you can redistribute it and/or             #
# modify it under the terms of the GNU General Public License               #
# as published by the Free Software Foundation; either version 2            #
# of the License, or (at your option) any later version.                    #
#                                                                           #
# This program is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
# GNU General Public License for more details.                              #
#                                                                           #
# You should have received a copy of the GNU General Public License         #
# along with this program; if not, write to the Free Software Foundation,   #
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.           #
#############################################################################

use strict;
use CGI qw(:standard);
use CGI::Carp qw(fatalsToBrowser);
use Jcode;   # can be commented out if you don't need Japanese 

CGI::nph();   # Treat script as a non-parsed-header script

$ENV{PATH} = ""; # no PATH should be needed

################# USER-CONFIGURABLE SECTION ##################

# paths

my $scriptpath = '/home/om/www/webmail';
my $sendmail = '/usr/sbin/sendmail';
my $omaildir = '/var/omail';               # to store sessions
my $logfile = '/var/omail/omail.log';      # logfile
my $langsubdir = "lang";                   # in $scriptpath

# urls

my $scripturl = '/webmail/omail.pl';
my $prefsurl = '/webmail/omail-prefs.pl';          # for preferences setup
my $button_url = '/webmail/buttons/';              # absolute url

# mode

my $cgi_mode = "suid";                     # possible : user or suid

# needed if mode = suid:

my $sys_passwd = "/etc/passwd";
my $sys_shadow = "/etc/shadow"; 
my $sys_virtualdomains = "/var/qmail/control/virtualdomains";
my $minimal_uid = 100;                     # should be greater than 100
my $minimal_gid = 100;

# needed if mode = user:

my $domainname = '';                       # necessary for cgi_mode = user
my $homedir = '';                          # necessary for cgi_mode = user
my $passwdfile = '';                       # necessary for cgi_mode = user

# if you are using vmailmgr (

my $vmailmgr = 1;                          # if you use Bruce's package...
my $vmailmgr_old = 0;                      # ...with old style passwd file (flat text)
my $vauthenticate = '/usr/local/bin/vauthenticate';
                                           # ...or the new passwd.cdb (binary)  ($vmailmgr_old = 0)


# if you use vpopmail by inter7.com (
my $vchkpw    = "/var/qmail/vpopmail/bin/vchkpw";
my $vpopmail  = 0;

# if you have a single user running all your domains, set to $vpopasswd 0
# and set $vpopuid to the uid and $vpopgid to the respective gid
# this brings a minimal performance win, as vchkpw does not have to be called twice
my $vpoppasswd=1;
my $vpopuid=1001;
my $vpopgid=83;

# ) end of vpopmail stuff


# default configuration

my $check_remote_host = "yes";		   # check if remote host IP correspond with session IP
my $sessiontimeout = 45;                   # automatic logout after N minutes inactivity
my $numberofheaders = 15;
my $max_numberofheaders = 100;
my $quotestyle = 'quote';
my $maildir = "";
my $use_cookies = "yes";                       # to save the username only
my $empty_trash = "yes";              	   # to empty trash on logout/login (recommended!)
my $save_outgoing_mail = "yes";                # save outgoing mails to "sent" folder
my $display_html_attach = "no";               # display html attachments inline
my $email_display_font_type = "variable";
my $autorefresh = 0;                       # in minutes, turned off by default
my $scriptadmin = 'notdefined@notdefined.dom';            # for complaints / support
my $default_lang = "de";                   # default language
my @available_lang = ("en","fr","de", "it", "pl", "es", "dk", "cz", "cziso", "ja", "lt", "id", "tc", "no", "nl", "sr");
my %available_lang_names = (en => "English",
                            it => "Italiano",
                            fr => "Francais",
                            pl => "Polski",
                            dk => "Dansk",
                            es => "Castellano",
                            cz => "Czech",
                            cziso => "Czech Iso",
                            ja => "Japanese",
                            lt => "Lithuanian",
                            id => "Indonesian",
                            tc => "Chinese",
                            no => "Norwegian",
                            nl => "Dutch",
                            sr => "Srpski",     
                            de => "Deutsch");   
my @available_quotestyle = ("quote", "outlook");
my %available_quotestyle_names = ("quote" => "Unix style \"> \"",
                                  "outlook" => "Outlook style");

# fonts stuff

my $default_variable_font = "Arial,Helvetica,sans-serif";
my $default_fixed_font = "Courier";


############### END USER-CONFIGURABLE SECTION ################

################### PROGRAM INIT #############################


my $version = "0.98.4";

my $maxabooksize = 100;
my @sigfiles = ('signature', 'signature2', 'signature3', 'signature4', 'signature5', 'header', 'header2', 'header3', 'header4', 'header5');

my %prefshash; # will hold the hash returned by readuserprefs()

my $action = param("action");
my $thissession = param("sessionid");
my $user = '';
$thissession =~ /^(.+)\-session\-0.*/ && ($user = $1);   # Grab userid from sessionid and untaint

writelog("started - version $version");

my $prefsdir = "";
my $lang = "";
my ($user_uid, $user_gid, $user_realname);   # only for suid mode needed

# check 
verifysession();

$prefsdir = "$maildir/config"; 

# read user parameters (if any)
my %prefshash = %{&readprefs};        # we need to know language...

# select language to use
my $lang = "";
$lang = $prefshash{'lang'} if ($prefshash{'lang'});
if (!$lang) {  # should happens only on firsttime
  $lang = param("lang") if (defined(param("lang"))); 
}
($lang =~ /^(.{2,10})$/) && ($lang = $1);            # untaint var
$lang = $default_lang if (!$lang);  # if no valid language found, and no lang in user config

my %TXT;
# try to open lang file

# check rem in omail.pl
#if (-f "$scriptpath/$langsubdir/$lang.lang") {
#  require "$scriptpath/$langsubdir/$lang.lang";
#  %TXT = getstrings();
#} else {
#  omailerror("couldn't open language file $scriptpath/$langsubdir/$lang.lang");
#}

if (-f "./$langsubdir/$lang.lang") {
  require "./$langsubdir/$lang.lang";
  %TXT = getstrings();
} else {
  omailerror("couldn't open language file ./$langsubdir/$lang.lang");
}


$default_fixed_font = $TXT{'default_fixed_font'} if ($TXT{'default_fixed_font'});
$default_variable_font = $TXT{'default_variable_font'} if ($TXT{'default_variable_font'});


my $firstmessage;
if (param("firstmessage")) {
   $firstmessage = (param("firstmessage") -1);  # -1 deals with "Jump" numbers
} else {
   $firstmessage = 0;
}

my $sort;
if (param("sort")) {
   $sort = param("sort");
} else {
   $sort = 'date';
}

my $folder;
if (param("folder")) {
  $folder = param("folder");
  omailerror ("No such folder: $folder") unless ( ($folder eq 'INBOX') ||
						    ($folder eq 'SENT') || ($folder eq 'SAVED') || ($folder eq 'TRASH') );
} else {
  $folder = "INBOX";
}


my $firsttimeuser = param("firsttimeuser") || ''; # Don't allow cancel if 'yes'

# once we print the header, we don't want to do it again if there's an error
my $headerprinted = 0;




############# abook stuff #############

my @abook_fields = ("id","title","firstname","lastname","emailhome",
"emailwork","phonehome","phonehomemobile","phonework","phoneworkmobile",
"faxhome","faxwork","homepagehome","homepagework","addressline1",
"addressline2","addressline3","zipcode","city","state","country",
"birthdateday","birthdatemonth","birthdateyear","category",
"displayname","nickname","organisation","departement","function",
"icq","aim","msn","irc","notes","custom1","custom2",
"custom3","custom4","custom5");

#######################################


########################## MAIN ##############################
if ($action) {      # an action has been chosen
   if ($action =~ /^(\w+)$/) {
      $action = $1;
      if ($action eq "saveprefs") {
         saveprefs();
      } elsif ($action eq "addressbook") {
         addressbook();
      } elsif ($action eq "editaddresses") {
         editaddresses();
      } elsif ($action eq "importabook") {
         importabook();
      } elsif ($action eq "viewaddress") {
         viewaddress();
      } elsif ($action eq "editaddress") {
         editaddress();
      } elsif ($action eq "addaddresslist") {
         modaddresslist();
      } elsif ($action eq "addaddress") {
         modaddress("add");
      } elsif ($action eq "deleteaddress") {
         modaddress("delete");
      }
   } else {
      omailerror($TXT{'action_has_illegal_chars'});
   }
} else {            # no action has been taken, display prefs page
   printheader();

   print start_form();
   print '<table border="0" align="center" cellpadding="2">';
   print '<tr><td bgcolor="#002266" align="left">',
   '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',$TXT{'config_title'},'</b></font>',
   "</td><td bgcolor=\"#002266\" align=\"right\"> <a href=\"$scripturl?action=displayheaders&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage\"><font color=\"#ffffff\">[ ",$TXT{'back_to_mailbox'}," ]</font></a> ";
   print '</td></tr>';

   ### Spacing
   print '<tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">';

   print hidden(-name=>'action',
                -default=>'saveprefs',
                -override=>'1');
   print hidden(-name=>'sessionid',
                -default=>$thissession,
                -override=>'1');
   print hidden(-name=>'firstmessage',
                -default=>$firstmessage,
                -override=>'1');
   print hidden(-name=>'sort',
                -default=>$sort,
                -override=>'1');
   print hidden(-name=>'folder',
                -default=>$folder,
                -override=>'1');
   print '<B>',$TXT{'realname'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'realname',
                    -default=>$prefshash{"realname"} || $TXT{'realname'},
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'from_conf'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'from',
                    -default=>$prefshash{"from"} || '',
                    -size=>'40',
                    -override=>'1'),
				  '</td></tr><tr><td align="left" valign="top" bgcolor="#dddddd">',
				    submit($TXT{'save'}),
					 '</td><td bgcolor="#dddddd" valign="middle" align="center">&nbsp;',
					   $TXT{'multi_email_info'},
      	'</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
          '<B>',$TXT{'realname'},' 2:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'realname2',
                    -default=>$prefshash{"realname2"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'from_conf'},' 2:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'from2',
                    -default=>$prefshash{"from2"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'realname'},' 3:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'realname3',
                    -default=>$prefshash{"realname3"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'from_conf'},' 3:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'from3',
                    -default=>$prefshash{"from3"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'realname'},' 4:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'realname4',
                    -default=>$prefshash{"realname4"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'from_conf'},' 4:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'from4',
                    -default=>$prefshash{"from4"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'realname'},' 5:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'realname5',
                    -default=>$prefshash{"realname5"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'from_conf'},' 5:</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'from5',
                    -default=>$prefshash{"from5"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'replyto'},':</B><br>(reply-to)</td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'replyto',
                    -default=>$prefshash{"replyto"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'send_cc_to'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'send_cc_to',
                    -default=>$prefshash{"send_cc_to"} || '',
                    -size=>'40',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'numberofheaders'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'numberofheaders',
                    -default=>$prefshash{"numberofheaders"} || '15',
                    -size=>'4',
                    -override=>'1'),
         $TXT{'messages'},' </td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',

         '<B>',$TXT{'autorefresh'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'autorefresh', 
                    -default=>$prefshash{"autorefresh"} || '0',
                    -size=>'4',
                    -override=>'1'),
         $TXT{'minutes'},' </td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
         '<B>',$TXT{'default_sort'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">';
   my %sortlabels = ('date'=>$TXT{'sort_date'},
                     'date_rev'=>$TXT{'sort_date_rev'},
                     'sender'=>$TXT{'sort_sender'},
                     'sender_rev'=>$TXT{'sort_sender_rev'},
                     'size'=>$TXT{'sort_size'},
                     'size_rev'=>$TXT{'sort_size_rev'},
                     'subject'=>$TXT{'sort_subject'},
                     'subject_rev'=>$TXT{'sort_subject_rev'},
                     'status'=>$TXT{'sort_status'},
                     'status_rev'=>$TXT{'sort_status_rev'}
                    );
   print popup_menu(-name=>'sort',
                    -"values"=>['date','date_rev','sender','sender_rev',
                              'size','size_rev','subject','subject_rev'],
                    -default=>$prefshash{"sort"} || 'date',
                    -labels=>\%sortlabels,
                    -override=>'1');

   print '<td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
   
   '<B>',$TXT{'language'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
   popup_menu(-name=>'lang',
                    -values=>\@available_lang,
                    -default=>$lang,
                    -labels=>\%available_lang_names);
   
   print '<td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
   
   '<B>',$TXT{'quote_style'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
   popup_menu(-name=>'quotestyle',
                    -values=>\@available_quotestyle, 
                    -default=>$prefshash{'quotestyle'},
                    -labels=>\%available_quotestyle_names);


   my @yes_no_array = ("yes","no");
   my %yes_no = ('yes'=>$TXT{'yes'},
		 'no'=>$TXT{'no'});

   my @fixed_variable_array = ("variable","fixed");
   my %fixed_variable = ('variable'=>$TXT{'variable'},
		 'fixed'=>$TXT{'fixed'});
   
   print '<td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
   '<B>',$TXT{'email_display_font_type'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
   popup_menu(-name=>'email_display_font_type',
                    -values=>\@fixed_variable_array, 
                    -default=>$prefshash{'email_display_font_type'} || $use_cookies,
                    -labels=>\%fixed_variable);

   print '<td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
   '<B>',$TXT{'save_username_in_cookies'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
   popup_menu(-name=>'use_cookies',
                    -values=>\@yes_no_array, 
                    -default=>$prefshash{'use_cookies'} || $use_cookies,
                    -labels=>\%yes_no);


   print '<td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
   '<B>',$TXT{'empty_trash_on_logout'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
   popup_menu(-name=>'empty_trash',
                    -values=>\@yes_no_array, 
                    -default=>$prefshash{'empty_trash'} || $empty_trash,
                    -labels=>\%yes_no);

   print '<td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
   '<B>',$TXT{'save_outgoing_mail'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
   popup_menu(-name=>'save_outgoing_mail',
                    -values=>\@yes_no_array, 
                    -default=>$prefshash{'save_outgoing_mail'} || $save_outgoing_mail,
                    -labels=>\%yes_no);

   print '<td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">',
   '<B>',$TXT{'display_html_attach'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
   popup_menu(-name=>'display_html_attach',
                    -values=>\@yes_no_array, 
                    -default=>$prefshash{'display_html_attach'} || $display_html_attach,
                    -labels=>\%yes_no);

  print '</td></tr><tr><td align="left" valign="top" bgcolor="#dddddd">',
                   submit($TXT{'save'});
   print '</td><td bgcolor="#dddddd">&nbsp;';


   foreach my $sigfile (@sigfiles) {
    if ($sigfile =~ /header/) {
       print '</td></tr><tr><td colspan="2" bgcolor="#DDDDDD">',
          '<B>',$TXT{'header'},'</B><BR>',$TXT{'name'},': ',
          textfield(-name=>"name$sigfile",
                    -default=>$prefshash{"name$sigfile"} || $sigfile,   
                    -size=>'10',
                    -override=>'1'),		
          '<BR>',
          textarea(-name=>$sigfile,
                   -default=>$prefshash{$sigfile} ||
                             "",
                   -rows=>'5',
                   -columns=>'76',
                   -wrap=>'hard',
                   -override=>'1');
      }
   }

   foreach my $sigfile (@sigfiles) {
    if ($sigfile =~ /signature/) {
       print '</td></tr><tr><td colspan="2" bgcolor="#DDDDDD">',
          '<B>',$TXT{'signature'},'</B><BR>',$TXT{'name'},': ',
          textfield(-name=>"name$sigfile",
                    -default=>$prefshash{"name$sigfile"} || $sigfile,   
                    -size=>'10',
                    -override=>'1'), '<br>',		
          textarea(-name=>$sigfile,
                   -default=>$prefshash{$sigfile} ||
                             "",
                   -rows=>'5',
                   -columns=>'76',
                   -wrap=>'hard',
                   -override=>'1');
      }
   }



   print '</td></tr><tr><td align="left" valign="top" bgcolor="#dddddd">',
                   submit($TXT{'save'});


   print '</td>', end_form();

   if ( $firsttimeuser eq 'yes' ) {
      print '<td>&nbsp;</td></tr></table>';
   } else {
      print startform(-action=>"$scripturl");
         print hidden(-name=>'action',
                      -default=>'displayheaders',
                      -override=>'1');
         print hidden(-name=>'sessionid',
                      -default=>$thissession,
                      -override=>'1');
         print hidden(-name=>'sort',
                       -default=>$sort,
                      -override=>'1');
         print hidden(-name=>'firstmessage',
                      -default=>$firstmessage,
                      -override=>'1');
         print hidden(-name=>'folder',
                      -default=>$folder,
                      -override=>'1');
      print '<td align="right" valign="top" bgcolor="#dddddd">',
            submit($TXT{'cancel'}),
            '</td></tr></table>';
      print end_form();
   }
   printfooter();
}
###################### END MAIN ##############################


#################### EDITADDRESSES ###########################
sub editaddresses {
   my %addresses;
   my %addresses2;
   my %names;
   my ($name, $email,$id);

  if ( -f "$prefsdir/addressbook") {
    # old style addressbook available -> convert it!
    writelog("old style addressbook found: converting...");
    convertaddr();
  }

   if ( -f "$prefsdir/addressbook2" ) {
      open (ABOOK,"$prefsdir/addressbook2") or
         omailerror("Can't open your address book, but it exists!");
      while (<ABOOK>) {
         my @row = split(/\t/, $_);
	    $name = $row[3] . " " . $row[2];
	    $id = $row[0];
	    $addresses{"$id"} = $row[4];   # home
	    $addresses2{"$id"} = $row[5];  # work
	    $names{"$id"} = $name;  # name
	  }
      close (ABOOK) or omailerror("Couldn't close your address book!");
   }

   my $abooksize = ( -s "$prefsdir/addressbook2" ) || 0;
   my $freespace = int($maxabooksize - ($abooksize/1024) + .5);

   printheader();
   print '<table border="0" align="center" width="90%" cellpadding="0" cellspacing="0">';

   print '<tr><td colspan="1" bgcolor="#002266" align="left">',
   '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',$TXT{'edit_addr_title'},'</b>&nbsp;</font>',
   '</td><td colspan="2" bgcolor="#002266" align="right"><font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',
   $freespace,' KB Available</b></font></td></tr><tr><td colspan="3" bgcolor="#DCDCDC" align="left">';
   print "<a href=\"$scripturl?action=displayheaders&amp;sessionid=$thissession&amp;sort=$sort&amp;firstmessage=$firstmessage\">",$TXT{'back_to_mailbox'},"</a> | ";
###   print "<a href=\"$prefsurl?action=importabook&amp;sessionid=$thissession&amp;sort=$sort&amp;firstmessage=$firstmessage\">",$TXT{'import'},"</a> | ";
   print "<a href=\"$scripturl?action=grabaddr&amp;sessionid=$thissession&amp;sort=$sort&amp;firstmessage=$firstmessage\">",$TXT{'grabaddr'},"</a>";
   print '</td></tr>';
### Spacing
   print '<tr><td colspan="3">&nbsp;</td></tr>';
   print '<tr><td align="center" bgcolor="D3DCE3"><B>',$TXT{'name'},'</B> (',$TXT{'click_to_display'},')</td><td align="center" bgcolor="D3DCE3">
          <B>',$TXT{'email_addresses'},'</B> (',$TXT{'click_to_mail'},')</td><td align="center" bgcolor="D3DCE3"><B>',$TXT{'action'},'</B></td></tr>';

   print startform(-name=>'newaddress');

   print '<tr><td bgcolor="#EEEEEE" width="200">',
          hidden(-name=>'action',
                 -value=>'addaddress',
                 -override=>'1'),
          hidden(-name=>'sessionid',
                 -value=>$thissession,
                 -override=>'1'),
          hidden(-name=>'sort',
                 -default=>$sort,
                 -override=>'1'),
          hidden(-name=>'new',
                 -default=>'1',
                 -override=>'1'),
          hidden(-name=>'firstmessage',
                 -default=>$firstmessage,
                 -override=>'1'),
          textfield(-name=>'lastname',
                    -default=>'',
                    -size=>'25',
                    -override=>'1'),
         '</td><td bgcolor="#EEEEEE" width="200">',
          textfield(-name=>'emailhome',
                    -default=>'',
                    -size=>'25',
                    -override=>'1'),
         '</td><td align="center" bgcolor="#EEEEEE" width="100">',
         submit($TXT{'add_modifiy'}),'</td></tr>';
   print end_form();

   my $current_letter = "";
   foreach my $key (sort { lc($names{$a}) cmp lc($names{$b}) } (keys %names)) {
	next if (!$key || !$names{$key} || ($key !~ /^\d*$/));
	
	if (lc(substr($names{$key},0,1)) ne lc($current_letter)) {
	  $current_letter = substr($names{$key},0,1);
	  if ($current_letter) {
	
	    print '<tr><td bgcolor="#DDDDDD" width="500" colspan="3"><font size="+1"><b>',
		 uc($current_letter),
		 '</b></font></td></tr>';
	  }
	}
	
	print '<tr><td bgcolor="#EEEEEE" width="200">',
	  "<a href=\"$prefsurl?action=viewaddress&amp;firstmessage=$firstmessage&amp;sort=$sort&amp;sessionid=$thissession&amp;addrid=$key\">",$names{$key},"</a>",
	    '&nbsp;</td><td bgcolor="#EEEEEE" width="200">',
		 "<a href=\"$scripturl?action=composemessage&amp;firstmessage=$firstmessage&amp;sort=$sort&amp;sessionid=$thissession&amp;composetype=sendto&amp;to=",
		   $addresses{$key},'">',$addresses{$key},'</a>';
	
	if ($addresses2{$key}) {
	  print ", <a href=\"$scripturl?action=composemessage&amp;firstmessage=$firstmessage&amp;sort=$sort&amp;sessionid=$thissession&amp;composetype=sendto&amp;to=",$addresses2{$key},'">',$addresses2{$key},'</a>';
	}
	
	
	print '&nbsp;</td>';

      print '<td bgcolor="#EEEEEE" align="center" width="100">',
	 "<nobr>[ <a href=\"$prefsurl?action=editaddress&amp;firstmessage=$firstmessage&amp;sort=$sort&amp;sessionid=$thissession&amp;addrid=$key\">$TXT{'edit'}</a> ]  ",

	 "[ <a href=\"$prefsurl?action=viewaddress&firstmessage=$firstmessage&sort=$sort&sessionid=$thissession&addrid=$key\">",
	 $TXT{'view_addr'},$TXT{'view'},'</a> ]',
	 "[ <a href=\"$prefsurl?action=deleteaddress&firstmessage=$firstmessage&sort=$sort&sessionid=$thissession&addrid=$key\">",
	 $TXT{'delete_addr'},'</a> ]</nobr>';
      print '</td></tr>';
   }
   print '</table>';

   printfooter();
}
################### END EDITADDRESSES ########################


##################### IMPORTABOOK ############################
sub importabook {
   my ($name, $email);
   my %addresses;
   my $abookupload = param("abook") || '';
   my $abooktowrite='';
   my $mua = param("mua") || '';
   if ($abookupload) {
      no strict 'refs';
      my $abookcontents = '';
      while (<$abookupload>) {
         $abookcontents .= $_;
      }
      close($abookupload);
      if ($mua eq 'outlookexp5') {
         unless ($abookcontents =~ /^Name,E-mail Address/) {
            omailerror("Sorry, this file doesn't appear to be an address book
                       exported from Microsoft Outlook Express 5 in CSV format.
                       Make sure that when viewing it in a text editor, the first
                       line starts out \"Name,E-mail Address\", then go
                       <a href=\"$prefsurl?action=importabook&amp;sessionid=$thissession&amp;sort=$sort&amp;firstmessage=$firstmessage\">back</a>
                       and try again.");
         }
      }
      unless ( -f "$prefsdir/addressbook2" ) {
         open (ABOOK, ">>$prefsdir/addressbook2"); # Create if nonexistent
         close(ABOOK);
      }
      open (ABOOK,"+<$prefsdir/addressbook2") or
         omailerror("Can't open your address book!");
      unless (flock(ABOOK, 2)) {
         omailerror("Can't get write lock on address book!");
      }
      while (<ABOOK>) {
         ($name, $email) = split(/\t/, $_);
         chomp($email);
         $addresses{"$name"} = $email;
      }
      my @fields;
      my $quotecount;
      my $tempstr;
      my @processed;

      foreach (split(/\r*\n/, $abookcontents)) {
         next if ( ($mua eq 'outlookexp5') && (/^Name,E-mail Address/) );
         $quotecount = 0;

	    if ($mua eq "outlookexp5" || $mua =~ /nsmail/) {
		 @fields = split(/,/);
		 @processed = ();
		 $tempstr = '';
		 foreach my $str (@fields) {
		   if ( ($str =~ /\"/) && ($quotecount == 1) ) {
               $tempstr .= ',' . $str;
               $tempstr =~ s/\"//g;
               push (@processed, $tempstr);
               $tempstr = '';
               $quotecount = 0;
		   } elsif ($str =~ /\"/) {
               $tempstr .= $str;
               $quotecount = 1;
		   } elsif ($quotecount == 1) {
               $tempstr .= ',' . $str;
		   } else {
               push (@processed, $str);
		   }
		 }
		 if ( ($mua eq 'outlookexp5') && ($processed[0]) && ($processed[1]) ) {
		   $processed[0] =~ s/^\s*//;
		   $processed[0] =~ s/\s*$//;
		   $processed[0] =~ s/://;
		   $processed[0] =~ s/</&lt;/g;
		   $processed[0] =~ s/>/&gt;/g;
		   $processed[1] =~ s/</&lt;/g;
		   $processed[1] =~ s/>/&gt;/g;
		   $addresses{"$processed[0]"} = $processed[1];
		 } elsif ( ($mua =~ /nsmail/) && ($processed[0]) && ($processed[6]) ) {
		   $processed[0] =~ s/^\s*//;
		   $processed[0] =~ s/\s*$//;
		   $processed[0] =~ s/://;
		   $processed[0] =~ s/</&lt;/g;
		   $processed[0] =~ s/>/&gt;/g;
		   $processed[6] =~ s/</&lt;/g;
		   $processed[6] =~ s/>/&gt;/g;
		   $addresses{"$processed[0]"} = $processed[6];
		 }
	    }

	    if ($mua eq "mutt") { 
		 chomp $abookcontents;
		 @fields = split(/\ /);		 
		 if ($fields[0] eq "alias" && $fields[1] && $fields[2]) {
		   $addresses{$fields[1]} = $fields[2];
		 }
	    }
	  }
	 
      seek (ABOOK, 0, 0) or
         omailerror("Couldn't seek to the beginning of your address book!");



      while ( ($name, $email) = each %addresses ) {
         $abooktowrite .= "$name:$email\n";
      }

      if (length($abooktowrite) > ($maxabooksize * 1024)) {
         omailerror("Sorry, importing these addresses would take your address
                       book over its size limit of $maxabooksize KB.  Please go
                       <a href=\"$prefsurl?action=importabook&amp;sessionid=$thissession&amp;sort=$sort&amp;firstmessage=$firstmessage\">back</a>
                       and try again, after freeing up some space.");
      }
      print ABOOK $abooktowrite;
      truncate(ABOOK, tell(ABOOK));
      close (ABOOK) or omailerror("Couldn't close your address book!");

      print "Location: $prefsurl?action=editaddresses&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage\n\n";

   } else {
      my %mualabels = ('outlookexp5' => ' Outlook Express 5 ',
                       'nsmailwin'      => ' Win Netscape Mail 4.x ',
                       'nsmaillinux'      => ' Linux Netscape Mail 4.x ',
                       'mutt'      => ' Mutt ',
                       'pine'      => ' Pine ');


      my $abooksize = ( -s "$prefsdir/addressbook2" ) || 0;
      my $freespace = int($maxabooksize - ($abooksize/1024) + .5);

      printheader();
      print '<table border="0" align="center" width="75%" cellpadding="0" cellspacing="0">';

      print '<tr><td bgcolor="#002266" align="left">',
      '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>IMPORT ADDRESS BOOK</b></font>',
      '</td><td bgcolor="#002266" align="right"><font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',
      $freespace,' KB Available</b></font></td></tr>';
      print '<tr><td colspan="2" align="center" bgcolor="#DDDDDD">';
      print '<table border="0" align="center" width="75%" cellpadding="0" cellspacing="0">';
      print '<tr><td bgcolor="#DDDDDD" align="left">';

      print '<BR> Outlook Express 5 and Netscape Mail can export their address
             books in a format known as CSV, or Comma Separated Values.  Omail
             can import these files into your personal address book to save you
             hours of tediously typing them into your adddress book by hand.<ul><li>
             In Netscape, open your address book, and select File-&gt;Export, then
             under "Save as type:" choose "Comma Separated (*.csv)."  <BR><BR><li>For Outlook
             Express, in the main Outlook Express window, choose File-&gt;Export-&gt;Address
             Book, and select the export type "Text file."  Make sure that the Name and E-mail
             address fields come first, which is the default setting.</ul>';
      print '</td></tr></table>';
      print start_multipart_form(),
      hidden(-name=>'action',
             -value=>'importabook',
             -override=>'1'),
      hidden(-name=>'sessionid',
             -value=>$thissession,
             -override=>'1'),
      hidden(-name=>'sort',
             -default=>$sort,
             -override=>'1'),
      hidden(-name=>'firstmessage',
             -default=>$firstmessage,
             -override=>'1'),
      radio_group(-name=>'mua',
                  -"values"=>['outlookexp5','nsmailwin','mutt'],   # todo: add pine and nsmaillinux support
                  -default=>'outlookexp5',
                  -labels=>\%mualabels),
      '<BR>',
      filefield(-name=>'abook',
                -default=>'',
                -size=>'30',
                -override=>'1');
      print '</td></tr>';
      print '<tr><td width="50%" align="right" bgcolor="#DDDDDD">';
      print submit($TXT{'import'}), ' &nbsp;&nbsp;';
      print '</td>';
      print end_form();
      print start_form(),
      hidden(-name=>'action',
             -value=>'editaddresses',
             -override=>'1'),
      hidden(-name=>'sessionid',
             -value=>$thissession,
             -override=>'1'),
      hidden(-name=>'sort',
             -default=>$sort,
             -override=>'1'),
      hidden(-name=>'firstmessage',
             -default=>$firstmessage,
             -override=>'1');
      print '<td width="50%" align="left" bgcolor="#DDDDDD">';
      print ' &nbsp;&nbsp;', submit($TXT{'cancel'});
      print '</td></tr></table>';
      print end_form();
      printfooter();
   }
}
#################### END IMPORTABOOK #########################

##################### EDITADDRESS ############################

sub editaddress {

  my @row = (); 
  my $found = 0;
  my $id = param("addrid");
  if ($id !~ /^(\d*)$/) { omailerror("Input format error"); }

  if ( -f "$prefsdir/addressbook2" ) {
    open (ABOOK,"$prefsdir/addressbook2") or
	 omailerror("Can't open your address book, but it exists!");
    while (<ABOOK>) {
	 @row = split(/\t/, $_);
	 if ($id == $row[0]) { $found = 1; last; }
    }
    close (ABOOK) or omailerror("Couldn't close your address book!");
  }
  if (!$found) { omailerror("Address ID $id not found"); }


  printheader();

  print start_form();

  print '<table border="0" align="center" cellpadding="2">';
  print '<tr><td bgcolor="#002266" align="left">',
    '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',$TXT{''},'&nbsp;</b></font>',
	 "</td><td bgcolor=\"#002266\" align=\"right\"> <a href=\"$prefsurl?action=editaddresses&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage\"><font color=\"#ffffff\">[ ",$TXT{'cancel'}," ]</font></a> ";
  print '</td></tr>';

  ### Spacing
  print '<tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">';

  print hidden(-name=>'action',
                -default=>'addaddress',
                -override=>'1');
   print hidden(-name=>'sessionid',
                -default=>$thissession,
                -override=>'1');
   print hidden(-name=>'firstmessage',
                -default=>$firstmessage,
                -override=>'1');
   print hidden(-name=>'sort',
                -default=>$sort,
                -override=>'1');
   print hidden(-name=>'folder',
                -default=>$folder,
                -override=>'1');
   print hidden(-name=>'addrid',
                -default=>$id,
                -override=>'1');

  my $current = "";
  for (my $i = 0; $i<=$#abook_fields; $i++) {
    $current = $abook_fields[$i];
    if ($i == 0) { next; }
    
    print '<B>',$TXT{$current},':</B></td>',
	 '<td valign="middle" align="left" bgcolor="#DDDDDD">',
	   textfield(-name=>$current,
			   -default=>$row[$i] || '',
			   -size=>'40',
			   -override=>'1'),
				'</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">';
    
  }
  
  print submit($TXT{'save'});
    
    print '</td>', end_form();
    print '<td bgcolor="#DDDDDD">&nbsp;</td></tr></table>';


  printfooter();
}

##############################################################

##################### VIEWADDRESS ############################

sub viewaddress {

  my @row = (); 
  my $found = 0;
  my $id = param("addrid");
  if ($id !~ /^(\d*)$/) { omailerror("Input format error"); }

  if ( -f "$prefsdir/addressbook2" ) {
    open (ABOOK,"$prefsdir/addressbook2") or
	 omailerror("Can't open your address book, but it exists!");
    while (<ABOOK>) {
	 @row = split(/\t/, $_);
	 if ($id == $row[0]) { $found = 1; last; }
    }
    close (ABOOK) or omailerror("Couldn't close your address book!");
  }
  if (!$found) { omailerror("Address ID $id not found"); }


  printheader();

  print '<table border="0" align="center" cellpadding="2">';
  print '<tr><td bgcolor="#002266" align="left">',
    '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',$TXT{''},'&nbsp;</b></font>',
	 "</td><td bgcolor=\"#002266\" align=\"right\"> <a href=\"$prefsurl?action=editaddresses&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage\"><font color=\"#ffffff\">[ ",$TXT{'cancel'}," ]</font></a> ";
  print '</td></tr>';

  ### Spacing
  print '<tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">';
  
  my $current = "";
  for (my $i = 0; $i<=$#abook_fields; $i++) {
    $current = $abook_fields[$i];
    if ($i == 0) { next; }
    
    print '<B>',$TXT{$current},':</B></td>',
	 '<td valign="middle" align="left" bgcolor="#DDDDDD">',$row[$i],
	   '&nbsp;</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="140">';
    
  }

    print '&nbsp;</td><td bgcolor="#DDDDDD">&nbsp;</td></tr></table>';

  printfooter();
}

##############################################################


################### MODADDRESS ###############################
sub modaddress {

  my $mode = shift;
  my @row = ();
  my @newrow = ();
  my $current = "";
  my $entry = "";
  my $line = "";
  my $id = param("addrid");
  my $tmpid = 0;
  my $highest_id = 100;
  my $empty = 1;

  if ($id !~ /^(\d*)$/ && $id) { omailerror("Input format error"); }
  
  # read address in an array
  for (my $i = 0; $i<=$#abook_fields; $i++) {
    $current = $abook_fields[$i];
    if ($i == 0 && $id) {
	 $current = "addrid";
    }

    $newrow[$i] = param("$current");
    $newrow[$i] =~ s/^\s*//; # strip beginning and trailing spaces
    $newrow[$i] =~ s/\s*$//;
    if ($newrow[$i] ne "") { $empty = 0; }
    $entry .= $newrow[$i] . "\t";
  }

  my %addresses;
  if ( -f "$prefsdir/addressbook2" ) {
    my $abooksize = ( -s "$prefsdir/addressbook2" );
    if ( (($abooksize +  length($entry) + 2) >= ($maxabooksize * 1024) ) && ($mode ne "delete") ) {
	 omailerror("Sorry, this would exceed the size limit for your
                          address book! Click
                          <a href=\"$prefsurl?action=editaddresses&amp;sessionid=$thissession&amp;sort=$sort&amp;firstmessage=$firstmessage\">here</a>
                          to return to your address book and delete some entries, then
                          try again.");
    }
    open (ABOOK,"$prefsdir/addressbook2") or
	 omailerror("Can't open your address book, but it exists!");

    while ($line = <ABOOK>) {
	 chomp ($line);
	 @row = split(/\t/, $line);
	 $tmpid = $row[0];
	 if ($tmpid) {
	   $addresses{"$tmpid"} = $line;
	   if ($tmpid > $highest_id) { $highest_id = $tmpid; }
	 }
    }
    close(ABOOK);

    if ($mode eq 'delete') {  # delete
	 writelog("deleting addr $id");
	 delete $addresses{"$id"};
	 #$addresses{"$id"} = "DELETED";
    } elsif ($id) { # edit
	 $addresses{"$id"} = $entry;
    } elsif (!$empty) { # add
	 $highest_id++;
	 $entry = substr($entry,1,2000000);
	 $entry = $highest_id . "\t" . $entry;
	 $addresses{"$highest_id"} = $entry;
    }

    open (ABOOK,">$prefsdir/addressbook2") or
	 omailerror("Couldn't write to your address book!");
    
    foreach $tmpid (sort(keys %addresses)) {
	 if ($tmpid !~ /^(\d*)$/) { writelog("XXX" . $tmpid . "YYY"); next; }	 
	 if ($addresses{"$tmpid"} eq "DELETED") { next; }
	 print ABOOK $addresses{"$tmpid"} . "\n" if $tmpid;
    }
    close (ABOOK) or omailerror("Couldn't close your address book!");

  } else {
    open (ABOOK, ">$prefsdir/addressbook2" ) or
	 omailerror("Can't create your address book!");
    $entry = substr($entry,1,2000000);
    print ABOOK "$highest_id\t$entry\n";
    close (ABOOK) or omailerror("Couldn't close your address book!");
  }

  if ((param("new") eq "1") && !$empty) {
     print "Location: $prefsurl?action=editaddress&sessionid=$thissession&addrid=$highest_id&sort=$sort&firstmessage=$firstmessage\n\n";
   
  } else {
    print "Location: $prefsurl?action=editaddresses&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage\n\n";
  }
}

################## END MODADDRESS ###########################

################### MODADDRESSLIST ##############################
sub modaddresslist {

  my ($realname, $address, $mail, $line, $tmpid);
  my @list = param("addresses");
  my %addresses; my @row; my @newrow;
  my $highest_id = 100;
  
  # findout highest_id
  if ( -f "$prefsdir/addressbook2" ) {
    open (ABOOK,"$prefsdir/addressbook2") or
	 omailerror("Can't open your address book, but it exists!");
    while ($line = <ABOOK>) {
	 chomp ($line);
	 @row = split(/\t/, $line);
	 $tmpid = $row[0];
	 if ($tmpid) {
	   if ($tmpid > $highest_id) { $highest_id = $tmpid; }
	 }
	
	# check if entry not already exists (name+mail)

    }
    close(ABOOK);
  }
  
  # add the new addresses
  
  open (ABOOK,">>$prefsdir/addressbook2") or
    omailerror("Couldn't write to your address book!");
  
  foreach $mail (@list) {
    chomp $mail;
    ($realname,$address) = split(/\|/, $mail);
    
    $realname =~ s/\t//;
    $realname =~ s/^\s*//; # strip beginning and trailing spaces from hash key
    $realname =~ s/\s*$//;
    #$address =~ s/[#&=\?]//g;
    
    if ($address) {
	 $highest_id++;
	 print ABOOK $highest_id++ . "\t\t\t" . $realname . "\t" . $address . "\t\t\t\t\t\t\t\t\t\t\t\t\t\n";
    }
  }
  
  close (ABOOK) or omailerror("Couldn't close your address book!");
  print "Location: $prefsurl?action=editaddresses&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage\n\n";
}

################## END MODADDRESSLIST ###########################


################### CONVERTADDR ##############################
sub convertaddr {

  my ($realname, $address, $mail, $line, $tmpid);
  my @list = param("addresses");
  my %addresses; my @row; my @newrow;
  my $highest_id = 100;
  
  if ( -f "$prefsdir/addressbook" ) {
    open (ABOOK,"$prefsdir/addressbook") or
	 omailerror("Can't open your address book, but it exists!");
    
    # add the new addresses
    
    open (ABOOK2,">>$prefsdir/addressbook2") or
	 omailerror("Couldn't write to your address book!");
    
    while ($line = <ABOOK>) {
	 chomp ($line);
	 $line =~ s/\t//g;
	 ($realname,$address) = split(/:/, $line);
	 
	 if ($address) {
	   $highest_id++;
	   print ABOOK2 $highest_id . "\t\t\t" . $realname . "\t" . $address . "\t\t\t\t\t\t\t\t\t\t\t\t\t\n";
	 }
    }
    
    close (ABOOK) or omailerror("Couldn't close your address book!");
    close (ABOOK2) or omailerror("Couldn't close your address book2!");

    # rename old book
    system("/bin/mv", "$prefsdir/addressbook", "$prefsdir/addressbook.old");
    writelog("conversion done.");
  }
}

################## END CONVERTADDR ###########################



###################### READPREFS #########################
sub readprefs {
   my ($key,$value);
   my %prefshash;
   my %addresses;
   if ( -f "$prefsdir/config" ) {
      open (CONFIG,"$prefsdir/config") or
         omailerror("Can't open your user config file, but it exists!");
      while (<CONFIG>) {

          # ($key, $value) = split(/=/, $_);
          # 2000/10/03 koj@syon.co.jp
          if(/^\s*(.+)\s*=\s*(.*)/) {
               ($key, $value) = ($1, $2);
          }
          # koj #chomp($value);

         $prefshash{"$key"} = $value;
      }
      close (CONFIG) or omailerror("Couldn't close your config file!");
   }

   my $sigfile = '';
   foreach $sigfile (@sigfiles) {
     if ( -f "$prefsdir/$sigfile" ) {
        $prefshash{$sigfile} = '';
        open (SIGNATURE, "$prefsdir/$sigfile") or
           omailerror("Can't open your signature ($sigfile), but file exists!");
        while (<SIGNATURE>) {
           $prefshash{$sigfile} .= $_;
        }
        close (SIGNATURE) or omailerror("Couldn't close your signature ($sigfile) file!");

        $prefshash{$sigfile} =~ s/\\'/'/g;
     }
   }
   return \%prefshash;
}
##################### END READPREFS ######################

###################### SAVEPREFS #########################
sub saveprefs {
   unless ( -d "$prefsdir" ) {
      mkdir ("$prefsdir", oct(700)) or
         omailerror("Couldn't create your user directory! ($prefsdir)");
   }


   # save signatures / headers

   foreach my $sigfile (@sigfiles) {
     open (SIGNATURE,">$prefsdir/$sigfile") or
        omailerror("Couldn't write your signature file!");
     my $value = param($sigfile) || '';
     $value =~ s/'/\\'/g;
     if (length($value) > 5000) {  # truncate signature to 5000 chars
        $value = substr($value, 0, 5000);
     }
     print SIGNATURE $value;
     close (SIGNATURE) or omailerror("Couldn't close your signature file!");
  }


   # save config

   open (CONFIG,">$prefsdir/config") or
      omailerror("Couldn't write your configuration file!");
   print CONFIG "lastchange=" . localtime() . "\n";
   print CONFIG "changedby=" . $ENV{REMOTE_ADDR} . "-" .$ENV{REMOTE_HOST} . "\n";


   foreach my $key (qw(realname realname2 realname3 realname4 realname5 replyto sort quotestyle lang numberofheaders autorefresh from from2 from3 from4 from5 send_cc_to use_cookies display_html_attach email_display_font_type empty_trash save_outgoing_mail)) {
      my $value = param("$key") || '';
      $value =~ s/[\n|=|\/|\|\`]//; # Strip out any sort of nastiness.
      if ($value) {
	print CONFIG "$key=$value\n";
      }
    }


   # for the sigs/headers
   foreach my $key (@sigfiles) {
      $key = "name" . $key;
      my $value = param("$key") || '';
      $value =~ s/[\n|=]//;
      if ($value) {
	print CONFIG "$key=$value\n";
      }
    }


   close (CONFIG) or omailerror("Couldn't close your configuration file!");
 


  printheader();
   print '<BR><BR><BR><BR><BR><BR>';
   print '<table border="0" align="center" width="40%" cellpadding="1" cellspacing="1">';

   print '<tr><td bgcolor="#DDDDDD" align="center">',
   '<font size="3"><b>OK!</b></font>',
   '</td></tr><tr><td align="center" bgcolor="#EEEEEE">';
   print "<BR>" . $TXT{'pref_saved'};
   print startform(-action=>"$scripturl");
   print hidden(-name=>'action',
		-default=>'displayheaders',
		-override=>'1');
   print hidden(-name=>'sessionid',
		-default=>$thissession,
		-override=>'1');
   print hidden(-name=>'sort',
		-default=>$sort,
		-override=>'1');
   print hidden(-name=>'firstmessage',
		-default=>$firstmessage,
		-override=>'1');
   print hidden(-name=>'folder',
		-default=>$folder,
		-override=>'1');
   print submit($TXT{'continue'});
   print end_form();
   print '</td></tr></table>';
   printfooter();
}
##################### END SAVEPREFS ######################
 
#################### ADDRESSBOOK #######################
sub addressbook {
   print header(-pragma=>'no-cache'),
   start_html(-"title"=>"Omail-webmail Address Book",
	      -author=>'emiller@hhs.net',
	      -BGCOLOR=>'#FFFFFF');

   my %addresses;
   my %addresses2;
   my ($name, $email);
   my $field=param("field");
   my $preexisting = param("preexisting") || '';
   $preexisting =~ s/(\s+)?,(\s+)?/,/g;
   print startform(-action=>"javascript:Update('" . $field . "')",
                   -name=>'addressbook'
                  );

   print '<table border="0" align="center" width="90%" cellpadding="1" cellspacing="1">';
   print '<tr><td colspan="2" bgcolor="#002266" align="left">',
   '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',uc($field),': ADDRESS BOOK</b></font>',
   '</td></tr>';

   if ( -f "$prefsdir/addressbook2" ) {
	open (ABOOK,"$prefsdir/addressbook2") or
	  omailerror("Can't open your address book, but it exists!");
	while (<ABOOK>) {
	  my @row = split(/\t/, $_);
	  $name = $row[3] . " " . $row[2];
   	  $addresses{"$name"} = $row[4];   # home
   	  $addresses2{"$name"} = $row[5];  # work

	}
	close (ABOOK) or omailerror("Couldn't close your address book!");


	foreach my $key (sort(keys %addresses)) {
	  next if (!$key || !$addresses{$key});

	  print '<tr><td bgcolor="#DDDDDD" width="20"><input type="checkbox" name="to" value="',
	  $addresses{"$key"}, '"';
	  if ($preexisting =~ s/$addresses{"$key"},?//g) {
	    print " checked";
	  }
	  print "></td><td width=\"100%\" bgcolor=\"#DDDDDD\">$key</td></tr>\n";

	  if ($addresses2{"$key"} ne "") {
	    print '<tr><td bgcolor="#DDDDDD" width="20"><input type="checkbox" name="to" value="',
		 $addresses2{"$key"}, '"';
	    if ($preexisting =~ s/$addresses2{"$key"},?//g) {
		 print " checked";
	    }
	    print "></td><td width=\"100%\" bgcolor=\"#DDDDDD\">$key [2]</td></tr>\n";
	  }

	}

   }

   print '</td></tr><tr><td align="center" colspan="2" bgcolor="#DDDDDD">';
   print '<input type="hidden" name="remainingstr" value="', $preexisting, '">';
   print '<input type="submit" name="mailto.x" value="OK">';
   print '<input type="button" value="',$TXT{'cancel'},'" onClick="window.close();">';
   print '</td></tr></table>';
   print '<script language="JavaScript">
      <!--
      function Update(whichfield)
      {
         var e2 = document.addressbook.remainingstr.value;
         for (var i = 0; i < document.addressbook.elements.length; i++)
         {
            var e = document.addressbook.elements[i];
            if (e.name == "to" && e.checked)
            {
               if (e2)
                  e2 += ",";
               e2 += e.value;
            }
         }
         if (whichfield == "to")
            window.opener.document.composeform.to.value = e2;
         else if (whichfield == "cc")
            window.opener.document.composeform.cc.value = e2;
         else
            window.opener.document.composeform.bcc.value = e2;
         window.close();
      }
      //-->
      </script>';
   print end_form();
   print end_html();
}
################## END ADDRESSBOOK #####################

############## VERIFYSESSION ########################
sub verifysession {

  omailerror("Session contains invalid data") unless
    (($thissession =~ /^([\w\.\-]+)$/) && ($thissession = $1));
  
  if ( !(-e "$omaildir/$thissession")) {
    writelog("User attempted to access non existant session - $thissession");
    omailerror("Session expired (deleted) - please login again.");
    exit 0;
  }
  
  if ( -M "$omaildir/$thissession" > ($sessiontimeout/60/24)) {
    writelog("User attempted to access timed-out session - $thissession - deleting file");
    $thissession =~ s/\///g;  # just in case someone gets tricky ...
    unlink "$omaildir/$thissession";
    writelog("User FORCED logout - $thissession");
    omailerror("Session expired (timeout) - please login again.");
    exit 0;
  }

  # session ok -> read maildir & uid if any

  my $tmp_line = "";
  open (SESSION,"$omaildir/$thissession") || omailerror("couldn't read session file 2");
  if ($tmp_line = <SESSION>) {
    # host
        chomp $tmp_line;
        my $host_test = $tmp_line;

    if (($host_test ne $ENV{REMOTE_ADDR}) && ($check_remote_host eq "yes")) {
         writelog("ALERT: wrong host : should be: $host_test, rejected value: " . $ENV{REMOTE_ADDR});
         unlink "$omaildir/$thissession";
         writelog("User FORCED logout - $thissession");
         omailerror("Security failure - Please conctact your sysadmin");
         exit(0);
    }

   }
   if ($tmp_line = <SESSION>) { 
     chomp $tmp_line; 
     $tmp_line =~ /(.*)/;      # untaint
     $maildir = $1;
   }
   if ($tmp_line = <SESSION>) { 
     chomp $tmp_line; 
     $tmp_line =~ /(.*)/;      # untaint
     $domainname = $1;
   }
   if ($tmp_line = <SESSION>) { 
     chomp $tmp_line; 
     $tmp_line =~ /(.*)/;      # untaint
     $user_uid = $1;
   }
   if ($tmp_line = <SESSION>) { 
     chomp $tmp_line; 
     $tmp_line =~ /(.*)/;      # untaint
     $user_gid = $1;
   }

  system("/usr/bin/touch","$omaildir/$thissession");
  
  #open (SESSION, '>' . $omaildir . $thissession) or omailerror("Couldn't open session file!\n");
  #print SESSION "$ENV{REMOTE_ADDR}";
  #close (SESSION);

  if ($cgi_mode="suid" && $user_uid && $user_gid && $user_uid ne "-1" && $user_gid ne "-1") {

    if ($user_uid < $minimal_uid || $user_gid < $minimal_gid) {
         writelog("ALERT: uid/gid too small : uid = $user_uid, gid = $user_gid");
         omailerror("UID Security failure - Please conctact your sysadmin");
         exit(0);
    }

    writelog("switching to user euid : UID -> $user_uid, GID -> $user_gid");
    writelog("--- [end of root log] --- continues in $maildir/.omail.log");
    $logfile = "$maildir/.omail.log";

    $< = $user_uid;
    $> = $user_uid;
    $( = $user_gid;
    $) = $user_gid;
    
    writelog("hello ------------------prefs------------- just logged in!");
  }
  
  return 1;
}
############# END VERIFYSESSION #####################


##################### WRITELOG ############################
sub writelog {
  unless ( ($logfile eq 'no') || ( -l "$logfile" ) ) {
    open (LOGFILE,">>$logfile") or omailerror("Couldn't open $logfile!\n");  
    my $timestamp = localtime();
    my $logaction = shift;
    my $loggeduser = $user || 'UNKNOWNUSER';
    print LOGFILE "$timestamp - [$$-P] ($ENV{REMOTE_ADDR},$ENV{REMOTE_HOST}) $loggeduser - $action - $logaction\n";
    close (LOGFILE);
  }
}
#################### END WRITELOG #########################

##################### OMAILERROR ##########################
sub omailerror {

  printheader();
  print '<br><br><table border="0" align="center"><tr><td valign="middle" align="center" bgcolor="#DDDDDD">';
  print '<FONT SIZE="+2"><B>Problem</B></FONT></td></tr>';
  print '<tr><td valign="middle" align="center" bgcolor="#EEEEEE" width="150" height="150">';
  print shift;
  print '</td></tr>';
  print '<tr><td valign="middle" align="center" bgcolor="#DDDDDD">';
  print '[ <a href="mailto:' . $scriptadmin . '?SUBJECT=Omail_Problem">support</a> ] &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;';
  print '[ <a href="' . $scripturl . '">login</a> ]</td></tr>';
 
  print '</table>';

  printfooter();
  exit(0);
  
}
################### END OMAILERROR #######################
##################### PRINTHEADER #########################

sub printheader {
   unless ($headerprinted) {

      $headerprinted = 1;
  
      if (!defined($TXT{'charset'})) { $TXT{'charset'} = "iso-8859-1"; }
      if (!defined($TXT{'charset_http'})) { $TXT{'charset_http'} = "iso-8859-1"; }

      my $charset_http ='';
      $charset_http = $TXT{'charset_http'} ? $TXT{'charset_http'}:$TXT{'charset'};

      print header(-pragma=>'no-cache',
                   -type=>"text/html; charset=$charset_http"),
                   # -type=>"text/html; charset=$TXT{'charset'}"),

            start_html(-"title"=>"oMail-webmail version $version [$lang]",
                       -BGCOLOR=>'#FFFFFF');

print <<STYLES;
<style type="text/css">
<!--
body {  font-family: $default_variable_font; font-size: 10pt}
th   {  font-family: $default_variable_font; font-size: 10pt; font-weight: bold; background-color: #D3DCE3;}
td   {  font-family: $default_variable_font; font-size: 10pt; }
test.td   {  font-family: $default_variable_font; font-size: 10pt;}
form   {  font-family: $default_variable_font; font-size: 10pt}
h1   {  font-family: $default_variable_font, sans-serif; font-size: 16pt; font-weight: bold}
A:link    {  font-family: $default_variable_font; font-size: 10pt; text-decoration: none; color: blue}
A:visited {  font-family: $default_variable_font; font-size: 10pt; text-decoration: none; color: blue}
A:hover   {  font-family: $default_variable_font; font-size: 10pt; text-decoration: underline; color: red}
A:link.nav {  font-family: $default_variable_font; color: #000000}
A:visited.nav {  font-family: $default_variable_font; color: #000000}
A:hover.nav {  font-family: $default_variable_font; color: red;}
.nav {  font-family: $default_variable_font; color: #000000}
-->
</style>
STYLES

      print "<FONT FACE=\"$default_variable_font\">\n";
   }
}
################### END PRINTHEADER #######################

################### PRINTFOOTER ###########################
sub printfooter {
   print "<p align=\"center\"><FONT SIZE=\"1\"><BR>
          <a href=\"$scripturl?action=about\" target=\"_new\">
          oMail-webmail</a><br>Version $version<BR>
          </FONT></FONT></p></BODY></HTML>";
}
################# END PRINTFOOTER #########################




################## JAPANESE HANDLING ######################
sub convert2euc {
  my $str = shift;
  $str = &Jcode::convert(\$str, 'euc', Jcode::getcode(\$str), "z");
  return $str;
}
sub convert2jis {
  my $str = shift;
  $str = &Jcode::convert(\$str, 'jis', Jcode::getcode(\$str), "z");
  return $str;
}
############## END JAPANESE HANDLING ######################

