#!/usr/bin/sudo /usr/bin/perl

#############################################################################
# omail.pl - Provides a web interface to user maildir spools                #
# Copyright (C) 2001 Olivier Mller <om@omnis.ch>                           #
# Copyright (C) 2000 Ernie Miller  (original GPL project: Neomail)          #
# See the CREDITS file for project contributors.                            #
#                                                                           #
# $Id: omail.pl,v 1.71 2001/08/26 12:35:27 swix Exp $                       #
# $Source: /cvsroot/omail/webmail/omail.pl,v $                              #
#                                                                           #
# This program is free software; you can redistribute it and/or             #
# modify it under the terms of the GNU General Public License               #
# as published by the Free Software Foundation; either version 2            #
# of the License, or (at your option) any later version.                    #
#                                                                           #
# This program is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
# GNU General Public License for more details.                              #
#                                                                           #
# You should have received a copy of the GNU General Public License         #
# along with this program; if not, write to the Free Software Foundation,   #
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.           #
#############################################################################

use strict;
use CGI qw(:standard);
use CGI::Carp qw(fatalsToBrowser);
use Date::Parse;
use MIME::Base64;
use Jcode;        # can be commented out if you don't need Japanese

CGI::nph();   # Treat script as a non-parsed-header script
$ENV{PATH} = ""; # no PATH should be needed

################# USER-CONFIGURABLE SECTION ##################

# paths

my $scriptpath = '/home/om/www/webmail';
my $sendmail = '/usr/sbin/sendmail';
my $omaildir = '/var/omail';               # to store sessions
my $logfile = '/var/omail/omail.log';      # logfile
my $langsubdir = "lang";                   # in $scriptpath

# urls

my $scripturl = '/webmail/omail.pl';
my $prefsurl = '/webmail/omail-prefs.pl';          # for preferences setup
my $button_url = '/webmail/buttons/';              # absolute url


# one-login  (cf. /contrib/onelogin)
my $secret_key = "CHANGETHAT_MUSTBEUNIQUE"; 
my $allow_one_login = 0;
my $one_login_startpage = "http://127.0.0.1/webmail/contrib/onelogin/login.html";
my $force_login_startpage_on_errors_or_logout = 0;

# mode

my $cgi_mode = "suid";                     # possible : user or suid

# needed if mode = suid:

my $sys_passwd = "/etc/passwd";
my $sys_shadow = "/etc/shadow"; 
my $sys_virtualdomains = "/var/qmail/control/virtualdomains";
my $minimal_uid = 100;                     # should be greater than 100
my $minimal_gid = 100;

# needed if mode = user:

my $domainname = '';                       # necessary for cgi_mode = user
my $homedir = '';                          # necessary for cgi_mode = user
my $passwdfile = '';                       # necessary for cgi_mode = user

# if you are using vmailmgr ( 

my $vmailmgr = 1;                          # if you use Bruce's package...
my $vmailmgr_old = 0;                      # ...with old style passwd file (flat text)
my $vauthenticate = '/usr/local/bin/vauthenticate';  
                                           # ...or the new passwd.cdb (binary)  ($vmailmgr_old = 0)

# if you use vpopmail by inter7.com (
my $vchkpw    = "/var/qmail/vpopmail/bin/vchkpw";
my $vpopmail  = 0;

# if you have a single user running all your domains, set to $vpopasswd 0
# and set $vpopuid to the uid and $vpopgid to the respective gid
# this brings a minimal performance win, as vchkpw does not have to be called twice
my $vpoppasswd=1;
my $vpopuid=1001;
my $vpopgid=83;

# ) end of vpopmail stuff


# default configuration

my $check_remote_host = "yes";             # check if remote host IP correspond with session IP
my $sessiontimeout = 45;                   # automatic logout after N minutes inactivity
my $numberofheaders = 15;
my $max_numberofheaders = 100;
my $quotestyle = 'quote';
my $maildir = "";
my $use_cookies = "yes";                       # saves only username
my $empty_trash = "yes";                       # to empty trash on login (recommended!)
my $save_outgoing_mail = "yes";                # save outgoing mails to "sent" folder
my $display_html_attach = "no";		   # display html attachments
my $email_display_font_type = "variable";
my $autorefresh = 0;                       # in minutes, turned off by default
my $scriptadmin = 'notdefined@notdefined.dom';            # for complaints / support
my $default_lang = "de";                   # default language
my @available_lang = ("en","fr","de", "it", "pl", "es", "dk", "cz", "cziso", "ja", "lt", "id", "tc", "no", "nl", "sr");
my %available_lang_names = (en => "English", 
                            it => "Italiano",
                            fr => "Francais",
                            pl => "Polski",
                            dk => "Dansk",
                            es => "Castellano",
                            cz => "Czech",
                            cziso => "Czech Iso",
			    ja => "Japanese",
			    lt => "Lithuanian",
			    id => "Indonesian",
			    tc => "Chinese",
			    no => "Norwegian",
			    nl => "Dutch",
			    sr => "Srpski",
                            de => "Deutsch");
my @available_quotestyle = ("quote", "outlook");
my %available_quotestyle_names = ("quote" => "Unix style \"> \"",
                                  "outlook" => "Outlook style");

my $timezone = 0;                          # experimental 
my $mimewords_decode = 1;                  # apply MIME::Words::decode_mimewords on subject and from/to/cc
                                           #  M=?ISO-8859-1?Q?=fcller_Olivier_?=<omueller@8304.ch> 
                                           #  -> Mller Olivier <omueller@8304.ch> 

# fonts stuff

my $default_variable_font = "Arial,Helvetica,sans-serif";
my $default_fixed_font = "Courier";

# will be shown on first screen just after login
my $welcome_message = "The webmail software has been updated, enjoy!";


############### END USER-CONFIGURABLE SECTION ################


################### PROGRAM INIT #############################

my $version = "0.98.4";

my %month = qw(Jan   1
               Feb   2
               Mar   3
               Apr   4
               May   5
               Jun   6
               Jul   7
               Aug   8
               Sep   9
               Oct   10
               Nov   11
               Dec   12);

if ($mimewords_decode) {
  use MIME::Words;
}

if ($allow_one_login) {
  use Digest::MD5  qw(md5_base64 md5_hex);
}

my @sigfiles = ('signature', 'signature2', 'signature3', 'signature4', 'signature5', 'header', 'header2', 'header3', 'header4', 'header5');

my $action = param("action");
my $thissession = param("sessionid");
my $user = $thissession;
$user =~ s/\-session\-0.*$//; # Grab userid from sessionid
($user =~ /^(.*)$/) && ($user = $1);  # otherwise, I get an Insecure dependency 
                                      # in open while running with -T switch in configsavex

writelog("started - version $version");

my $prefsdir = "";
my %prefs = ();
my $lang = "";
my ($user_uid, $user_gid, $user_realname);   # only for suid mode needed
my ($vuser_UID, $vuser_GID, $vuser_USER, $vuser_HOME, $vuser_MAILDIR, $vuser_VUSER);  # only for vmailmgr not old
my $loginmessage = "";
my $onelogin_active = 0;

# check 
if ($action && ($action ne "login") && ($action ne "onelogin") && ($action ne "about") && ($action ne "logout")) {
  verifysession();
}
  
$prefsdir = "$maildir/config"; 
%prefs = %{&readprefs};

# select language to use
$lang = $prefs{'lang'} if ($prefs{'lang'}); 
($lang =~ /^(.{2,10})$/) && ($lang = $1);            # untaint var

if (!$lang) { # no language found -> we try to guess the best one based on browser preferences
  my $langfound = 0;
  if (defined($ENV{'HTTP_ACCEPT_LANGUAGE'})) { 
    my $langaccept = $ENV{'HTTP_ACCEPT_LANGUAGE'} ;
    foreach my $tmplang (split(/\,/, $langaccept)) {
      foreach my $tmplangok (@available_lang) {
	$tmplang =~ s/\s//g; chomp $tmplang;
	if ($tmplangok eq substr($tmplang,0,2)) {  # this way fr-CH = fr-DE = fr, etc.
	  $lang = $tmplangok; 
	  $langfound = 1; last;
	}
      }
      if ($langfound) { last; }
    }
  }
}

$lang = $default_lang if (!$lang);  # if no valid language found, and no lang in user config, take default one (mostly english)


my %TXT;

# since perl 5.6 ??? maybe that default @INC is something else
# try to open lang file
#if (-f "$scriptpath/$langsubdir/$lang.lang") {
#  require "$scriptpath/$langsubdir/$lang.lang";
#  %TXT = getstrings();
#} else {
#  omailerror("couldn't open language file $scriptpath/$langsubdir/$lang.lang");
#}

if (-f "./$langsubdir/$lang.lang") {
  require "./$langsubdir/$lang.lang";
  %TXT = getstrings();
  $TXT{'charset'} ||= 'iso-8859-1';
} else {
  omailerror("couldn't open language file ./$langsubdir/$lang.lang");
}


$default_fixed_font = $TXT{'default_fixed_font'} if ($TXT{'default_fixed_font'});
$default_variable_font = $TXT{'default_variable_font'} if ($TXT{'default_variable_font'});


my $firstmessage;
if (param("firstmessage")) {
  $firstmessage = (param("firstmessage") -1); # -1 deals with "Jump" numbers
} else {
  $firstmessage = 0;
}

my $sort;
if (param("sort")) {
  $sort = param("sort");
} else {
  $sort = $prefs{"sort"} || 'date';
}

my $folder;
if (param("folder")) {
  $folder = param("folder");
  omailerror ("No such folder: $folder") unless ( ($folder eq 'INBOX') ||
						    ($folder eq 'SENT') || ($folder eq 'SAVED') || ($folder eq 'TRASH') );
} else {
  $folder = "INBOX";
}

my $selected_mail_font = "";

$numberofheaders = $prefs{"numberofheaders"} if (defined($prefs{"numberofheaders"}));
$autorefresh = $prefs{"autorefresh"} if (defined($prefs{"autorefresh"}));
$use_cookies = $prefs{"use_cookies"} if (defined($prefs{"use_cookies"}));
$empty_trash = $prefs{"empty_trash"} if (defined($prefs{"empty_trash"}));
$save_outgoing_mail = $prefs{"save_outgoing_mail"} if (defined($prefs{"save_outgoing_mail"}));
$display_html_attach = $prefs{"display_html_attach"} if (defined($prefs{"display_html_attach"}));
$email_display_font_type = $prefs{"email_display_font_type"} if (defined($prefs{"save_outgoing_mail"}));

if ($email_display_font_type eq "fixed") {
	$selected_mail_font = $default_fixed_font;
} else {
	$selected_mail_font = $default_variable_font;
}

# once we print the header, we don't want to do it again if there's an error
my $headerprinted = 0;
my $total_size = 0;
my $total_new_mails = 0;
my $savedattsize;
my $attlimit;

################### END PROGRAM INIT #########################

########################## MAIN ##############################
if ($action) {  # an action has been chosen
  if ($action =~ /^(\w+)$/) {
    $action = $1;
    if ($action eq "login") {
      login();
    } elsif ($action eq "onelogin") {
	 onelogin();
    } elsif ($action eq "displayheaders") {
      displayheaders();
    } elsif ($action eq "readmessage") {
      readmessage();
    } elsif ($action eq "deletemessage") {
      deletemessage();
    } elsif ($action eq "viewattachment") {
      viewattachment();
    } elsif ($action eq "composemessage") {
      composemessage();
    } elsif ($action eq "sendmessage") {
      sendmessage();
    } elsif ($action eq "savemessage") {
      savemessage();
    } elsif ($action eq "toggle") {
      toggleflag();
    } elsif ($action eq "logout") {
      logout();
    } elsif ($action eq "about") {
      about();
    } elsif ($action eq "grabaddr") {
      grabaddr();
    } else {
	 omailerror("Action not valid");
    }
  } else {
    omailerror("Action has illegal characters");
  }
} else {            # no action has been taken, display login page

  # get userid in cookie if any
  my $cookie_userid = cookie(-name=>'omaillogin');

  printheader();

  print "<br><br><br><FONT FACE=\"$default_variable_font\">\n";
  print '<table border="0" align="center" cellpadding="2" width="300">
          <tr><td valign="middle" align="center" bgcolor="#DDDDDD">';
  print '<FONT SIZE="+2"><B>oMail-webmail - ',$TXT{'login'},'</B></FONT>';
  print '</td></tr><tr><td valign="middle" align="center" bgcolor="#EEEEEE">';
  print '<form method="post" action="',$scripturl,'" name="login1" enctype="application/x-www-form-urlencoded">';
  print hidden("action","login");
  print hidden("lang","$lang");
  print '<table border="0">';
  print p("<tr><td align=\"right\">$TXT{'userid'}:</td><td> ", textfield(-name=>'userid',
									 -default=>$cookie_userid,
									 -size=>'22',
									 -override=>'1'),
	  "</td></tr><tr><td align=\"right\">$TXT{'passwd'}:</td><td> ", password_field(-name=>'password',
											-default=>'',
											-size=>'22',
											-override=>'1'),'</td></tr></table>',
	  submit("$TXT{'login'}"));
  print end_form();
  print '</td></tr>';
  
  print'<tr><td valign="middle" align="right" bgcolor="#DDDDDD">';
  print '&nbsp;</td></tr>';
   
  print '</table>';

  if ($cookie_userid ne "") {
    print <<JSCRIPT1;
      <SCRIPT LANGUAGE="JavaScript"><!--
	   document.login1.password.focus();
      //--></SCRIPT>
JSCRIPT1
  } else {
    print <<JSCRIPT2;
    <SCRIPT LANGUAGE="JavaScript"><!--
	 document.login1.userid.focus();
    //--></SCRIPT>
JSCRIPT2
  }


  print "<p align=\"center\"><FONT SIZE=\"1\"><BR>
          <a href=\"$scripturl?action=about\" target=\"_new\">
          About oMail-webmail</a><br>Version $version</font></p>";
  print end_html();
}
###################### END MAIN ##############################

####################### LOGIN ########################
sub login {
  my $userid = param("userid");
  $userid =~ s/\s//g;            # remove blanks

  if (($vmailmgr || $vpopmail) && $cgi_mode eq "suid") {

    if (!($userid =~ /(.*)\@(.*)/)) {
      omailerror("domain name is missing : format userid\@domain");
    } else {
      $userid = $1;
      $domainname = $2;
    }
  }

  
  my $password = param("password");
  $password =~ /^(.+)$/;
  $password = $1;

  # 8 possible cases : 
  #
  # user cgi mode   -> $cgi_mode = "user"
  # 
  # -> $homedir and $passwd are correct 
  #
  # a) just a Maildir : no auth                                                vmailmgr=0  vmailmgr_old=0
  # b) vmailmgr old : against $homedir/passwd                                  vmailmgr=1  vmailmgr_old=0
  # c) vmailmgr new : against $homedir/passwd.cdb via vcheckuserid             vmailmgr=1  vmailmgr_old=1
  #
  # suid cgi mode   -> $cgi_mode = "suid"
  #
  # -> findout $homedir and $passwd using /etc/passwd and /var/qmail/control/virtualdomains
  #
  # d) just qmail  ->  auth against /etc/passwd and /etc/shadow                vmailmgr=0  vmailmgr_old=0
  # e) vmailmgr old -> auth against /homedir/passwd                            vmailmgr=1  vmailmgr_old=0
  # f) vmailmgr new -> auth against /homedir/passwd.cdb via vcheckuserid       vmailmgr=1  vmailmgr_old=1
  #
  # x1) vpopmail -> auth against $vchkpw
  # x2) vpopmail with /etc/passwd accounts -> auth against $vchkpw

  
  if ($cgi_mode eq "suid") {

    my $username_found = 0;
    my $tmp_line = "";
    my $username = "";

    if ($vmailmgr && $vmailmgr_old && !$vpopmail) {  # multiple domain system -> we need to know which user own the domain 
      
      # check if domain exists. if yes, get username
      open(VDOM, $sys_virtualdomains) || omailerror("cannot open $sys_virtualdomains");
      while ($tmp_line = <VDOM>) {
	chomp $tmp_line;
	if ($tmp_line =~ /^$domainname:(.*)/) { 
	    $username = $1;	# Untainted, too
	  
	  if ($username) {
	    writelog("login: username found for domain $domainname : $username");
	    $username_found = 1;
	  } else {
	    omailerror("empty username for domain $domainname");
	  }
	  last;
	}
      }
      close (VDOM);
      if (!$username_found) {
	omailerror("domain $domainname not found!");
      }
    } else {  # single domain system (don't forget to set up $domainname)
      
      $username = $userid;
      
    }

    if (!($vmailmgr && !$vmailmgr_old) &&!$vpopmail) {      
      
      # get homedir, numerical uid and gid
      
      open(PWS, $sys_passwd) || omailerror("cannot open $sys_passwd");
      while ($tmp_line = <PWS>) {
	chomp $tmp_line;
	if ($tmp_line =~ /^$username:(.*):(.*):(.*):(.*):(.*):(.*)$/) {
	  ($user_uid, $user_gid, $homedir, $user_realname) =
	    ($2, $3, $5, $4);	# Parse and untaint	  
	  writelog("user info : uid = $user_uid , gid = $user_gid, homedir = $homedir, descr = $user_realname");
	  last;
	}
      }
      close(PWS);
      
      if (!($user_uid)) { omailerror("[9] user $username not in found ..."); }
      
      if (!$vmailmgr) {
	$passwdfile = $sys_shadow;
      } else {
	$passwdfile = $homedir . "/passwd";
      }
    } 
  }


  if (checklogin($userid,$password,$domainname)) {

    if ((!$vmailmgr_old && $vmailmgr) || $vpopmail) {  # get the uid/gid from vauthenticate output
      $homedir = $vuser_HOME;
      if ($vuser_MAILDIR && $vuser_MAILDIR =~ /^(\.\/)?(.*)/) {
	  $maildir = $vuser_HOME . '/' . $2;
	  writelog("Maildir set to $maildir");
      }
      $user_uid = $vuser_UID;
      $user_gid = $vuser_GID;
    }

    # findout Maildir path
    
    if ($vmailmgr && !$vpopmail) {
	if ($vmailmgr_old) {	# new set maildir above
	    my $qmail_userid = $userid;   # replace "." by ":" in mailbox name
	    $qmail_userid =~ s/\./\:/g;    
	    $maildir = $homedir . "/users/" . $qmail_userid;
	}
    } else {
      $maildir = $homedir . "/Maildir";
    }      
    
    $thissession = $userid . "-session-" . rand();
    writelog("Successful login for user $userid \@ $domainname - $thissession");
    cleanupoldsessions();
    
    open (SESSION, '>' . $omaildir . "/". $thissession) or omailerror("Couldn't write to session file!\n");
    print SESSION "$ENV{REMOTE_ADDR}\n";
    print SESSION "$maildir\n";
    print SESSION "$domainname\n";

    if ($user_uid && $user_gid) {
      print SESSION "$user_uid\n$user_gid\n";
    } else {
      print SESSION "-1\n-1\n";
    }
		
    close (SESSION);

    if ($cgi_mode="suid" && $user_uid && $user_gid && $user_uid ne "-1" && $user_gid ne "-1") {
      writelog("switching to user euid : UID -> $user_uid, GID -> $user_gid");
      writelog("--- [end of root log] --- continues in $maildir/.omail.log");
      $logfile = "$maildir/.omail.log";
      
      $< = $user_uid;
      $> = $user_uid;
      $( = $user_gid;
      $) = $user_gid;
      

	 # make sure the logfile is not bigger than 20kb

	 if ((-s $logfile) > 20000) {
	   writelog("rotating logs.");
	   system("/bin/mv", "-f", "$logfile", "$logfile.old");
	   writelog("log rotation done.");
	 }
	 
      writelog("hello ------- just logged in! ------ session: $thissession");

    }
  
    
    $user = $userid;
    $prefsdir = "$maildir/config"; 


    
    if ( -d "$prefsdir" ) {
      %prefs = %{&readprefs};
      $sort = $prefs{"sort"} || 'date';
      $numberofheaders = $prefs{"numberofheaders"} if (defined($prefs{"numberofheaders"}));
      $autorefresh = $prefs{"autorefresh"} if (defined($prefs{"autorefresh"}));
      $use_cookies = $prefs{"use_cookies"} if (defined($prefs{"use_cookies"}));
      $empty_trash = $prefs{"empty_trash"} if (defined($prefs{"empty_trash"}));
      $save_outgoing_mail = $prefs{"save_outgoing_mail"} if (defined($prefs{"save_outgoing_mail"}));
      $email_display_font_type = $prefs{"email_display_font_type"} if (defined($prefs{"save_outgoing_mail"}));
      $display_html_attach = $prefs{"display_html_attach"} if (defined($prefs{"display_html_attach"}));


      # empty trash if necessary:
	writelog("$empty_trash --- $maildir");
      emptytrash($maildir) if ($empty_trash eq "yes");

      if ($prefs{'lang'}) {
	$lang = $prefs{'lang'};
	($lang =~ /^(.{2,10})$/) && ($lang = $1);            # untaint var
	# try to open lang file
	if (-f "./$langsubdir/$lang.lang") {
	  require "./$langsubdir/$lang.lang";
	  %TXT = getstrings();
	  $TXT{'charset'} ||= 'iso-8859-1';
	} else {
	  omailerror("#1: couldn't open language file lang/$lang.lang");
	}
      }

      $default_fixed_font = $TXT{'default_fixed_font'} if ($TXT{'default_fixed_font'});
      $default_variable_font = $TXT{'default_variable_font'} if ($TXT{'default_variable_font'});

      if ($email_display_font_type eq "fixed") {
          $selected_mail_font = $default_fixed_font;
      } else {
          $selected_mail_font = $default_variable_font;
      }

    my $lastloginfile = "$prefsdir/lastlogin";
    
    # get last_login infos, and save new login informations

    my $logincount = 0;
    if ( -f "$lastloginfile" ) {
	 if (open(LASTLOG, "$lastloginfile")) {
	   my $lasttmp = <LASTLOG>;
	   chomp $lasttmp;
	   my @lastarr = split(/\t/, $lasttmp);
	   # format: unixtime   ip   host   
	   $loginmessage = "<b>" . $TXT{'welcome_back'} . " - " . $TXT{'lastlogin'} . ": " . $TXT{'lastlogin_on'}
		. " " . localtime($lastarr[0]) . " " . $TXT{'lastlogin_from'} . " " . $lastarr[1];
	   if ($lastarr[2]) { $loginmessage .= " (" . $lastarr[2] . ")"; }
	   $loginmessage .= " - " . $TXT{'days_ago_before'} . " " 
		. (int(time/86400) - int($lastarr[0]/86400))  
		. " " . $TXT{'days_ago_after'} . "</b>";
	   close(LASTLOG);
	   if ($lastarr[3]) { $logincount = $lastarr[3]; }
	 }
    }

    # is there a welcome_message ?
    if ($welcome_message) {
         if ($loginmessage && $welcome_message) {
	      $loginmessage .= '<br>';
         }
         $loginmessage .= $welcome_message;
    }
    

    # write current infos
    
    if (open(LASTLOG, ">$lastloginfile")) {
	 $logincount++;
	 print LASTLOG time . "\t" . $ENV{'REMOTE_ADDR'} . "\t" . $ENV{'REMOTE_HOST'} . "\t$logincount\t\n";
	 close(LASTLOG);
    } else {
	 writelog("login: can't write to $lastloginfile");
    }
	 
	 # need for multiple-attachment support
	 if (!(-d "$maildir/omailtmp")) { 
	   system("/bin/mkdir $maildir/omailtmp");
	   writelog("getattlist: created directory $maildir/omailtmp");
	 }
	 
	 displayheaders();
	 
    } else {
      firsttimeuser();
    }
    
  } else {
    omailerror($TXT{'sorry_wrong_pw'});
    exit 0;
  }
}
#################### END LOGIN #####################

sub logout {
   omailerror($TXT{'err_session_invalid'}) unless
      (($thissession =~ /^(.+?\-\d?\.\d+)$/) && ($thissession = $1));
   $thissession =~ s/\///g;  # just in case someone gets tricky ...

   unlink "$omaildir/$thissession";
   writelog("User logout - $thissession");

   if ($onelogin_active || $force_login_startpage_on_errors_or_logout) {
	print "Location: $one_login_startpage\n\n";
   } else {
	print "Location: $scripturl\n\n";
   }
}
################## END LOGOUT ######################

#################### CHECKLOGIN ####################
sub checklogin {
  my ($username, $password, $domainname, $usr, $pswd);
  my $passcorrect = 1; # default to correct, see end of sub for reason
  $username = shift;
  $password = shift;
  $domainname = shift;

  if ($vmailmgr && !$vmailmgr_old && !$vpopmail) {
    
    my $command = 'echo -e "'. $domainname .'\000'. $username .'\000'. $password .'\000" | ' .$vauthenticate;
    my $output = `$command`;

    if (!$output) { 
      $passcorrect = 0;
      writelog("vauth login FAILED for $username \@ $domainname");
      
    } else {
      
      # login ok!
      
      foreach my $dataline (split(/\n/,$output)) {
	
	chomp $dataline;
	$dataline =~ /(.*?)=(.*)/;
	my ($entname, $entvalue) = ($1, $2); # Parse and untaint

	$vuser_UID = $entvalue if $entname eq "UID";
	$vuser_GID = $entvalue if $entname eq "GID";
	$vuser_USER = $entvalue if $entname eq "USER";
	$vuser_HOME = $entvalue if $entname eq "HOME";
	$vuser_MAILDIR = $entvalue if $entname eq "MAILDIR";
	$vuser_VUSER = $entvalue if $entname eq "VUSER";
      }	

      if ($vuser_UID && $vuser_GID && $vuser_MAILDIR && $vuser_HOME && $vuser_USER) {
	writelog("vauth login OK for $username \@ $domainname : $vuser_UID,$vuser_GID,$vuser_USER,$vuser_MAILDIR,$vuser_VUSER");
      } else {
	$passcorrect = 0;
	writelog("vauth login PROBLEM with $username \@ $domainname");
      }
    }

    } elsif ($vpopmail) {
      my $tmpcmd='echo -e "'.$username."\@".$domainname.'\000'.$password.'\000\000" | '.$vchkpw.' ';
      my $home_cmd=$tmpcmd.'/bin/pwd 3<&0';
      $vuser_HOME=`$home_cmd`;
      $vuser_HOME=~s/\n//g;
      $passcorrect=0;
      if ($vuser_HOME ne ""){
        # login seems to be correct, if /etc/passwd accounts are used get uid & gid, otherwise set to default
        if ($vpoppasswd){
          my $gid_cmd=$tmpcmd.' /usr/bin/id 3<&0';
          my $tmpid=`$gid_cmd`;
          if ($tmpid ne "" && $tmpid=~/uid=(\d+)\(.+\) gid=(\d+)\(.+\).*/){
            $vuser_UID=$1;
            $vuser_GID=$2;
            $passcorrect=1;
          } else {
            writelog("vpopmail - unexpected /usr/bin/id output: $tmpid");
          }
        } else {
          $vuser_UID=$vpopuid;
          $vuser_GID=$vpopgid;
          $passcorrect=1;
        }
      } 
      writelog("vpopmail - uid: $vuser_UID - gid: $vuser_GID - home: $vuser_HOME "); 

  } else { 
    
    # works for both /etc/shadow and vmailmgr oldstyle /home/USER/passwd 
    # authentication
    
    open (PASSWD, $passwdfile) or omailerror("Couldn't open $passwdfile!\n");
    while (defined(my $tmp_line = <PASSWD>)) {
	if ($tmp_line =~ /^$username:(.*?):/) {
	    $pswd = $1;
	    $usr = $username;
	    last;
	}
    }
    close (PASSWD);
    
    chomp $pswd;
    
    if (($usr ne $username) or (crypt($password, $pswd) ne $pswd)) {
      $passcorrect = 0; # User/Pass combo is WRONG!
    }   
  }

  if (!$password) {  # no null passwords allowed
    $passcorrect = 0;
  }
  
  unless ($passcorrect) {
    writelog("FAILED login - $username");
  }
  return $passcorrect; 
}
################ END CHECKLOGIN #####################

################ CLEANUPOLDSESSIONS ##################
sub cleanupoldsessions {
  opendir (OMAILDIR, "$omaildir") or omailerror("Couldn't opendir $omaildir!\n");
  while (defined(my $sessionid = readdir(OMAILDIR))) {
    if ($sessionid =~ /^(\w+\-session\-0\.\d*.*)$/) {
      $sessionid = $1;

      if ( -M "$omaildir/$sessionid" > ($sessiontimeout/60/24) ) {
	writelog("Cleaning up old session file - $sessionid");
	unlink "$omaildir/$sessionid";
      }
    }
  }
  closedir (OMAILDIR);
}
############## END CLEANUPOLDSESSIONS ################

############## VERIFYSESSION ########################
sub verifysession {

  omailerror("Session contains invalid data") unless
    (($thissession =~ /^([\w\.\-]+)$/) && ($thissession = $1));
  
  if ( !(-e "$omaildir/$thissession")) {
    writelog("User attempted to access non existant session - $thissession");
    omailerror("Session expired (deleted) - please login again.");
    exit 0;
  }
  
  if ( -M "$omaildir/$thissession" > ($sessiontimeout/60/24)) {
    writelog("User attempted to access timed-out session - $thissession - deleting file");
    $thissession =~ s/\///g;  # just in case someone gets tricky ...
    unlink "$omaildir/$thissession";
    writelog("User FORCED logout - $thissession");
    omailerror("Session expired (timeout) - please login again.");
    exit 0;
  }

  # session ok -> read maildir & uid if any

  my $tmp_line = "";
  open (SESSION,"$omaildir/$thissession") || omailerror("couldn't read session file 2"); 
  if ($tmp_line = <SESSION>) { 
    # host
	chomp $tmp_line;
	my $host_test = $tmp_line;

    if (($host_test ne $ENV{REMOTE_ADDR}) && ($check_remote_host eq "yes")) {
	 writelog("ALERT: wrong host : should be: $host_test, rejected value: " . $ENV{REMOTE_ADDR});
	 unlink "$omaildir/$thissession";
         writelog("User FORCED logout - $thissession");
	 omailerror("Host Security failure - Please conctact your sysadmin");
	 exit(0);    
    }

  }
  if ($tmp_line = <SESSION>) { 
    chomp $tmp_line; 
    $tmp_line =~ /(.*)/;	# untaint
    $maildir = $1;
  }
  if ($tmp_line = <SESSION>) { 
    chomp $tmp_line; 
    $tmp_line =~ /(.*)/;	# untaint
    $domainname = $1;
  }
  if ($tmp_line = <SESSION>) { 
    chomp $tmp_line; 
    $tmp_line =~ /(.*)/;	# untaint
    $user_uid = $1;
  }
  if ($tmp_line = <SESSION>) { 
    chomp $tmp_line; 
    $tmp_line =~ /(.*)/;	# untaint
    $user_gid = $1;
  }

  system("/usr/bin/touch","$omaildir/$thissession");
  
  #open (SESSION, '>' . $omaildir . $thissession) or omailerror("Couldn't open session file!\n");
  #print SESSION "$ENV{REMOTE_ADDR}";
  #close (SESSION);

  if ($cgi_mode="suid" && $user_uid && $user_gid && $user_uid ne "-1" && $user_gid ne "-1") {

    if ($user_uid < $minimal_uid || $user_gid < $minimal_gid) {
	 writelog("ALERT: uid/gid too small : uid = $user_uid, gid = $user_gid");
	 omailerror("UID Security failure - Please conctact your sysadmin");
	 exit(0);
    }

    writelog("switching to user euid : UID -> $user_uid, GID -> $user_gid");
    writelog("--- [end of root log] --- continues in $maildir/.omail.log");
    $logfile = "$maildir/.omail.log";

    $< = $user_uid;
    $> = $user_uid;
    $( = $user_gid;
    $) = $user_gid;
    
    writelog("hello from verifysession ---- connected! - action: $action - session: $thissession");
  }
  
  return 1;
}
############# END VERIFYSESSION #####################

################ DISPLAYHEADERS #####################
sub displayheaders {
  printheader("displayheaders");
  
  my ($bgcolor, $status, $message_size, $searchstring);

  if (defined(param("searchstring"))) {
    $searchstring = param("searchstring");
  }

  my @headers = @{&getheaders($user, $searchstring)};

  $numberofheaders = $max_numberofheaders if ($numberofheaders > $max_numberofheaders); 
  my $page_total = ($#headers+1) / $numberofheaders;
  $page_total = int($page_total) + 1 if ($page_total != int($page_total));
  
  if (defined(param("custompage"))) {
    $firstmessage = (param("custompage")-1)*$numberofheaders;
  }     

  
  if ($firstmessage > $#headers) {
    $firstmessage = ($page_total-1)*$numberofheaders;
  }
  
  if ($firstmessage < 0) {
    $firstmessage = 0;
  }

  my $lastmessage = $firstmessage + $numberofheaders - 1;
  if ($lastmessage > $#headers) {
    $lastmessage = $#headers;
  }
  
  my $base_url = "$scripturl?sessionid=$thissession&sort=$sort";
  
  my $page_nb;
  if ($#headers > 0) {
    $page_nb = ($firstmessage+1) * ($#headers / $numberofheaders) / $#headers;
    $page_nb = int($page_nb) + 1 if ($page_nb != int($page_nb));
  } else {
    $page_nb = 1;
  }
  
  $page_nb = int($page_nb) + 1 if ($page_nb != int($page_nb));
  
  if ($total_size > 1048575){ 
    $total_size = int(($total_size/1048576)+0.5) . " MB"; 
  } elsif ($total_size > 1023) {
    $total_size =  int(($total_size/1024)+0.5) . " kB"; 
  } else {
    $total_size = $total_size . " " . $TXT{'bytes'};
  }
  
  #     print '</td></tr><tr><td>';  # page table
  
  # headers table
    
  print '<table border="0" align="center" width="95%" cellpadding="1" cellspacing="1">';
  
  # first row - mbox name / message nb
  
  print '<tr><td colspan="7" bgcolor="#002266" align="center"><table border="0" width="100%" cellpadding="0" cellspacing="0">',
  '<tr><td align="left" width="33%"><font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',uc($TXT{$folder}),'</b>';
  
  print "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; ";  
  print " [ <a href=\"$base_url&action=displayheaders&firstmessage=0&folder=INBOX\"><font color=\"#ffffff\">$TXT{'INBOX'}</font></a> ] " 
    if $folder ne "INBOX";
  print " [ <a href=\"$base_url&action=displayheaders&firstmessage=0&folder=SENT\"><font color=\"#ffffff\">$TXT{'SENT'}</font></a> ] " 
    if $folder ne "SENT";
  print " [ <a href=\"$base_url&action=displayheaders&firstmessage=0&folder=SAVED\"><font color=\"#ffffff\">$TXT{'SAVED'}</font></a> ] " 
    if $folder ne "SAVED";
  print " [ <a href=\"$base_url&action=displayheaders&firstmessage=0&folder=TRASH\"><font color=\"#ffffff\">$TXT{'TRASH'}</font></a> ] "
    if $folder ne "TRASH";
  
  print '  </font></td><td align=right width="33%"><font color="#ffffff" face="',$default_variable_font,'" size="3">',
  "<b>", $firstmessage + 1, " - ", $lastmessage + 1, " $TXT{'of'} ", $#headers+1, " $TXT{'messages'}";
  
  if ($total_new_mails) {
    print " ($total_new_mails $TXT{'new'})" if ($total_new_mails == 1);
    print " ($total_new_mails $TXT{'news'})" if ($total_new_mails > 1);
  } 

  print " - $TXT{'mbox_size'}: $total_size</b></font></td></tr></table></td></tr>";


  # info row - optional

  if ($loginmessage) {
    print '<tr><td bgcolor="lightblue" colspan="7" align="center">';
    print '<font face="',$default_variable_font,'" size=2>';
    print $loginmessage;
    print '</td></tr>';
  }

  
  # second row : navigatation
  
  print '<tr><td bgcolor="#dcdcdc" colspan=7><table border="0" width="100%" cellpadding="0" cellspacing="0">';
  print '<td align=center><font face="',$default_variable_font,'" size=2>';
  
  print start_form();
  
  print '<table border=0 width="100%">';
  print '<tr><td valign="middle" align="left">';
  print "<a href=\"$base_url&folder=$folder&action=composemessage&firstmessage=",$firstmessage+1,"\">$TXT{'compose'}</a> | ";
  print "<a href=\"$prefsurl?sessionid=$thissession&folder=$folder&sort=$sort&firstmessage=",$firstmessage+1,"\">$TXT{'userconfig'}</a> | ";
  print "<a href=\"$prefsurl?action=editaddresses&folder=$folder&sessionid=$thissession&sort=$sort&firstmessage=",$firstmessage+1,"&lang=$lang\">$TXT{'addressbook'}</a> | ";
  print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=",$firstmessage+1,"\">$TXT{'refresh'}</a> | ";

  print "<a href=\"$base_url&folder=$folder&action=logout&firstmessage=",$firstmessage+1,"\">$TXT{'logout'}</a>";
  


  print '</td><td valign="middle" align="center">';
  
  print start_form();
  print '',$TXT{'find'} ,': ';
  print hidden(-name=>'action',
	       -default=>'displayheaders',
	       -override=>'1');
  print hidden(-name=>'sessionid',
	       -default=>$thissession,
	       -override=>'1');
  print hidden(-name=>'folder',
	       -default=>$folder,
	       -override=>'1');
  print hidden(-name=>'sort',
	       -default=>$sort,
	       -override=>'1');
  print textfield(-name=>'searchstring',
		  -default=>$searchstring,
		  -size=>'8',
		  -override=>'1');

  print '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td>', end_form(), '';





  print '</td><td valign="middle" align="center"><nobr>';
  
  print start_form();
  if ($firstmessage != 0) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=0\">";
    print "<img src=\"$button_url/first.gif\" border=\"0\" alt=\"First page\"></a>";
  } else {
    print "<img src=\"$button_url/first-grey.gif\" border=\"0\" alt=\"\">";      
  }
  
  if (($firstmessage - $numberofheaders) >= 0) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=" . ($firstmessage +1 - $numberofheaders) . "\">";
    print "<img src=\"$button_url/left.gif\" border=\"0\" alt=\"Previous page\"></a>";
  } else {
    print "<img src=\"$button_url/left-grey.gif\" border=\"0\" alt=\"\">";      
  }
  
  print '[',$TXT{'page'} ,' ';
  print hidden(-name=>'action',
	       -default=>'displayheaders',
	       -override=>'1');
  print hidden(-name=>'sessionid',
	       -default=>$thissession,
	       -override=>'1');
  print hidden(-name=>'folder',
	       -default=>$folder,
	       -override=>'1');
  print hidden(-name=>'sort',
	       -default=>$sort,
	       -override=>'1');
  print textfield(-name=>'custompage',
		  -default=>$page_nb,
		  -size=>'2',
		  -override=>'1');
  print ' ', $TXT{'of'} ,' ', $page_total ,']';
  
  if (($firstmessage + $numberofheaders) <= $#headers) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=" . ($firstmessage + 1 + $numberofheaders) ."\">";
    print "<img src=\"$button_url/right.gif\"  border=\"0\" alt=\"Next page\"></a>";
  } else {
    print "<img src=\"$button_url/right-grey.gif\" border=\"0\" alt=\"\">"; 
  }
  
  if (($firstmessage + $numberofheaders) <= $#headers) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=" . ($#headers+1). "\">";
    print "<img src=\"$button_url/last.gif\" border=\"0\" alt=\"Last page\"></a>";
  } else {
    print "<img src=\"$button_url/last-grey.gif\" border=\"0\" alt=\"\">";      
  }
  
  print '</nobr></td>', end_form(), '<td valign="middle" align="right">';
  # delete button
  
  print start_form(-onSubmit=>"return confirm('" . $TXT{'js_confirm_delete'} . "')",
		   -name=>'Formdel');
#  print hidden(-name=>'action',
#	       -default=>'deletemessage',
#	       -override=>'1');
  print hidden(-name=>'sessionid',
	       -default=>$thissession,
	       -override=>'1');
  print hidden(-name=>'folder',
	       -default=>$folder,
	       -override=>'1');
  print hidden(-name=>'firstmessage',
	       -default=>$firstmessage +1,
	       -override=>'1');
  print hidden(-name=>'sort',
	       -default=>$sort,
	       -override=>'1');
  
  print '</td></tr></table>';
  print '</td><td>&nbsp;</td>';
  print '</tr></table></td></tr>';
  
  # row names
  
  print '<tr><td colspan="7" bgcolor="#ffffff">&nbsp;</td></tr>';

  print '<tr><td valign="middle" width="20" bgcolor="#D3DCE3">',$TXT{'nr'},'</td>';
  
  
  print '<td valign="middle" width="50" bgcolor="#D3DCE3"><B>';
  print "<a href=\"$scripturl?action=displayheaders&folder=$folder&firstmessage=",
  $firstmessage+1,"&sessionid=$thissession&sort=";
  
  if ($sort eq "status") {
    print "status_rev\">$TXT{'status'} <IMG SRC=\"$button_url/up.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"^\"></a></B></td>";
  } elsif ($sort eq "status_rev") {
    print "status\">$TXT{'status'} <IMG SRC=\"$button_url/down.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"v\"></a></B></td>";
  } else {
    print "status\">$TXT{'status'}</a></B></td>";
  }
  
  print '<td valign="middle" width="100" bgcolor="#D3DCE3"><B>';
  print "<a href=\"$scripturl?action=displayheaders&folder=$folder&firstmessage=",
  $firstmessage+1,"&sessionid=$thissession&sort=";
  
  if ($sort eq "date") {
    print "date_rev\">$TXT{'date'} <IMG SRC=\"$button_url/up.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"^\"></a></B></td>";
  } elsif ($sort eq "date_rev") {
    print "date\">$TXT{'date'} <IMG SRC=\"$button_url/down.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"v\"></a></B></td>";
  } else {
    print "date\">$TXT{'date'}</a></B></td>";
  }
  
  print '<td valign="middle" width="150" bgcolor="#D3DCE3"><B>';
  print "<a href=\"$scripturl?action=displayheaders&folder=$folder&firstmessage=",
  $firstmessage+1,"&sessionid=$thissession&sort=";
  
  if ($folder ne "SENT") {
    if ($sort eq "sender") {
      print "sender_rev\">$TXT{'sender'} <IMG SRC=\"$button_url/down.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"v\"></a></B></td>";
    } elsif ($sort eq "sender_rev") {
      print "sender\">$TXT{'sender'} <IMG SRC=\"$button_url/up.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"^\"></a></B></td>";
    } else {
      print "sender\">$TXT{'sender'}</a></B></td>";
    }
  } else {
    if ($sort eq "sender") {
      print "recipient_rev\">$TXT{'recipient'} <IMG SRC=\"$button_url/down.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"v\"></a></B></td>";
    } elsif ($sort eq "recipient_rev") {
      print "recipient\">$TXT{'recipient'} <IMG SRC=\"$button_url/up.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"^\"></a></B></td>";
    } else {
      print "sender\">$TXT{'recipient'}</a></B></td>";
    }
  }
  
  print '<td valign="middle" width="350" bgcolor="#D3DCE3"><B>';
  print "<a href=\"$scripturl?action=displayheaders&folder=$folder&firstmessage=",
  $firstmessage+1,"&sessionid=$thissession&sort=";
  
  if ($sort eq "subject") {
    print "subject_rev\">$TXT{'subject'} <IMG SRC=\"$button_url/down.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"v\"></a></B></td>";
  } elsif ($sort eq "subject_rev") {
    print "subject\">$TXT{'subject'} <IMG SRC=\"$button_url/up.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"^\"></a></B></td>";
  } else {
    print "subject\">$TXT{'subject'}</a></B></td>";
  }
  
  print '<td valign="middle" width="50" bgcolor="#D3DCE3"><B>';
  print "<a href=\"$scripturl?action=displayheaders&folder=$folder&firstmessage=",
  $firstmessage+1,"&sessionid=$thissession&sort=";
  
  if ($sort eq "size") {
    print "size_rev\">$TXT{'size'} <IMG SRC=\"$button_url/up.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"^\"></a></B></td>";
  } elsif ($sort eq "size_rev") {
    print "size\">$TXT{'size'} <IMG SRC=\"$button_url/down.gif\" width=\"12\"
                height=\"12\" border=\"0\" alt=\"v\"></a></B></td>";
  } else {
    print "size\">$TXT{'size'}</a></B></td>";
  }
  
  
  
  print '<td bgcolor="#D3DCE3"><img src="' . $button_url. '/chkb.gif"></td></tr><tr>';
  
  foreach my $messnum ($firstmessage .. $lastmessage) {
    last if !(defined($headers[$messnum]));
    ${$headers[$messnum]}{subject} =~ s/</&lt;/g;
    ${$headers[$messnum]}{subject} =~ s/>/&gt;/g;

   if (length(${$headers[$messnum]}{subject}) < 4) {  
       # usefall in case a mail have no subject (or a very small one)
       ${$headers[$messnum]}{subject} .= "&nbsp;&nbsp;&nbsp;<i>[ $TXT{'read'} ]</i>"; 
   }

   if (length(${$headers[$messnum]}{subject}) > 50) {  
       # subject too long
       ${$headers[$messnum]}{subject} = substr(${$headers[$messnum]}{subject},0,50) . "..."; 
   }


   if (${$headers[$messnum]}{foundstatus}) {
       ${$headers[$messnum]}{subject} .= " &nbsp;&nbsp; <b> [" . ${$headers[$messnum]}{foundstatus}  . "]</b> "; 
   }

   if ( $messnum % 2 ) {
     $bgcolor = "#EEEEEE";
   } else {
     $bgcolor = "#DDDDDD";
   }

   $message_size = ${$headers[$messnum]}{messagesize};
   if ($message_size > 1048575){ 
     $message_size = int(($message_size/1048576)+0.5) . "MB"; 
   } elsif ($message_size > 1023) {
     $message_size =  int(($message_size/1024)+0.5) . "kB"; 
   } else {
     $message_size = $message_size;
   }

   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
     localtime( ${$headers[$messnum]}{date} );
   $mon++; $year += 1900;
   
   $min = "0" . $min if ($min<10); $mday = "0" . $mday if ($mday<10);
   $hour = "0" . $hour if ($hour<10); $mon = "0" . $mon if ($mon<10);
   my $datestring = "$mday/$mon/$year ${hour}h$min";
		  
   $status = "";
   unless ( ${$headers[$messnum]}{status} =~ /S/) { # seen (read)
      $status  .= "<img src=\"$button_url/new.gif\">"; 
	 ${$headers[$messnum]}{subject} = "<b>" . ${$headers[$messnum]}{subject} . "</b>";
	 ${$headers[$messnum]}{from} = "<b>" . ${$headers[$messnum]}{from} . "</b>";
 	 ${$headers[$messnum]}{to} = "<b>" . ${$headers[$messnum]}{to} . "</b>";
      $datestring = "<b>" . $datestring . "</b>";
  }

   if ( ${$headers[$messnum]}{status} =~ /R/) { # replied (answered)
       $status  .= " <img src=\"$button_url/read1.gif\"> "; 
   }

   if ( ${$headers[$messnum]}{status} =~ /F/) {  # flag
       $status  .= " <img src=\"$button_url/smiley1.gif\"> "; 
   }
   
   if (${$headers[$messnum]}{new}) {
     $status  .= " <b>*</b> ";
   } 
     
   if ( ${$headers[$messnum]}{content_type} =~ /^multipart/i ) {
     $status .= "<img src=\"$button_url/attach.gif\">";
   }
	
   print "<td valign=\"middle\" width=\"20\" bgcolor=\"$bgcolor\"><b>" . ($messnum + 1). "</b>&nbsp;</td>";
   print "<td valign=\"middle\" width=\"50\" bgcolor=\"$bgcolor\"> $status &nbsp;</td>";

   print "<td valign=\"middle\" width=\"100\" bgcolor=\"$bgcolor\"><nobr>";
   print $datestring,'</nobr>&nbsp;</td>';

   if ($folder ne "SENT") {
     print "<td valign=\"middle\" width=\"150\" bgcolor=\"$bgcolor\">",
     ${$headers[$messnum]}{from},'&nbsp;</td>';
   } else {
     print "<td valign=\"middle\" width=\"150\" bgcolor=\"$bgcolor\">",
     ${$headers[$messnum]}{to},'&nbsp;</td>';
   }

   print "<td valign=\"middle\" width=\"350\" bgcolor=\"$bgcolor\">",
         "<a href=\"$scripturl?action=readmessage&folder=$folder&firstmessage=",
         $firstmessage+1,"&sessionid=$thissession&sort=$sort&headers=",
         "normal&message_id=",${$headers[$messnum]}{message_id},'">',
         ${$headers[$messnum]}{subject},'</a></td>';

    print "<td valign=\"middle\" width=\"40\" bgcolor=\"$bgcolor\">";
    print $message_size . "</td>";

    print "<td align=\"center\" valign=\"middle\" bgcolor=\"$bgcolor\">";
    print checkbox(-name=>'message_ids',
		   -value=>${$headers[$messnum]}{message_id},
                   -label=>'');
    print '</td>';

    print '</tr>';
  }



  # second row : navigatation
  
  print '<tr><td colspan="7" bgcolor="#ffffff">&nbsp;</td></tr>';
  print '<tr><td bgcolor="#dcdcdc" colspan=7><table border="0" width="100%" cellpadding="0" cellspacing="0">';
  print '<td "valign"=middle align=center><font face="',$default_variable_font,'" size=2>';
  
  print '<table border=0 width="100%">';
  print '<tr><td valign="middle" align="right">';
  
  if ($firstmessage != 0) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=0\">";
    print "<img src=\"$button_url/first.gif\" border=\"0\" alt=\"First page\"></a>";
  } else {
    print "<img src=\"$button_url/first-grey.gif\" border=\"0\" alt=\"\">";      
  }
  
  if (($firstmessage - $numberofheaders) >= 0) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=" . ($firstmessage +1 - $numberofheaders) . "\">";
    print "<img src=\"$button_url/left.gif\" border=\"0\" alt=\"Previous page\"></a>";
  } else {
    print "<img src=\"$button_url/left-grey.gif\" border=\"0\" alt=\"\">";      
  }
  
  print '[',$TXT{'page'} ,' ' . $page_nb;
  print ' ', $TXT{'of'} ,' ', $page_total ,']';
  
  if (($firstmessage + $numberofheaders) <= $#headers) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=" . ($firstmessage + 1 + $numberofheaders) ."\">";
    print "<img src=\"$button_url/right.gif\"  border=\"0\" alt=\"Next page\"></a>";
  } else {
    print "<img src=\"$button_url/right-grey.gif\" border=\"0\" alt=\"\">"; 
  }
  
  if (($firstmessage + $numberofheaders) <= $#headers) {
    print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=" . ($#headers+1). "\">";
    print "<img src=\"$button_url/last.gif\" border=\"0\" alt=\"Last page\"></a>";
  } else {
    print "<img src=\"$button_url/last-grey.gif\" border=\"0\" alt=\"\">";      
  }
  
  print '</nobr></td><td valign="middle" align="right">';
  # delete button
  

  print $TXT{'check_all'},': <input name="allbox" type="checkbox" value="1" onClick="CheckAll();">';
#  print '<input type="hidden" name="action" value="">';

  print ' | <input type="radio" name="action" value="deletemessage" CHECKED> ';
  print $TXT{'delete'} if ($folder ne "TRASH");
 print $TXT{'delete_real'} if ($folder eq "TRASH");
  print ' | <input type="radio" name="action" value="savemessage"> ' if ($folder ne "SAVED");
  print $TXT{'save'} if ($folder ne "SAVED");
 print ' | <a href="javascript:document.Formdel.submit();" >',$TXT{'move'},'</a>';
# print ' | <a href="javascript:document.Formdel.action=\'deletemessage\';document.Formdel.submit()" >',$TXT{'delete'},'</a>' if ($folder ne "TRASH");
# print ' | <a href="javascript:document.Formdel.action=\'deletemessage\';document.Formdel.submit()" >',$TXT{'delete_real'},'</a>' if ($folder eq "TRASH");
# print ' | <a href="javascript:document.Formdel.action=\'savemessages\';document.Formdel.submit()" >',$TXT{'save'},'</a>'  if ($folder ne "SAVED");

  print '</td></tr></table>';
  print '</td><td>&nbsp;&nbsp;&nbsp;</td>';
  print '</tr></table></td></tr>';
  




  print end_form();
  print '</table>';


  print '<script language="JavaScript">
<!--

             <!--
             function CheckAll()
             {
                for (var i=0;i<document.Formdel.elements.length;i++)
                {
                   var e = document.Formdel.elements[i];
                   if ((e.name != "allbox") && (e.name != "action"))
                   e.checked = document.Formdel.allbox.checked;
                }
             }
             //-->
</script>';

  printfooter();
} 
############### END DISPLAYHEADERS ##################


#################### READMESSAGE ####################
sub readmessage {
  printheader();
  my $messageid = param("message_id");
  
  my %message = %{&getmessage($user,$messageid)};
  
  if (%message) {
    
    $messageid = $message{messageid};  # message id (eventualy corrected by checkid)
    unless ($message{status} =~ /S/) {
      ($message{status},$message{messageid}) = updatestatus($user,$messageid,"+S");
    }
     
    my ($from, $replyto, $to, $cc, $subject, $body); # will hold web-ified headers
    
    my $headers = param("headers") || 'normal';
    $messageid = $message{messageid};
    $from = $message{from} || '';
    $from =~ s/</&lt;/g;
    $from =~ s/>/&gt;/g;
    $replyto = $message{replyto} || '';
    $replyto =~ s/</&lt;/g;
    $replyto =~ s/>/&gt;/g;
    $to = $message{to} || '';
    $to =~ s/</&lt;/g;
    $to =~ s/>/&gt;/g;
    $cc = $message{cc} || '';
    $cc =~ s/</&lt;/g;
    $cc =~ s/>/&gt;/g;
    $subject = $message{subject} || '';
    $subject =~ s/</&lt;/g;
    $subject =~ s/>/&gt;/g;
    if ( ($message{contenttype} =~ /^text\//i) &&
	 ($message{encoding} =~ /^quoted-printable/i) ) {

      $message{"body"} = decode_qp($message{"body"});
      
	$from = decode_qp($from);
	$to = decode_qp($to);
	$cc = decode_qp($cc);
	$replyto = decode_qp($replyto);
	$subject = decode_qp($subject);
    }

    if ($mimewords_decode) {
	$from = MIME::Words::decode_mimewords($from);
	$to = MIME::Words::decode_mimewords($to);
	$cc = MIME::Words::decode_mimewords($cc);
	$replyto = MIME::Words::decode_mimewords($replyto);
	$subject = MIME::Words::decode_mimewords($subject);
    } 
    
    if ($lang eq 'ja') {
      $from = &Jcode::convert(\$from, 'euc', Jcode::getcode(\$from), "z");
      $to = &Jcode::convert(\$to, 'euc', Jcode::getcode(\$to), "z");
      $cc = &Jcode::convert(\$cc, 'euc', Jcode::getcode(\$cc), "z");
      $replyto = &Jcode::convert(\$replyto, 'euc', Jcode::getcode(\$replyto), "z");
      $subject = &Jcode::convert(\$subject, 'euc', Jcode::getcode(\$subject), "z");
    }

    # transform mail addresses to links -> addressbook
    
    $body = $message{"body"} || '';
    $body = Jcode->new($body)->euc if($lang eq 'ja');
    $body =~ s/</&lt;/g;
    $body =~ s/>/&gt;/g;
    $body =~ s/\n/<BR>\n/g;
    
    if (($message{contenttype} =~ /^multipart/i) &&
	(${$message{attachment}[0]}{contenttype} =~ /^text\/plain/i)) {
      if (${$message{attachment}[0]}{encoding} =~ /^quoted-printable/i) {
        ${$message{attachment}[0]}{contents} =
	  decode_qp(${$message{attachment}[0]}{contents});
      }
      $message{"body"} .= "\n" . ${$message{attachment}[0]}{contents};
    }


    my $base_url = "$scripturl?sessionid=$thissession&folder=$folder&firstmessage=" . ($firstmessage+1) . 
                   "&sort=$sort&message_id=$messageid&headers=$headers";

    my $base_url2 = "$scripturl?sessionid=$thissession&folder=$folder&firstmessage=" . ($firstmessage+1) .
                    "&sort=$sort";

    my $messageafterdelete;
    if (defined($message{"next"})) {
      $messageafterdelete = $message{"next"};
    } elsif (defined($message{"prev"})) {
      $messageafterdelete = $message{"prev"};
    }
    
    
    print '<table border="0" align="center" width="85%" cellpadding="1" cellspacing="1">';
    
    # first row - mbox name / message nb
    
    print '<tr><td bgcolor="#002266" align="center"><table border="0" width="100%" cellpadding="0" cellspacing="0">',
    '<tr><td align="left" width="33%"><font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',uc($TXT{$folder}),
    ' - ',$TXT{'maildisplay_title'},'</b></font>',
    '</td><td align=right width="33%"><font color="#ffffff" face="',$default_variable_font,'" size="3">',
    "<b>$TXT{'message_nr'} " . $message{"number"} ."</b></font></td></tr></table></td></tr>";
    
    # second row : navigatation
    
    print '<tr><td bgcolor="#dcdcdc"><table border="0" width="100%" cellpadding="0" cellspacing="0">';
    print '<td align=center><font face="',$default_variable_font,'" size=2>';
    
    print start_form();
    
    print '<table border=0 width="100%">';
    print '<tr><td valign="middle" align="left">';
    print "<a href=\"$base_url&action=displayheaders\">",$TXT{$folder},"</a> | ";

    unless ($folder eq 'SAVED') {
      print "<a href=\"$base_url2&action=savemessage&message_ids=$messageid";
      if ($messageafterdelete) {
	print "&messageafterdelete=1&message_id=$messageafterdelete";
      }
      print '">',$TXT{'save'},'</a> | ';
    }
    print "<a href=\"$base_url&action=composemessage\">$TXT{'compose'}</a> | ";
    print "<a href=\"$base_url&action=composemessage&composetype=reply\">$TXT{'reply'}</a> | ";
    print "<a href=\"$base_url&action=composemessage&composetype=replyall\">$TXT{'replyall'}</a> | ";
    print "<a href=\"$base_url&action=composemessage&composetype=forward\">$TXT{'forward'}</a> | ";

    if ($folder ne "TRASH") {
	 print "<a href=\"$base_url2&action=deletemessage&message_ids=$messageid";
	 if ($messageafterdelete) {
	   print "&messageafterdelete=1&message_id=$messageafterdelete";
	 }
	 print "\">$TXT{'delete'}</a> | ";
    } 

    if ($folder eq "TRASH") {
	 print "<a href=\"$base_url2&action=deletemessage&message_ids=$messageid";
	 
	 if ($messageafterdelete) {
	   print "&messageafterdelete=1&message_id=$messageafterdelete";
	 }
	 print "\">$TXT{'delete_real'}</a> | ";
    }

    print "<a href=\"$base_url&action=logout\">$TXT{'logout'}</a>";

    print '</td><td valign="middle" align="center">';

    print '<table border="0"><tr><td>';
    if (defined($message{"prev"})) {
      print "<a href=\"$base_url2&action=readmessage&message_id=" . $message{"prev"}. "&headers=$headers\">";
      print "<img src=\"$button_url/left.gif\" border=\"0\" alt=\"Previous mail\"></a>";
    } else {
      print "<img src=\"$button_url/left-grey.gif\" border=\"0\" alt=\"\">";      
    }
    
    print "</td><td>" . $message{"number"} . "</td><td>";
    
    if (defined($message{"next"})) {
      print "<a href=\"$base_url2&action=readmessage&message_id=" . $message{"next"}. "&headers=$headers\">";
      print "<img src=\"$button_url/right.gif\"  border=\"0\" alt=\"Next mail\"></a>";
    } else {
      print "<img src=\"$button_url/right-grey.gif\" border=\"0\" alt=\"\">"; 
    }
    
    print '</td></tr></table>';
    print '</td><td valign="middle" align="right">';
  
    my $flagtxt = $TXT{'flag'};
    if ($message{status} =~ /F/) {
      $flagtxt = $TXT{'unflag'};
    }
    
    print "<a href=\"$base_url&action=toggle&flag=F&after=msg\">$flagtxt</a> | ";

    if ($headers eq "all") {
      print "<a href=\"$base_url2&action=readmessage&message_id=$messageid&headers=normal\">$TXT{'normal_headers'}</a>";
    } else {
      print "<a href=\"$base_url2&action=readmessage&message_id=$messageid&headers=all\">$TXT{'all_headers'}</a>";
    }


    if ( $message{status} =~ /R/) { # replied (answered)
      print " | <img align=\"absmiddle\" src=\"$button_url/read1.gif\"> "; 
    }


    print '</td></tr></table>';
    print '</td><td>&nbsp;</td>';
    print '</tr></table></td></tr>';

    print '<tr><td>&nbsp;</td></tr>';
     
    print '<tr><td width="420" valign="middle" bgcolor="#DDDDDD"><font face="',$selected_mail_font,'">';
    
    if ($headers eq "all") {
      $message{"header"} =~ s/</&lt;/g;
      $message{"header"} =~ s/>/&gt;/g;
      $message{"header"} =~ s/\n/<BR>\n/g;
      $message{"header"} =~ s/\n([-\w]+?:)/\n<B>$1<\/B>/g;
      print $message{"header"};
    } else {
      print "<B>$TXT{'date'}:</B> $message{date}<BR>\n";
      print "<B>$TXT{'from'}:</B> $from<BR>\n";
      if ($replyto) {
	print "<B>$TXT{'replyto'}:</B> $replyto<BR>\n";
      }
      if ($to) {
	print "<B>$TXT{'to'}: </B> $to<BR>\n";
      }
      if ($cc) {
	print "<B>$TXT{'cc'}: </B> $cc<BR>\n";
      }
      if ($subject) {
	print "<B>$TXT{'subject'}:</B> <b>$subject</b>\n";
      }
    }
    
    print '</font></td></tr><tr><td bgcolor="#EEEEEE"><font face="',$selected_mail_font,'">';
    foreach (qw(Http Ftp http ftp nntp news gopher telnet)) {
      $body =~ s/($_:[\d|\w|\/|\.|\-|?|;|&|+|=|~]*[\d|\w|\/])([\b|\n| ]*)/<A HREF=\"$1\">$1<\/A>$2/gs;
    }
    $body =~ s/([\b|\n| ]+)(www\.[-\w\.]+\.[-\w]{2,3})([\b|\n| ]*)/$1<a href=\"http:\/\/$2\">$2<\/a>$3/gs;
    $body =~ s/([\b|\n| ]+)(ftp\.[-\w\.]+\.[-\w]{2,3})([\b|\n| ]*)/$1<a href=\"ftp:\/\/$2\">$2<\/a>$3/gs;


    print $body;

    # Handle the messages generated if sendmail is set up to send MIME error reports
      if ($message{contenttype} =~ /^multipart\/report/i) {
         foreach my $attnumber (0 .. $#{$message{attachment}}) {
            if (defined(${$message{attachment}[$attnumber]}{contents})) {
               ${$message{attachment}[$attnumber]}{contents} =~ s/&/&amp;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/\"/&quot;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/</&lt;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/>/&gt;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/\n/<BR>\n/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/ {2}/ &nbsp;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/\t/ &nbsp;&nbsp;&nbsp;&nbsp;/g;
               print ${$message{attachment}[$attnumber]}{contents} .
                     hr();
            }
         }
      } elsif ( ($message{contenttype} ne 'N/A') && !($message{contenttype} =~ /^text/i )) {
         foreach my $attnumber (0 .. $#{$message{attachment}}) {
            next unless (defined(%{$message{attachment}[$attnumber]}));
            if (($attnumber == 0) &&
               (${$message{attachment}[$attnumber]}{contenttype} =~ /^text\/plain/i)) {
               if (${$message{attachment}[$attnumber]}{encoding} =~ /^quoted-printable/i) {
                  ${$message{attachment}[$attnumber]}{contents} = decode_qp(${$message{attachment}[$attnumber]}{contents});
               } elsif (${$message{attachment}[$attnumber]}{encoding} =~ /^base64/i) {
                  ${$message{attachment}[$attnumber]}{contents} = decode_base64(${$message{attachment}[$attnumber]}{contents});
               }
               print hr();
               ${$message{attachment}[$attnumber]}{contents} =~ s/&/&amp;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/\"/&quot;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/</&lt;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/>/&gt;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/\n/<BR>\n/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/ {2}/ &nbsp;/g;
               ${$message{attachment}[$attnumber]}{contents} =~ s/\t/ &nbsp;&nbsp;&nbsp;&nbsp;/g;
               foreach (qw(Http Ftp http https ftp nntp news gopher telnet)) {
                  ${$message{attachment}[$attnumber]}{contents} =~
                  s/($_:\/\/[\w\.\-]+?\/?[^\s<>]*[\w\/])([\b|\n| ]*)/<A HREF=\"$1\" TARGET=\"_blank\">$1<\/A>$2/gs;
               }
               ${$message{attachment}[$attnumber]}{contents} =~ s/([\b|\n| ]+)(www\.[-\w\.]+\.[-\w]{2,3})([\b|\n| ]*)/$1<a href=\"http:\/\/$2\" TARGET=\"_blank\">$2<\/a>$3/gs;
               print ${$message{attachment}[$attnumber]}{contents} . "<BR><BR>";
            } else {
               my $escapedfilename = CGI::escape(${$message{attachment}[$attnumber]}{filename});
               if (${$message{attachment}[$attnumber]}{filename} =~
                  /\.(jpg|jpeg|gif|png)$/i) {
                  print "<table border=\"0\" align=\"center\" cellpadding=\"2\">";
                  print "<tr><td valign=\"middle\"";
                  print "bgcolor=\"#DDDDDD\" align=\"center\">$TXT{'attachment'} $attnumber: ${$message{attachment}[$attnumber]}{filename}";
                  print "</td></tr><td valign=\"middle\" bgcolor=\"#E5E5E5\" align=\"center\">";
                  print "<IMG BORDER=\"0\" SRC=\"$scripturl?action=viewattachment&amp;sessionid=$thissession&amp;message_id=$messageid&amp;attachment_number=$attnumber\">";

                  print '</td></tr><tr><td valign="middle" bgcolor="#DDDDDD" align="center">';
                  print "<a href=\"$scripturl?action=viewattachment&amp;sessionid=$thissession&amp;message_id=$messageid&amp;attachment_number=$attnumber\" target=\"_new\">$TXT{'download'}</a>";
                  print '</td></tr></table><br>';

               } elsif ((${$message{attachment}[$attnumber]}{contenttype} eq "text/html") && ($display_html_attach eq "yes")) {

                  print "<table border=\"0\" align=\"center\" cellpadding=\"2\"><tr><td valign=\"middle\" bgcolor=\"#DDDDDD\" align=\"center\">";
		  print 'HTML ' . $TXT{'attachment'} . '</td></tr><tr><td valign="middle" bgcolor="#E5E5E5" align="left">';

	               if (${$message{attachment}[$attnumber]}{encoding} =~ /^quoted-printable/i) {
	                  ${$message{attachment}[$attnumber]}{contents} = decode_qp(${$message{attachment}[$attnumber]}{contents});
	               } elsif (${$message{attachment}[$attnumber]}{encoding} =~ /^base64/i) {
	                  ${$message{attachment}[$attnumber]}{contents} = decode_base64(${$message{attachment}[$attnumber]}{contents});
	               }

		  # remove javascript tags (better than nothing...)
  		  ${$message{attachment}[$attnumber]}{contents} =~ s/<script>.*?<\/script>//igs;
		  ${$message{attachment}[$attnumber]}{contents} =~ s/script/sc\<nobr\>ri\<\/nobr\>pt/igs;

		  print  ${$message{attachment}[$attnumber]}{contents};

                  print '</td></tr><tr><td valign="middle" bgcolor="#DDDDDD" align="center">';
                  print "<a href=\"$scripturl?action=viewattachment&amp;sessionid=$thissession&amp;message_id=$messageid&amp;attachment_number=$attnumber\" target=\"_new\">$TXT{'download'}</a>";
                  print '</td></tr></table><br>';

               } else {
                  print "<table border=\"0\" align=\"center\" cellpadding=\"2\"><tr><td colspan=\"2\" valign=\"middle\" bgcolor=\"#DDDDDD\" align=\"center\">$TXT{'attachment'} $attnumber</td></tr><td valign=\"middle\" bgcolor=\"#E5E5E5\" align=\"center\">";
                  print "$TXT{'type'}: " . ${$message{attachment}[$attnumber]}{contenttype} . "<BR>";
                  print "$TXT{'filename'}: " . ${$message{attachment}[$attnumber]}{filename} . "<BR>";
                  print "$TXT{'encoding'}: " . ${$message{attachment}[$attnumber]}{encoding};
                  print '</td><td valign="middle" bgcolor="#E5E5E5" align="center">';
                  print "<a href=\"$scripturl?action=viewattachment&amp;sessionid=$thissession&amp;message_id=$messageid&amp;attachment_number=$attnumber\" target=\"_new\">$TXT{'download'}</a>";
                  print '</font></td></tr></table><br>';
               }
            }
         }
      }


# copy of menubar  (cut'n'paste)

    print '</td></tr><tr><td>&nbsp;</td></tr>';

    # second row : navigatation
    
    print '<tr><td bgcolor="#dcdcdc"><table border="0" width="100%" cellpadding="0" cellspacing="0">';
    print '<td align=center><font face="',$default_variable_font,'" size=2>';
    
    print start_form();
    



   print '<table border=0 width="100%">';
    print '<tr><td valign="middle" align="left">';
    print "<a href=\"$base_url&action=displayheaders\">",$TXT{$folder},"</a> | ";

    unless ($folder eq 'SAVED') {
      print "<a href=\"$base_url2&action=savemessage&message_ids=$messageid";
      if ($messageafterdelete) {
	print "&messageafterdelete=1&message_id=$messageafterdelete";
      }
      print '">',$TXT{'save'},'</a> | ';
    }
    print "<a href=\"$base_url&action=composemessage\">$TXT{'compose'}</a> | ";
    print "<a href=\"$base_url&action=composemessage&composetype=reply\">$TXT{'reply'}</a> | ";
    print "<a href=\"$base_url&action=composemessage&composetype=replyall\">$TXT{'replyall'}</a> | ";
    print "<a href=\"$base_url&action=composemessage&composetype=forward\">$TXT{'forward'}</a> | ";


    if ($folder ne "TRASH") {
	 print "<a href=\"$base_url2&action=deletemessage&message_ids=$messageid";
	 if ($messageafterdelete) {
	   print "&messageafterdelete=1&message_id=$messageafterdelete";
	 }
	 print "\">$TXT{'delete'}</a> | ";
    } 

    if ($folder eq "TRASH") {
	 print "<a href=\"$base_url2&action=deletemessage&message_ids=$messageid";
	 
	 if ($messageafterdelete) {
	   print "&messageafterdelete=1&message_id=$messageafterdelete";
	 }
	 print "\">$TXT{'delete_real'}</a> | ";
    }

    print "<a href=\"$base_url&action=logout\">$TXT{'logout'}</a>";

    print '</td><td valign="middle" align="center">';
    
    print '<table border="0"><tr><td>';
    if (defined($message{"prev"})) {
      print "<a href=\"$base_url2&action=readmessage&message_id=" . $message{"prev"}. "&headers=$headers\">";
      print "<img src=\"$button_url/left.gif\" border=\"0\" alt=\"Previous mail\"></a>";
    } else {
      print "<img src=\"$button_url/left-grey.gif\" border=\"0\" alt=\"\">";
    }
          
    print "</td><td>" . $message{"number"} . "</td><td>";
     
    if (defined($message{"next"})) {
      print "<a href=\"$base_url2&action=readmessage&message_id=" . $message{"next"}. "&headers=$headers\">";
      print "<img src=\"$button_url/right.gif\"  border=\"0\" alt=\"Next mail\"></a>";
    } else {
      print "<img src=\"$button_url/right-grey.gif\" border=\"0\" alt=\"\">";
    }
          
    print '</td></tr></table>';
    
    print '</td><td valign="middle" align="right">';
  
    my $flagtxt = $TXT{'flag'};
    if ($message{status} =~ /F/) {
      $flagtxt = $TXT{'unflag'};
    }
    
    print "<a href=\"$base_url&action=toggle&flag=F&after=msg\">$flagtxt</a> | ";

    if ($headers eq "all") {
      print "<a href=\"$base_url2&action=readmessage&message_id=$messageid&headers=normal\">$TXT{'normal_headers'}</a>";
    } else {
      print "<a href=\"$base_url2&action=readmessage&message_id=$messageid&headers=all\">$TXT{'all_headers'}</a>";
    }

    if ( $message{status} =~ /R/) { # replied (answered)
      print " | <img align=\"absmiddle\" src=\"$button_url/read1.gif\"> "; 
    }

    print '</td></tr></table>';
    print '</td><td>&nbsp;</td>';
    print '</tr></table></td></tr>';

# end of table
    
    print '</table>';
    
  } else {

    my $errormsg = "$TXT{'err_message_gone'} : ' $messageid ' !\n";
    $errormsg .= start_form(), hidden(-name=>'action',
				      -default=>'displayheaders',
				      -override=>'1'),
      hidden(-name=>'firstmessage', 
	     -default=>$firstmessage +1,
	     -override=>'1'),
      hidden(-name=>'sort',
	     -default=>$sort,
	     -override=>'1'),
      hidden(-name=>'sessionid',
	     -default=>$thissession,
	     -override=>'1');

      $errormsg .= submit($TXT{'inbox'});

    $errormsg .= end_form();
    omailerror($errormsg);
  }

  printfooter();
}
############### END READMESSAGE ##################


############### COMPOSEMESSAGE ###################
sub composemessage {
  no strict 'refs';
  my $messageid = param("message_id");
  my %message;
  my $from = $thissession;
  $from =~ s/\-session\-0.*$/\@$domainname/; # create from: address
  my %sigs = %{&readsigs};
  my @attlist;
  my $attnumber;
  my $from2 = "";
  my $from3 = "";
  my $from4 = "";
  my $from5 = "";
  my $realname = "";
  my $realname2 = "";
  my $realname3 = "";
  my $realname4 = "";
  my $realname5 = "";

  # multiattachsupport

     if (defined(param($TXT{'add'}))) {
  printheader();
      @attlist = @{&getattlist()};
      my $attachment = param("attachment");
      my $attname = $attachment;
      my $attcontents = '';
      if ($attachment) {
         if ( ($attlimit) && ( ( $savedattsize + (-s $attachment) ) > ($attlimit * 1048576) ) ) {
            omailerror("att overflow!");
         }
         my $content_type;
### Convert :: back to the ' like it should be.
         $attname =~ s/::/'/g;
### Trim the path info from the filename
         $attname =~ s/^.*\\//;
         $attname =~ s/^.*\///;
         $attname =~ s/^.*://;
         $/ = undef; # Force a single input to grab the whole spool
         $attcontents = <$attachment>;
         $/ = "\n";
         $attcontents = encode_base64($attcontents);
         $savedattsize += length($attcontents);
         if (defined(uploadInfo($attachment))) {
            $content_type = ${uploadInfo($attachment)}{'Content-Type'} || 'application/octet-stream';
         } else {
            $content_type = 'application/octet-stream';
         }
         my $attnum = ($#attlist +1) || 0;
         open (ATTFILE, ">$maildir/omailtmp/$thissession-att$attnum");
         print ATTFILE "Content-Type: ", $content_type,";\n";
         print ATTFILE "\tname=\"$attname\"\nContent-Transfer-Encoding: base64\n\n";
         print ATTFILE "$attcontents";
         close ATTFILE;
         $attname =~ s/&/&amp;/g;
         $attname =~ s/\"/&quot;/g;
         $attname =~ s/</&lt;/g;
         $attname =~ s/>/&gt;/g;
         $attname =~ s/^(.*)$/<em>$1<\/em>/;
         push (@attlist, $attname);
      }
   } elsif ( !(defined(param($TXT{'add'})))) {
      deleteattachments();
   }


  # 3 from's should be enough. if not, this should be programmed with an array
  
  if ($prefs{"from"}) {
    $from = $prefs{"from"};
  }
  if ($prefs{"from2"}) {
    $from2 = $prefs{"from2"};
  }
  if ($prefs{"from3"}) {
    $from3 = $prefs{"from3"};
  }
  if ($prefs{"from4"}) {
    $from4 = $prefs{"from4"};
  }
  if ($prefs{"from5"}) {
    $from5 = $prefs{"from5"};
  }
  
  if ($prefs{"realname"}) {
    $realname = $prefs{"realname"};
    $from =~ s/^(.+)$/$realname <$1>/;
  }
  if ($prefs{"realname2"}) {
    $realname2 = $prefs{"realname2"};
    $from2 =~ s/^(.+)$/$realname2 <$1>/;
  }
  if ($prefs{"realname3"}) {
    $realname3 = $prefs{"realname3"};
    $from3 =~ s/^(.+)$/$realname3 <$1>/;
  }
  if ($prefs{"realname4"}) {
    $realname4 = $prefs{"realname4"};
    $from4 =~ s/^(.+)$/$realname4 <$1>/;
  }
  if ($prefs{"realname5"}) {
    $realname5 = $prefs{"realname5"};
    $from5 =~ s/^(.+)$/$realname5 <$1>/;
  }

  my $escapedfrom = $from;
  $escapedfrom =~ s/</&lt;/g;
  $escapedfrom =~ s/>/&gt;/g;
  my $escapedfrom2 = $from2;
  $escapedfrom2 =~ s/</&lt;/g;
  $escapedfrom2 =~ s/>/&gt;/g;
  my $escapedfrom3 = $from3;
  $escapedfrom3 =~ s/</&lt;/g;
  $escapedfrom3 =~ s/>/&gt;/g;
  my $escapedfrom4 = $from4;
  $escapedfrom4 =~ s/</&lt;/g;
  $escapedfrom4 =~ s/>/&gt;/g;
  my $escapedfrom5 = $from5;
  $escapedfrom5 =~ s/</&lt;/g;
  $escapedfrom5 =~ s/>/&gt;/g;

  my $to = '';
  my $cc = '';
  my $bcc = '';
  my $subject = '';
  my $body = '';
  my $composetype = param("composetype") || '';
  my $references = '';
  my $inreplyto = '';
  
  if ($composetype) {
    $to = param("to") || '';
    $cc = param("cc") || '';
    $bcc = param("bcc") || '';
    $subject = param("subject") || '';
    $body = param("body") || '';
    
    if (($composetype eq "reply") || ($composetype eq "replyall") || ($composetype eq "forward") ) {
      %message = %{&getmessage($user,$messageid)};
      
      ### Handle mail programs that send the body of a message quoted-printable
      if ( ($message{contenttype} =~ /^text\//i) && ($message{encoding} =~ /^quoted-printable/i) ) {
	$message{"body"} = decode_qp($message{"body"});
      }
      $body = $message{"body"} || '';
      ### If the first attachment is text, assume it's the body of a message
      ### in multi-part format
      if (($message{contenttype} =~ /^multipart/i) &&
	  (defined(${$message{attachment}[0]}{contenttype})) && 
	    (${$message{attachment}[0]}{contenttype} =~ /^text\/plain/i)) {
        if (${$message{attachment}[0]}{encoding} =~ /^quoted-printable/i) {
	   ${$message{attachment}[0]}{contents} = decode_qp(${$message{attachment}[0]}{contents});
        } 
        $body .= "\n" . ${$message{attachment}[0]}{contents};
        shift @{$message{attachment}};
      }
      # Handle the messages generated if sendmail is set up to send MIME error reports
      if ($message{contenttype} =~ /^multipart\/report/i) {
        foreach my $attnumber (0 .. $#{$message{attachment}}) {
        if (defined(${$message{attachment}[$attnumber]}{contents})) {
          $body .= ${$message{attachment}[$attnumber]}{contents};
          shift @{$message{attachment}};
        }
      }
    } # endif composetype = reply|replyall|forward


    if (($composetype eq "reply") || ($composetype eq "replyall")) {
      $subject = $message{"subject"} || '';
      $subject = "Re: " . $message{"subject"} unless ($message{"subject"} =~ /^re:/i);
      $message{references} = '' if($message{references} eq 'N/A');
      $references = " $message{references} <$message{messageid_header}>";
      $inreplyto = " <$message{messageid_header}>";
      if (defined($message{"replyto"})) {
	$to = $message{"replyto"} || '';
      } else {
	$to = $message{"from"} || '';
      }
      if ($composetype eq "replyall") {
	$to .= "," . $message{"to"} if (defined($message{"to"}));
	$to .= "," . $message{"cc"} if (defined($message{"cc"}));
      }
      
      $quotestyle = $prefs{"quotestyle"} if $prefs{"quotestyle"};
      
      if ($quotestyle eq "outlook") {
	$body = "\n\n------------- Original message follows -------------\n\n$body";
      } else {
	$body = "\n" . $body;
	$body =~ s/\n/\n\> /g;
	$body = "\n\nOn $message{'date'}, $message{'from'} wrote :\n" . $body;
      }
    }  

    if ($composetype eq "forward") {

         if (defined(${$message{attachment}[0]}{header})) {
            foreach my $attnumber (0 .. $#{$message{attachment}}) {
               ($attnumber =~ /^(\d+)$/) && ($attnumber = $1);  # untaint it
               open (ATTFILE, ">$maildir/omailtmp/$thissession-att$attnumber");
               print ATTFILE ${$message{attachment}[$attnumber]}{header}, "\n\n", ${$message{attachment}[$attnumber]}{contents};
               close ATTFILE;
            }
            @attlist = @{&getattlist()};
         }

 
      $subject = $message{"subject"} || '';
      $subject = "Fwd: "  . $subject unless ($subject =~ /^fw:/i);
      $body = "\n\n------------- Forwarded message follows -------------\n\n$body";
    }
  } # if ($composetype)

  # add signature to body

} # TODO ?????????????????????????????????  for which if is this for ? (well, without it, systax error...)


  if ((defined($sigs{"signature"})) && ($composetype ne 'continue') )  {
    if ($sigs{"signature"} && !($sigs{"signature2"} || $sigs{"signature3"} || $sigs{"signature4"} || $sigs{"signature5"})) {
     # add signature1 automaticaly only if there are no sig2 and sig3
        $sigs{"signature"} =~ s/\\'/'/g;
        $sigs{"signature"} =~ s/\\n/\n/g; 
         $body .= "\n\n";
         unless ($sigs{'signature'} =~ /--\s/) {        # if signature is not "correct", add the official "-- \n"
           $body .= "-- \n";          # TODO: correct regexp: shoudl be /\n--\s\n/ but it doesn't work
         }
         $body .= $sigs{'signature'};
    }
  }


  printheader();
  print start_multipart_form(-name=>'composeform');

  print '<table border="0" align="center" cellpadding="2">';

  if ($messageid) {
    print '<tr><td bgcolor="#002266" align="left">',
    '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',$TXT{'compose_title'},'</b></font>',
    "</td><td bgcolor=\"#002266\" align=\"right\"> <a href=\"$scripturl?action=readmessage&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage&message_id=$messageid&folder=$folder\"><font color=\"#ffffff\">[ ",$TXT{'back_to_mailbox'}," ]</font></a> ";
    print '</td></tr>';
  } else {
    print '<tr><td bgcolor="#002266" align="left">',
    '<font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',$TXT{'compose_title'},'</b></font>',
    "</td><td bgcolor=\"#002266\" align=\"right\"> <a href=\"$scripturl?action=displayheaders&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage&folder=$folder\"><font color=\"#ffffff\">[ ",$TXT{'back_to_mailbox'}," ]</font></a> ";
    print '</td></tr>';
  }



  print '<tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="75">';

  print '<B>', $TXT{'from'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">';

if (!($escapedfrom2 || $escapedfrom3 || $escapedfrom4 || $escapedfrom5)) {
  print $escapedfrom;
  print '<input type="hidden" name="multifrom" value="0">';
} else {
    print '<select name="multifrom">';
  print '<option value="0">',$escapedfrom,'</option>';
  print '<option value="2">',$escapedfrom2,'</option>' if ($escapedfrom2);
  print '<option value="3">',$escapedfrom3,'</option>' if ($escapedfrom3);
  print '<option value="4">',$escapedfrom4,'</option>' if ($escapedfrom4);
  print '<option value="5">',$escapedfrom5,'</option>' if ($escapedfrom5);
  print '</select>';
}

  print '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="75">';

  print hidden(-name=>'action',
	       -default=>'sendmessage',
	       -override=>'1');
  print hidden(-name=>'sessionid',
	       -default=>$thissession,
	       -override=>'1');
  print hidden(-name=>'sort',
	       -default=>$sort,
	       -override=>'1');
  print hidden(-name=>'composetype',
	       -default=>'continue',
	       -override=>'1');
  print hidden(-name=>'message_id',
	       -default=>$messageid,
	       -override=>'1');
  print hidden(-name=>'folder',
	       -default=>$folder,
	       -override=>'1');
  print hidden(-name=>'references',
               -default=>$references,
               -override=>'1');
  print hidden(-name=>'inreplyto',
               -default=>$inreplyto,
	       -override=>'1');

  print '<a href="Javascript:GoAddressWindow(',"'to'",')"><B>',$TXT{'to'},':</B></a></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'to',
                    -default=>$to,
                    -size=>'70',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="75">',
         '<a href="Javascript:GoAddressWindow(',"'cc'", ')"><B>',$TXT{'cc'},':</B></a></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'cc',
                    -default=>$cc,
                    -size=>'70',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="75">',
         '<a href="Javascript:GoAddressWindow(',"'bcc'", ')"><B>',$TXT{'bcc'},':</B></a></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'bcc',
                    -default=>$bcc,
                    -size=>'70',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="75">',
         '<B>',$TXT{'replyto'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'replyto',
                    -default=>$prefs{"replyto"} || '',
                    -size=>'70',
                    -override=>'1'),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="75">',
         '<B>',$TXT{'attachment'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">';

   foreach my $filename (@attlist) {
      print "$filename<BR>";
   }
   if ( $savedattsize ) {
      print "<em>" . int($savedattsize/1024) . "KB";
      if ( $attlimit ) {
         print " $TXT{'of'} $attlimit MB";
      }
      print "</em><BR>";
   }

  
  print filefield(-name=>'attachment',
                    -default=>'',
                    -size=>'60',
                    -override=>'1',
                    -tabindex=>'-1'),
         submit(-name=>$TXT{'add'},
                       -value=>$TXT{'add'},
                       -tabindex=>'-1'
                      ),
         '</td></tr><tr><td valign="middle" align="right" bgcolor="#DDDDDD" width="75">',
         '<B>',$TXT{'subject'},':</B></td><td valign="middle" align="left" bgcolor="#DDDDDD">',
          textfield(-name=>'subject',
                    -default=>$subject,
                    -size=>'70',
                    -override=>'1');
  print '</td></tr><tr><td colspan="2" bgcolor="#EEEEEE">';

  print '<table border="0" width="100%"><tr><td align="left">';

  foreach my $sigfile (@sigfiles) {

    if ($sigs{$sigfile} && $sigfile =~ /header/) {
	 print button(-onClick=>'composeform.body.value = \'' . $sigs{$sigfile} . '\n\' + composeform.body.value',
	   	      -name=>$prefs{"name$sigfile"}
                     ), '&nbsp;';
      }
  }

  print '</td></tr></table>';

  print textarea(-name=>'body',
                   -default=>$body,
                   -rows=>'20',
                   -columns=>'76',
                   -wrap=>'hard',
                   -override=>'1');

  print '<table border="0" width="100%"><tr><td align="left">';


  foreach my $sigfile (@sigfiles) {

    if ($sigs{$sigfile} && $sigfile =~ /signature/ && !($sigfile eq "signature" && !($sigs{"signature1"} ||$sigs{"signature2"} || $sigs{"signature3"} || $sigs{"signature4"} || $sigs{"signature5"}))) {
        print button(-onClick=>'composeform.body.value = composeform.body.value + \'\n' . $sigs{$sigfile} . '\'',
       	             -name=>$prefs{"name$sigfile"}
                    ), '&nbsp;';
      }
  }

  print '</td></tr></table>';


  print '</td></tr><tr><td align="left" valign="top" bgcolor="#DDDDDD">',
                   submit($TXT{'send'});

  print '</td>';
  print '<td align="right" valign="middle" bgcolor="#DDDDDD">',reset($TXT{'delete_addr'}),'&nbsp;</td></tr></table>';

  print end_form();
  print '<script language="JavaScript">
   <!--
   if (self.focus != null)
      self.focus();

  function OnLoadHandler()
  {
           document.composeform.to.focus();
  }

  function GoAddressWindow(toccbcc)
  {
           var url = "',$prefsurl,'?action=addressbook&sessionid=',
           $thissession,'&field=" + toccbcc;
          if (toccbcc == "to")
            url += "&preexisting=" + escape (document.composeform.to.value);
          else if (toccbcc == "cc")
            url += "&preexisting=" + escape (document.composeform.cc.value);
          else if (toccbcc == "bcc")
            url += "&preexisting=" + escape (document.composeform.bcc.value);        

          var hWnd = window.open(url,"HelpWindow","width=300,height=360,resizable=yes,scrollbars=yes");
          if ((document.window != null) && (!hWnd.opener))
                   hWnd.opener = document.window;
  }

  //-->
  </script>';
  printfooter();
}
############# END COMPOSEMESSAGE #################


############### SENDMESSAGE ######################
sub sendmessage {
  no strict 'refs';

   if (defined(param($TXT{'add'}))) {
      composemessage();
   } else {

  my $messagecontents = "";
  my $date = localtime();
  my $from = $thissession;
  $from =~ s/\-session\-0.*$/\@$domainname/; # create from: address

  my $from2 = "";
  my $from3 = "";

  my $header_from = '';
  my $header_subject = '';  

  if ($prefs{"from"}) {
    $from = $prefs{"from"} || '';
  }
  my $realname = $prefs{"realname"} || '';

  if (param("multifrom")) {
    if (param("multifrom") ne "0") {
      $from = $prefs{("from" . param("multifrom"))};
      $realname = $prefs{("realname" . param("multifrom"))};
    }
  }
      
  $from =~ s/[\||\'|\"|\`]/ /g;  # Get rid of shell escape attempts
  $realname =~ s/[\||\'|\"|\`]/ /g;  # Get rid of shell escape attempts
  ($realname =~ /^(.+)$/) && ($realname = '"'.$1.'"');
  ($from =~ /^(.+)$/) && ($from = $1);

  my $savedatts = ''; # Will buffer saved attachments
  my $boundary = "----=OMAIL_ATT_" . rand();
  
  # for automatic bcc to another mail account (prefs : send_cc_to)
  my $bcc2 = "";
  if ($prefs{"send_cc_to"}) {
    $bcc2 = $prefs{"send_cc_to"};
  }

  my $to = param("to");
  my $cc = param("cc");
  my $bcc = param("bcc");
  if ($bcc2 && $bcc) { $bcc .= ",$bcc2"; }
  if ($bcc2 && !$bcc) { $bcc = $bcc2; }

  my $replyto = param("replyto");
  my $subject = param("subject");
  my $body = param("body") || '';
  my $messageid = param("message_id");
  my $references = param("references");
  my $inreplyto = param("inreplyto");
  $body =~ s/\r//g;  # strip ^M characters from message. How annoying!
  my $attachment = param("attachment");
   if ( $attachment ) {
       getattlist();
       if ( ($attlimit) && ( ( $savedattsize + (-s $attachment) ) > ($attlimit * 1048576) ) ) {
     omailerror ("att_overlimit");     
    }
  }

  ### Trim the path info from the filename
  my $attname = $attachment;
  $attname =~ s/^.*\\//;
  $attname =~ s/^.*\///;
  $attname =~ s/^.*://;
  
  $header_from = "From: $realname <$from>"; 
  $header_subject = "Subject: $subject";

  if ($lang eq 'ja') {
    $to = Jcode->new($to)->mime_encode;
    $header_from = Jcode->new($header_from)->mime_encode;
    $header_subject = Jcode->new($header_subject)->mime_encode;
    $body = &convert2jis($body);
  }

  open (SENDMAIL, "|" . $sendmail . " -oem -oi -F '$realname' -f '$from' -t  1>&2") or omailerror("Can't run $sendmail!");
  # print SENDMAIL "From: $realname <$from>\n";
  # $messagecontents .= "From: $realname <$from>\n";
  print SENDMAIL "$header_from\n";
  $messagecontents .= "$header_from\n";
  $messagecontents .= "Date: " . localtime() . "\n";
  print SENDMAIL "To: $to\n";
  $messagecontents .= "To: $to\n";
  print SENDMAIL "CC: $cc\n" if $cc;
  $messagecontents .= "CC: $cc\n" if $cc;
  print SENDMAIL "Bcc: $bcc\n" if $bcc;
  $messagecontents .= "Bcc: $bcc\n" if $bcc;
  print SENDMAIL "Reply-to: $replyto\n" if $replyto;
  $messagecontents .= "Reply-to: $replyto\n" if $replyto;
  # print SENDMAIL "Subject: $subject\n";
  # $messagecontents .= "Subject: $subject\n";
  print SENDMAIL "$header_subject\n";
  $messagecontents .= "$header_subject\n";
  print SENDMAIL "References: $references\n";
  $messagecontents .= "References: $references\n";
  print SENDMAIL "In-Reply-To: $inreplyto\n";
  $messagecontents .= "In-Reply-To: $inreplyto\n";
  print SENDMAIL "X-Mailer: oMail $version - http://webmail.omnis.ch\n";
  $messagecontents .= "X-Mailer: oMail $version - http://webmail.omnis.ch\n";
  print SENDMAIL "X-IPAddress: $ENV{REMOTE_ADDR}\n";
  $messagecontents .= "X-IPAddress: $ENV{REMOTE_ADDR}\n";
  print SENDMAIL "X-Sender: $user\@$domainname\n";
  $messagecontents .= "X-Sender: $user\@$domainname\n";

    print SENDMAIL "MIME-Version: 1.0\n";
   $messagecontents .= "MIME-Version: 1.0\n";

      opendir (OMAILDIR, "$maildir/omailtmp") or
         omailerror("coulndopen $maildir/omailtmp!");
      while (defined(my $currentfile = readdir(OMAILDIR))) {
         if ($currentfile =~ /^($thissession-att\d+)$/) {
            $currentfile = $1;
            open (ATTFILE, "$maildir/omailtmp/$currentfile");
            $/ = undef; # Force a single input to grab the whole spool
            $savedatts .= "\n--$boundary\n" . <ATTFILE>;
            $/ = "\n";
            close (ATTFILE);
         }
      }
      closedir (OMAILDIR);

   my $contenttype = '';
         if ($attachment || $savedatts) {
         $contenttype = 'multipart/mixed;';
         print SENDMAIL "Content-Type: multipart/mixed;\n";
         print SENDMAIL "\tboundary=\"$boundary\"\n\n";
         print SENDMAIL "This is a multi-part message in MIME format.\n\n";
         print SENDMAIL "--$boundary\n";
         print SENDMAIL "Content-Type: text/plain; charset=$TXT{'charset'}\n\n";
         print SENDMAIL $body;
         $body =~ s/^From />From /gm;
         print SENDMAIL "\n$savedatts\n";

         $messagecontents .= "Content-Type: multipart/mixed;\n".
                             "\tboundary=\"$boundary\"\n\nThis is a multi-part message in MIME format.\n\n".
                             "--$boundary\nContent-Type: text/plain; charset=$TXT{'charset'}\n\n$body\n$savedatts\n";
         if ($attachment) {
            my $attcontents = '';
            $/ = undef; # Force a single input to grab the whole spool
            $attcontents = <$attachment>;
            $/ = "\n";
            $attcontents = encode_base64($attcontents);
            my $content_type;
            if (defined(uploadInfo($attachment))) {
               $content_type = ${uploadInfo($attachment)}{'Content-Type'} || 'application/octet-stream';
            } else {
               $content_type = 'application/octet-stream';
            }
            print SENDMAIL "--$boundary\nContent-Type: ", $content_type,";\n";
            print SENDMAIL "\tname=\"$attname\"\nContent-Transfer-Encoding: base64\n\n";
            print SENDMAIL "$attcontents\n";
            $messagecontents .= "--$boundary\nContent-Type: $content_type;\n".
                             "\tname=\"$attname\"\nContent-Transfer-Encoding: base64\n\n".
                             "$attcontents\n";
         }
         print SENDMAIL "--$boundary--";
         $messagecontents .= "--$boundary--\n\n";
      } else {
         print SENDMAIL "Content-Type: text/plain; charset=$TXT{'charset'}\n\n$body\n";
         $body =~ s/^From />From /gm;
         $messagecontents .= "Content-Type: text/plain; charset=$TXT{'charset'}\n\n$body\n\n";
      }

      close(SENDMAIL) or senderror();
      deleteattachments();


  if ($messageid) { 
    my ($null,$null2) = updatestatus($user,$messageid,"+R");
  }
  


  # save the message to the sent folder - if needed

  if ($save_outgoing_mail ne "no") { 
    my $destfile;
    $destfile = "$maildir/sent"; 
    
    if (!(-d $destfile)) { # create directory if necessary
	 system("/bin/mkdir $destfile");
	 writelog("sendmessage: created directory $destfile");
    }
    
    if (-d $destfile) {
	 $destfile .= "/sentmail-" . $$ . "." . time();
	 open (SENT, ">$destfile") or
	   omailerror("There was a problem opening your SENT folder!");
	 print SENT "$messagecontents\n";
	 close (SENT) or omailerror ("Couldn't close SENT!");
	 writelog("sendmessage: saved mail to $destfile");
    } else {
	 writelog("oops ? sent directory doesn't exist, not saving");
    }
  } else {
    writelog("sendmessage: message not saved, as requested");
  }

  displayheaders();
}
}

############## END SENDMESSAGE ###################

#####################################################

sub senderror {
  printheader();
  print '<table border="0" align="center"><tr>
          <td valign="middle" align="center" bgcolor="#EEEEEE" width="250"
          height="250">';
  print "Sorry, there was a problem sending your message.  Please
          click your browser's Back button, check your message, and
          try again.  If the problem persists, contact the author at
          emiller\@hhs.net.";
  print '</td></tr></table>';
  printfooter();
  exit 0;
}

#####################################################

################ VIEWATTACHMENT ##################
sub viewattachment {
  my %message = %{&getmessage($user,param("message_id"))};
  if (%message) {
    my $attnumber = param("attachment_number");
    my $attachment;
    if (${$message{attachment}[$attnumber]}{encoding} =~ /^base64$/i) {
      $attachment = decode_base64(${$message{attachment}[$attnumber]}{contents});
    } elsif (${$message{attachment}[$attnumber]}{encoding} =~ /^quoted-printable$/i) {      
      $attachment = decode_qp(${$message{attachment}[$attnumber]}{contents});
    } else { ## Guessing it's 7-bit, at least sending SOMETHING back! :)
      $attachment = ${$message{attachment}[$attnumber]}{contents};
    }
    my $length = length($attachment);
    print "Content-Length: $length\n";
    print "Content-Transfer-Coding: binary\n";
    print "Connection: close\n";
    print "Content-Type: ${$message{attachment}[$attnumber]}{contenttype}; name=\"${$message{attachment}[$attnumber]}{filename}\"\n";
    unless (${$message{attachment}[$attnumber]}{contenttype} =~ /^text\//i) {
      print "Content-Disposition: attachment; filename=\"${$message{attachment}[$attnumber]}{filename}\"\n";
    }
    print "\n";
    print $attachment;
  } else {
    printheader();
    print "What the heck? Message seems to be gone!";
    printfooter();
  }
}
##################################################

################## GETHEADERS #######################
sub getheaders {
  my $user = shift;
  my $searchstring = shift;
  my $spool = '';
  my @message = ();
  my $currentheader;
  my $currentfile = '';
  my $messagesize = 0;
  my $spoolfile;
  my $spooldir;
  my $messageheader = '';
  my $messagebody = '';
  my $maildir_path = "";
  my $regexpmessnum = $firstmessage + 1;
  my $foundstatus = "";
  my $tmpline = 0;
  
  $maildir_path = $maildir; 
  $spooldir = $maildir; 
  
  my @submdirs = ();

  # folder support! yes, so easy it is :)  took 5 minutes to implement this!

  @submdirs = ('cur', 'new') if $folder eq "INBOX";
  @submdirs = ('sent') if $folder eq "SENT";
  @submdirs = ('trash') if $folder eq "TRASH";
  @submdirs = ('saved') if $folder eq "SAVED";
  
  my $messnum = 0;
  
  foreach my $submdir (@submdirs) {
    
    opendir (MAILDIR, "$maildir_path/$submdir") or
      omailerror("Couldn't opendir $maildir_path/$submdir\n");
    
    while (defined($currentfile = readdir(MAILDIR))) {
      if ( (-f "$maildir_path/$submdir/$currentfile") &&
	   !(-l "$maildir_path/$submdir/$currentfile") &&
	   ((stat "$maildir_path/$submdir/$currentfile")[3] == 1) ) {
	
	open (MESSAGE, "$maildir_path/$submdir/$currentfile") or
	  omailerror("Couldn't open $maildir_path/$submdir/$currentfile!\n");
	
	$messagesize = (stat "$maildir_path/$submdir/$currentfile")[7];
	$messageheader = " ";
	$messagebody = "";
	$foundstatus = "";
	
	while (<MESSAGE>) {
	  if ($_ eq "\n" || $_ eq "\n\r") { last; }  
	  # this way, _only_ the header will be read -> be nice with the RAM :)
	  
	  $messageheader .= $_;
	}

	# are we looking for something specific ? if yes, read the whole message. -> not
	# recommended on huge mailboxes... idea: maybe that using fgrep on the file would
	# be better ? current solution: only lookup in the first 80 lines of the file
	# this way it won't parse 1000MB attachments... :)
	if ($searchstring) {

	  # check for argument in header. if found: continue without looking at body
	  # if ($messageheader =~ /$searchstring/smig) {   (dangerous! regexp check needed) 
	  if ((index($messageheader, $searchstring) != -1) || (index($messageheader, lc($searchstring)) != -1) || (index($messageheader, uc($searchstring)) != -1) || (index($messageheader, ucfirst($searchstring)) != -1)) {
	    $foundstatus = $TXT{'header'};
	  }
	  
	  if (!$foundstatus) {
	    $tmpline = 0;
	    while (<MESSAGE>) {
		 # file body
		 $messagebody .= $_;
		 $tmpline++;
		 if ($tmpline > 80) { last; }
	    }

	    # if ($messagebody !~ /$searchstring/smig) { 
	    if ((index($messagebody, $searchstring) == -1) && (index($messagebody, lc($searchstring)) == -1) && (index($messagebody, uc($searchstring)) == -1) && (index($messagebody, ucfirst($searchstring)) == -1)) {
		 next;
	    }
	    $foundstatus = $TXT{'body'};
	  }
	}


	close (MESSAGE);
	
	if (($messnum == 1) && ((split(/\n\r*\n/, $messageheader))[0] =~
				/DON\'T DELETE THIS MESSAGE.+\nX-IMAP:/s) ) {
	  next;  # Remove those pesky IMAP messages :)
	}
	
	# get header infos
	
	my %header = qw(from    N/A
			date    N/A
			to      N/A
			subject N/A
			message_id N/A
			content_type N/A
		       );
	my $lastline = 'NONE';
	
	foreach (split(/\n/, $messageheader)) {
	  if (/^\s/) {  # that the heck is this for ?
	    if    ($lastline eq 'FROM') { $header{from} .= $_ }
	    elsif ($lastline eq 'DATE') { $header{date} .= $_ }
	    elsif ($lastline eq 'TO') { $header{to} .= $_ }
	    elsif ($lastline eq 'SUBJ') { $header{subject} .= $_ }
	  } elsif (/^from:\s+(.+)$/ig) {
	    if ($mimewords_decode) {
              # $header{from} = MIME::Words::decode_mimewords($1);
              $header{from} = $1;
	    } else {
	      $header{from} = decode_qp($1); 
	    }
	    
	    $lastline = 'FROM';
	  } elsif (/^date:\s+(.+)$/ig) {
	    $header{date} = $1;
	    $lastline = 'DATE';
	  } elsif (/^to:\s+(.+)$/ig) {
	    
	    if ($mimewords_decode) {
              # $header{to} = MIME::Words::decode_mimewords($1);
              $header{to} = $1;
	    } else {
	      $header{to} = decode_qp($1);
	    }
	    $lastline = 'TO';
	  } elsif (/^subject:\s+(.+)$/ig) {
	    
	    if ($mimewords_decode) {
              # $header{subject} = MIME::Words::decode_mimewords($1);
              $header{subject} = $1;
	    } else {                       
	      $header{subject} = decode_qp($1);
	    }
	    $lastline = 'SUBJ';
	  } elsif (/^content-type:\s+(.+)$/ig) {
	    $header{content_type} = $1;
	    $lastline = 'NONE';
	  } else {
	    $lastline = 'NONE';
	  }
	}	
	
	if ($foundstatus) {
	  $header{foundstatus} = $foundstatus;
	}

	if(!$mimewords_decode) {
          $header{from} = decode_qp($header{from});
          $header{to} = decode_qp($header{to});
          $header{cc} = decode_qp($header{cc});
          $header{subject} = decode_qp($header{subject});
        } else {
          $header{from} = MIME::Words::decode_mimewords($header{from});
          $header{to} = MIME::Words::decode_mimewords($header{to});
          $header{cc} = MIME::Words::decode_mimewords($header{cc});
          $header{subject} = MIME::Words::decode_mimewords($header{subject});
          if($lang eq 'ja') {
            $header{from} = convert2euc($header{from});
            $header{to} = convert2euc($header{to});
            $header{cc} = convert2euc($header{subject});
            $header{subject} = convert2euc($header{subject});
          }
        }

	$header{from1} = $header{from};
	$header{from2} = $header{from};



	($header{from} =~ s/^"?(.+?)"?\s+<(.*)>$/<a href="$scripturl\?action=composemessage&sort=$sort&firstmessage=$regexpmessnum&sessionid=$thissession&composetype=sendto&to=$2">$1<\/a>/) || ($header{from} =~ s/<?([^>]+)>?/<a href="$scripturl\?action=composemessage&sort=$sort&firstmessage=$regexpmessnum&sessionid=$thissession&composetype=sendto&to=$1">$1<\/a>/);

	($header{from1} =~ s/^"?(.+?)"?\s+<(.*)>$/$1/);
	($header{from2} =~ s/^"?(.+?)"?\s+<(.*)>$/$2/);
	
	# [old date support]
	#my @datearray = split(/\s/, $header{date});
	#if ($datearray[0] =~ /[A-Za-z,]/) {
	#  shift @datearray; # Get rid of the day of the week
	#}
	#$header{date} = "$month{$datearray[1]}/$datearray[0]/$datearray[2] $datearray[3] $datearray[4]";
	
	
	# status : 921508519.5533.orion:2,____       (ascii order!)
	# S = seen (read)
	# R = replied
	# F = flag
	# T = trashed
	
	if ($currentfile =~ /:2,([RSFT]+)$/) {
	  $header{status} = $1;
	} else {
	  $header{status} = "";
	}
	
	
	$header{date} = str2time($header{date}) + (3600*$timezone);
	$header{messagesize} = $messagesize;
	$total_size += $messagesize;
	$total_new_mails++ if ($header{status} !~ /S/);

	
	$header{message_id} = $submdir . "/" . $currentfile; 
	
	

	# move new mails directely in cur directory
	
	if ($submdir eq "new") {
	  my $sourcefile = "$spooldir/$submdir/$currentfile";
	  my $destfile = "$spooldir/cur/$currentfile";;
	  
	  system("/bin/cp","$sourcefile","$destfile");
	  
	  if ((-f $destfile) && ($destfile ne $sourcefile)) {
	    system("/bin/rm","$sourcefile");
	    
	    $header{message_id} = "cur/" . $currentfile; 
	    
	    if (-f $sourcefile) {
	      writelog("ERR: Moved new message from $sourcefile to $destfile but old file still there!");
	    } else {
	      writelog("OK: Moved new message from $sourcefile to $destfile");
	    }
	    $header{new} = 1;
	    
	  } else {
	    writelog("ERR: Problem moving $sourcefile to $destfile");
	    $header{new} = 0;
	  }
	}
	
	
	$message[$messnum] = \%header;
	
	$messnum++; 
      }
    }
    closedir(MAILDIR);
  }
  
  
  if ( $sort eq 'date' ) {
    @message = sort by_date @message;
  } elsif ( $sort eq 'date_rev' ) {
    @message = reverse(sort by_date @message);
  } elsif ( $sort eq 'sender' ) {
    @message = sort by_sender @message;
  } elsif ( $sort eq 'sender_rev' ) {
    @message = reverse(sort by_sender @message);
  } elsif ( $sort eq 'recipient' ) {
    @message = sort by_recipient @message;
  } elsif ( $sort eq 'recipient_rev' ) {
    @message = reverse(sort by_recipient @message);
  } elsif ( $sort eq 'status' ) {
    @message = sort by_status (sort by_date (@message));
  } elsif ( $sort eq 'status_rev' ) {
    @message = reverse(sort by_status (reverse(sort by_date @message)));
  } elsif ( $sort eq 'size' ) {
    @message = sort by_size @message;
  } elsif ( $sort eq 'size_rev' ) {
    @message = reverse(sort by_size @message);
  } elsif ( $sort eq 'subject' ) {
    @message = sort by_subject @message;
  } else {
    @message = reverse(sort by_subject @message);
  }
  
  return \@message;
}


sub by_date {
  
  my $datea = ${$a}{date};
  my $dateb = ${$b}{date};

  $dateb <=> $datea;
}

sub by_sender {
   my $sendera = ${$a}{from};
   my $senderb = ${$b}{from};

   lc($sendera) cmp lc($senderb);
}

sub by_recipient {
   my $reca = ${$a}{to};
   my $recb = ${$b}{to};

   lc($reca) cmp lc($recb);
}

sub by_subject {
   my $sendera = ${$a}{from};
   my $senderb = ${$b}{from};

   lc($sendera) cmp lc($senderb);
}


sub by_size {
   my $sizea = ${$a}{messagesize};
   my $sizeb = ${$b}{messagesize};

   $sizea <=> $sizeb;
}

sub by_status {  # new messages first

   my $statusa = (${$a}{status} =~ /S/);
   my $statusb = (${$b}{status} =~ /S/);

   $statusa <=> $statusb;
}

#################### END GETHEADERS #######################

#################### GETMESSAGE ###########################
sub getmessage {
  my $user = shift;
  my $spool = '';
  my $spoolfile; 
  my $currentfile = '';
  my $messageid = shift;
  my $maildirfound = 0;
  my $maildirmessageid = "";
  my $status = "";
  my $fileroot = "";
  
  # get filename & status & real message id (if changed)
  ($spoolfile, $messageid, $status) = &checkmailid($messageid,$user);
  
  my %message = ();
  my @attachment = ();
  
  my ($currentheader, $currentbody, $currentfrom, $currentdate,
      $currentsubject, $currentid, $currenttype, $currentto, $currentcc,
      $currentreplyto, $currentencoding, $currentreferences, 
      $currentinreplyto);
  
  if ( (-f "$spoolfile") &&
       !(-l "$spoolfile") &&
       ((stat "$spoolfile")[3] == 1) ) {
    
    open (MESSAGE, "$spoolfile") or omailerror("Couldn't open $spoolfile\n");
    
    while (<MESSAGE>) {
      $spool .= $_;
    }
    close (MESSAGE);
    $spool .= "\n\n";
    
    my $currmessage = $spool;
    
    $currentheader = $currentbody = $currentfrom = $currentdate =
      $currentsubject = $currentid = $currenttype = $currentto =
	$currentcc = $currentreplyto = $currentencoding =
          $currentreferences = $currentinreplyto = 'N/A';
    ($currentheader, $currentbody) = split(/\n\r*\n/, $currmessage, 2);
    my $lastline = 'NONE';
    foreach (split(/\n/, $currentheader)) {
      if (/^\s/) {
	if    ($lastline eq 'FROM') { $currentfrom .= $_ }
	elsif ($lastline eq 'REPLYTO') { $currentreplyto .= $_ }
	elsif ($lastline eq 'DATE') { $currentdate .= $_ }
	elsif ($lastline eq 'SUBJ') { $currentsubject .= $_ }
	elsif ($lastline eq 'TYPE') { $currenttype .= $_ }
	elsif ($lastline eq 'ENCODING') { $currentencoding .= $_ }
	elsif ($lastline eq 'TO')   { $currentto .= $_ }
	elsif ($lastline eq 'CC')   { $currentcc .= $_ }
	elsif ($lastline eq 'REFERENCES') { $currentreferences .= $_ }
	elsif ($lastline eq 'INREPLYTO') { $currentinreplyto.= $_ }
      } elsif (/^from:\s+(.+)$/ig) {
	$currentfrom = $1;
	$lastline = 'FROM';
      } elsif (/^reply-to:\s+(.+)$/ig) {
	$currentreplyto = $1;
	$lastline = 'REPLYTO';
      } elsif (/^to:\s+(.+)$/ig) {
	$currentto = $1;
	$lastline = 'TO';
      } elsif (/^cc:\s+(.+)$/ig) {
	$currentcc = $1;
	$lastline = 'CC';
      } elsif (/^date:\s+(.+)$/ig) {
	$currentdate = $1;
	$lastline = 'DATE';
      } elsif (/^subject:\s+(.+)$/ig) {
	$currentsubject = $1;
	$lastline = 'SUBJ';
      } elsif (/^message-id:\s+(.+)$/ig) {
	$currentid = $1;
	$currentid =~ s/[<>\#&=\?]//g;
	$lastline = 'NONE';
      } elsif (/^content-type:\s+(.+)$/ig) {
	$currenttype = $1;
	$lastline = 'TYPE';
      } elsif (/^content-transfer-encoding:\s+(.+)$/ig) {
	$currentencoding = $1;
	$lastline = 'ENCODING';
      } elsif (/^references:\s+(.+)$/ig) {
        $currentreferences = $1;
	$lastline = 'REFERENCES';
      } elsif (/^in-reply-to:\s+(.+)$/ig) {
        $currentinreplyto = $1;
	$lastline = 'INREPLYTO';
      } else {
	$lastline = 'NONE';
      }
    }
      
    # parse message and attachments
    

      if ($currenttype =~ /^multipart/i) {
         my @attachments;
         my ($attheader, $attcontents);
         my $boundary = $currenttype;
         $boundary =~ s/.*boundary="?([^"]+)"?.*$/$1/i;
          ($currentbody, @attachments) = split(/\-\-\Q$boundary\E\n*/,$currentbody);
         # ($attcontents, my @attachments) = split(/\-\-\Q$boundary\E\n*/,$attcontents);
         my $attnum = -1;  # So that the first increment gives 0
         my $outerattnum = -1;
         foreach my $bound (0 .. $#attachments) {
            $outerattnum++;
            $attnum++;
            my %temphash;
            ($attheader, $attcontents) = split(/\n\n/, $attachments[$outerattnum], 2);
            my $lastline='NONE';
            my ($attcontenttype, $attdisposition, $attfilename, $attencoding);
            foreach (split(/\n/, $attheader)) {
               if (/^\s/) {
                  if ($lastline eq 'TYPE') { $attcontenttype .= $_ }
               } elsif (/^content-type:\s+(.+)$/ig) {
                  $attcontenttype = $1;
                  $lastline = 'TYPE';
               } elsif (/^content-transfer-encoding:\s+(.+)$/ig) {
                  $attencoding = $1;
                  $lastline = 'NONE';
               } elsif (/^content-disposition:\s+(.+)$/ig) {
                  $attdisposition = $1;
                  $lastline = 'NONE';
               } else {
                  $lastline = 'NONE';
               }
            }

# Handle Outlook's funky "multipart inside multipart" encapsulation.  Leave
# it to MS to do something weird. TODO: Possibly implement recursion for
# deeper levels of multipart messages, to reduce code size?  Would have to
# avoid being exploited by someone sucking up all memory by deliberately
# sending messages with hundreds-of-levels deep attachments

            if ($attfilename || $attcontenttype) {
               if ($attcontenttype =~ /^multipart/i) {
                  my $boundary = $attcontenttype;
                  $boundary =~ s/.*boundary="?([^"]+)"?.*$/$1/i;
                  ($attcontents, my @attachments) = split(/\-\-\Q$boundary\E\n*/,$attcontents);
                  $attnum--; # avoid incrementing $attnum first time
                  foreach my $bound (0 .. $#attachments) {
                     last if ($attachments[$attnum + 1] =~ /^\-\-\n/);
                     $attnum++;
                     my %temphash;
                     my ($attheader, $attcontents) = split(/\n\n/, $attachments[$attnum], 2);
                     my $lastline='NONE';
                     my ($attcontenttype, $attdisposition, $attfilename, $attencoding);
                     foreach (split(/\n/, $attheader)) {
                        if (/^\s/) {
                           if ($lastline eq 'TYPE') { $attcontenttype .= $_ }
                        } elsif (/^content-type:\s+(.+)$/ig) {
                           $attcontenttype = $1;
                           $lastline = 'TYPE';
                        } elsif (/^content-transfer-encoding:\s+(.+)$/ig) {
                           $attencoding = $1;
                           $lastline = 'NONE';
                        } elsif (/^content-disposition:\s+(.+)$/ig) {
                           $attdisposition = $1;
                           $lastline = 'NONE';
                        } else {
                           $lastline = 'NONE';
                        }
                     }
                     if ($attfilename || $attcontenttype) {
                        $attfilename = $attcontenttype;
                        $attcontenttype =~ s/^(.+);.*/$1/g;
                        unless ($attfilename =~ s/^.+name="?([^"]+)"?.*$/$1/ig) {
                           $attfilename = $attdisposition || '';
                           unless ($attfilename =~ s/^.+filename="?([^"]+)"?.*$/$1/ig) {
                              $attfilename = "Unknown.html";
                           }
                        }
                        
                        $attfilename = decode_mimewords($attfilename);
                        $attheader = decode_mimewords($attheader);
                        $temphash{filename} = $attfilename;
                        $temphash{contenttype} = $attcontenttype || 'text/plain';
                        $temphash{encoding} = $attencoding;
                        $temphash{header} = $attheader;
                        $temphash{contents} = $attcontents;
                        $attachment[$attnum] = \%temphash;
                     }
                  }
               } else {
                  $attfilename = $attcontenttype;
                  $attcontenttype =~ s/^(.+);.*/$1/g;
                  unless ($attfilename =~ s/^.+name="?([^"]+)"?.*$/$1/ig) {
                     $attfilename = $attdisposition || '';
                     unless ($attfilename =~ s/^.+filename="?([^"]+)"?.*$/$1/ig) {
                        $attfilename = "Unknown.html";
                     }
                  }
                  $attfilename = decode_mimewords($attfilename);
                  $attheader = decode_mimewords($attheader);
                  $temphash{filename} = $attfilename;
                  $temphash{contenttype} = $attcontenttype || 'text/plain';
                  $temphash{encoding} = $attencoding;
                  $temphash{header} = $attheader;
                  $temphash{contents} = $attcontents;
                  $attachment[$attnum] = \%temphash;
               }
            }
         }
      } elsif ( ($currenttype ne 'N/A') && !($currenttype =~ /^text/i) ) {
         my ($attfilename, $attcontenttype, $attheader);
         my %temphash;
         $attcontenttype = $attfilename = $currenttype;
         $attcontenttype =~ s/^(.+);.*/$1/g;
         unless ($attfilename =~ s/^.+name="?([^"]+)"?.*$/$1/ig) {
            $attfilename = "Unknown.html";
         }
         $attfilename = decode_mimewords($attfilename);
         $attheader = decode_mimewords($attheader);
         $temphash{filename} = $attfilename;
         $temphash{contenttype} = $attcontenttype || 'text/plain';
         $temphash{encoding} = $currentencoding;
         $temphash{header} = $attheader;
         $temphash{contents} = $currentbody;
         $currentbody = " ";
         $attachment[0] = \%temphash;
      }

      $currentfrom = decode_mimewords($currentfrom);
      $currentto = decode_mimewords($currentto);
      $currentcc = decode_mimewords($currentcc);
      $currentreplyto = decode_mimewords($currentreplyto);
      $currentsubject = decode_mimewords($currentsubject);


    if ($lang ne 'ja') {
        $message{"header"} = $currentheader;
        $message{"body"} = $currentbody;
        $message{from} = $currentfrom;
        $message{replyto} = $currentreplyto unless ($currentreplyto eq "N/A");
        $message{to} = $currentto unless ($currentto eq "N/A");
        $message{cc} = $currentcc unless ($currentcc eq "N/A");
        $message{date} = $currentdate;
        $message{subject} = $currentsubject;
        $message{messageid} = $messageid;
        $message{contenttype} = $currenttype;
        $message{encoding} = $currentencoding;
        $message{attachment} = \@attachment;
        $message{status} = $status;
    } else{  #Japanese
        $message{"header"} = &convert2euc($currentheader);
        $message{"body"} = &convert2euc($currentbody);
        $message{from} = &convert2euc($currentfrom);
        $message{replyto} = &convert2euc($currentreplyto) unless ($currentreplyto eq "N/A");
        $message{to} = &convert2euc($currentto) unless ($currentto eq "N/A");
        $message{cc} = &convert2euc($currentcc) unless ($currentcc eq "N/A");
        $message{date} = &convert2euc($currentdate);
        $message{subject} = &convert2euc($currentsubject);
        $message{messageid} = $messageid;
        $message{contenttype} = $currenttype;
        $message{encoding} = $currentencoding;
        $message{attachment} = \@attachment;
        $message{status} = $status;
    }
    
    # Determine message's number and previous and next
    # message IDs.
    my @headers = @{&getheaders($user, "")};
    foreach my $messagenumber (0..$#headers) {
      if (${$headers[$messagenumber]}{message_id} eq $messageid) {
        $message{"prev"} = ${$headers[$messagenumber-1]}{message_id} if ($messagenumber > 0);
        $message{"next"} = ${$headers[$messagenumber+1]}{message_id} if ($messagenumber < $#headers);
        $message{"number"} = $messagenumber+1;
        last;
      } 
    }
 }
 return \%message;
}

#################### END GETMESSAGE #######################

##################### CHECKMAILID #########################

# input: msgid
# output: corrected msgid + status
# 
# + security checks on filename

sub checkmailid {
  
  my $filename = shift;
  my $user = shift;
  my $spooldir;
  my $spoolfile;
  my $submdir = "";
  my $mailroot = "";
  my $mailstatus = "";
  my $currentfile = "";

  $spooldir = $maildir;
  $spoolfile = "$maildir/$filename"; 

  # get submdir and mailroot
  
  if ($filename =~ /^(.*)\/(.*)$/ && $filename !~ /\.\./ && $filename !~ /^\//) {
    $submdir = $1;
    $mailroot = $2;
  } else {
    writelog("ALERT: $filename format uncorrect");
    omailerror("Error: $filename format uncorrect"); 
  }

  # get mailroot
  
  if ($mailroot =~ /(.*:2,)[RSFT]*$/) {
    $mailroot = $1
  }
  
  # get submdir
  
  if ($filename =~ /^(.*)\//) {
    $submdir = $1;
  }
  
  # case 1 : msg file exists : we return the same name + status
  
  if (-f "$spoolfile") {
    
    # everything ok
    
  } else {
    
    # case 2 : msg file doesn't exist anymore (status changed?) : try to guess right name
    
    opendir (MAILDIR, "$spooldir/$submdir") or
      omailerror("Couldn't opendir $spooldir/$submdir\n");
    
    while (defined($currentfile = readdir(MAILDIR))) {
      
      if ($currentfile =~ /$mailroot/) {
	
	$filename = "$submdir/$currentfile";
	$spoolfile = "$spooldir/$filename";
	
	# todo: check with -f
	last;
      }
    }
  }


  # check if file belongs to the correct user

  if (!(-O "$spoolfile" && -o "$spoolfile")) {
    writelog("ALERT: not file owner! $spoolfile");
    omailerror($TXT{'not_file_owner'});
  }
    
  # get real status
  
  if ($filename =~ /\/.*:2,([RSFT]*)$/) {
    $mailstatus = $1;
  } else {
    $mailstatus = "";
  }
  
  return ($spoolfile, $filename, $mailstatus);
}

#################### END CHECKMAILID ######################

#################### UPDATESTATUS #########################
sub updatestatus {

  my $user = shift;
  my $messageid = shift;
  my $newstatus = shift;   # format = "+X" where X = R, S, T or F
  my $spooldir;           # or "-X" to remove a status flag
  my $status = "";
  my $spoolfile;

  ($spoolfile, $messageid, $status) = &checkmailid($messageid,$user);
  my $origfile = $spoolfile;

  # handel the case the user view a new message,and tries to go back
  # using the [Back] button (for example after viewing an attachment)
  # -> it would normaly get a "msg not found" error, because the msg
  # has been moved to the /cur directory.
  
  if ( (-f "$spoolfile") &&
       !(-l "$spoolfile") &&
       ((stat "$spoolfile")[3] == 1) ) {
    
    if ($spoolfile !~ /:2,/) {
      
      # there are no status at all : just add it!
      $spoolfile .= ":2," . substr($newstatus,1,1);
      writelog("OK: spoolfile name have no :2,  ($spoolfile)");
      
    } else {
      
      writelog("OK: spoolfile name has a :2,  ($spoolfile)");
      
      # get current status in an array
      
      my @statusflags = (); 

      if (length($status)) {
	for (my $j=0; $j<length($status); $j++) {
	  push @statusflags, substr($status,$j,1);
	}
      }
      
      # add or remove status to array
      
      
      if (substr($newstatus,0,1) eq "-") {
	my @tmplist = ();
	foreach my $statustmp (@statusflags) {
	  unless ($statustmp eq substr($newstatus,1,1)) {
	    push (@tmplist, $statustmp);
	  }
	}
	@statusflags = @tmplist;
	
	
      } elsif (substr($newstatus,0,1) eq "+") {
	my $tmpregexp = substr($newstatus,1,1);
	if ($status !~ /$tmpregexp/) {
	  push (@statusflags, $tmpregexp);
	}
      }
      
      $status = "";

      if ($spoolfile =~  /^(.*):2,[RSFT]*$/) {
	$spoolfile = $1 . ":2,";
	
	if ($messageid =~  /^(.*):2,[RSFT]*$/) {
	  $messageid = $1 . ":2,";
	  
	  foreach my $statletter (sort(@statusflags)) {
	    $messageid .= $statletter;
	    $spoolfile .= $statletter;
	    $status .= $statletter;
	  } 
	}
      } 
    }
    
    if ($spoolfile ne $origfile) {
      system("/bin/mv", "-f", "$origfile", "$spoolfile");
      writelog("OK: updated status for file $origfile to $spoolfile");
    } else {
      writelog("ERR: in updatestatus, $spoolfile = $origfile");
    }
  }

  return ($status, $messageid);
}


################### END UPDATESTATUS ######################


#################### SAVEMESSAGE #######################
sub savemessage {
  
  my @messageids = param("message_ids");
  my $messageids = join("\n", @messageids);
  
  my $maildir_path = "";
  my $spoolfile = "";
  my ($messageid, $status);
  my $destfile;
  my $counter = 50;
  my $destfiletmp = "";

  omailerror ("You shouldn't be trying to save here!") if ($folder eq 'SAVED');
  $destfile = "$maildir/saved"; 
  
  if (!(-d $destfile)) { # create directory if necessary
    system("/bin/mkdir $destfile");
    writelog("savemessage: created directory $destfile");
  }

  foreach my $messagefile (@messageids) {
    
    ($spoolfile, $messageid, $status) = &checkmailid($messagefile,$user);
    
    if ( (-f "$spoolfile") &&
	 !(-l "$spoolfile") &&
	 ((stat "$spoolfile")[3] == 1) ) {
      

      if (-d $destfile) {
	$counter++;
	$destfiletmp = $destfile . "/saved-" . $counter . "." . $$ . "." . time() . ":2," . $status;
	
	system("/bin/mv", "$spoolfile", "$destfiletmp");
	writelog("User moved message $spoolfile to $destfiletmp");
      } else {
	writelog("oops ? saved directory doesn't exist, not saving");
      }
    }
  }
  
  if (param("messageafterdelete")) {
    readmessage();
  } else {
    displayheaders();
  }
}
#################### END SAVEMESSAGE #######################


#################### DELETEMESSAGE #######################
sub deletemessage {
  
  my @messageids = param("message_ids");
  my $messageids = join("\n", @messageids);
  
  my $maildir_path = "";
  my $spoolfile = "";
  my ($messageid, $status);
  my $counter = 70;
  my $destfile;
  my $destfiletmp = "";
  
  $destfile = "$maildir/trash"; 
  
  if (!(-d $destfile)) { # create directory if necessary
    system("/bin/mkdir $destfile");
    writelog("deletemessage: created directory $destfile");
  }
  
  
  foreach my $messagefile (@messageids) {
    
    ($spoolfile, $messageid, $status) = &checkmailid($messagefile,$user);
    
    if ( (-f "$spoolfile") &&
	 !(-l "$spoolfile") &&
	 ((stat "$spoolfile")[3] == 1) ) {

      if ($folder eq "TRASH") {
	# definitely remove a file
	system("/bin/rm", "$spoolfile");
	writelog("User deleted message from trash - $spoolfile");
      } else {
	# delete from a folder = move to trash
	
	if (-d $destfile) {
	  $counter++;
	  $destfiletmp = $destfile . "/saved-" . $counter . "." . $$ . "." . time() . ":2," . $status;
	  
	  system("/bin/mv", "$spoolfile", "$destfiletmp");
	  writelog("User moved message $spoolfile to $destfiletmp");
	} else {
	  writelog("oops ? saved directory doesn't exist, not saving");
	}	
      }
    }
  }
  
  if (param("messageafterdelete")) {
    readmessage();
  } else {
    displayheaders();
  }
}
#################### END DELETEMESSAGE #######################

###################### TOGGLEFLAG ############################
sub toggleflag {

  my $flag = param("flag");
  my ($spoolfile, $messageid, $status);
  my ($null,$null2);
  my $newmessageid = "";
  
  my @messageids = param("message_id");
  # my $messageids = join("\n", @messageids);
  
  foreach my $messageid (@messageids) {
    
    $newmessageid = $messageid;
    
    ($spoolfile, $messageid, $status) = &checkmailid($messageid,$user);
    
    if ($flag =~ /[RSFT]/) {   # check if it is a valid flag
      
      if ($status =~ /$flag/) {
	($null,$newmessageid) = updatestatus($user,$messageid,"-$flag");
      } else {
	($null,$newmessageid) = updatestatus($user,$messageid,"+$flag");
      }
    }
  }   
  
  if (param("after") eq "msg") {  
    
    readmessage();
  } else {
    displayheaders();
  }  
}

###################### END TOGGLEFLAG  ########################


###################### ENCODING/DECODING ##############################
		   
sub decode_qp ($) {
  my $res = shift;
  $res =~ s/[ \t]+?(\r?\n)/$1/g;  # rule #3 (trailing space must be deleted)
  $res =~ s/=\r?\n//g;            # rule #5 (soft line breaks)
  $res =~ s/=([\da-fA-F]{2})/pack("C", hex($1))/ge;
  $res;
}

sub convert2euc {
  my $str = shift;
  # $str = &Jcode::convert(\$str, 'euc', 'jis', "z");
  $str = Jcode->new($str)->mime_decode->euc;
  return $str;
}

sub convert2jis {
  my $str = shift;
  $str = Jcode->new($str)->iso_2022_jp;
  # $str = &Jcode::convert(\$str, 'jis', Jcode::getcode(\$str), "z");
  return $str;
}

sub decode_mimewords {
  my $str = shift;
  if ($mimewords_decode) {
	$str = MIME::Words::decode_mimewords($str);
  }
  return $str;
}



################### END ENCODING/DECODING ##########################

##################### GETATTLIST ###############################
sub getattlist {
   my $currentfile;
   my @attlist;
   $savedattsize = 0;

   opendir (OMAILDIR, "$maildir/omailtmp") or
      omailerror("couldnt_open $maildir/omailtmp!");
   while (defined($currentfile = readdir(OMAILDIR))) {
      if ($currentfile =~ /^($thissession-att\d+)$/) {
         $currentfile = $1;
         $savedattsize += ( -s "$maildir/omailtmp/$currentfile" );
         open (ATTFILE, "$maildir/omailtmp/$currentfile");
         while (defined(my $line = <ATTFILE>)) {
            if ($line =~ s/^.+name="?([^"]+)"?.*$/$1/i) {
               $line =~ s/&/&amp;/g;
               $line =~ s/\"/&quot;/g;
               $line =~ s/</&lt;/g;
               $line =~ s/>/&gt;/g;
               $line =~ s/^(.*)$/<em>$1<\/em>/;
               push (@attlist, $line);
               last;
            }
         }
         close (ATTFILE);
      }
   }
   closedir (OMAILDIR);
   return \@attlist;
}
##################### END GETATTLIST ###########################

##################### DELETEATTACHMENTS ############################
sub deleteattachments {
  my $currentfile;
  opendir (OMAILDIR, "$maildir/omailtmp") or
    omailerror("couldnt_open $maildir/omailtmp!");
  while (defined($currentfile = readdir(OMAILDIR))) {
    
    # remove files from current session, after sendmail
    if ($currentfile =~ /^($thissession-att\d+)$/) {
	 $currentfile = $1;
	 unlink ("$maildir/omailtmp/$currentfile");
    }
    
    # remove any too old attachments from any session
    if (( -M "$maildir/omailtmp/$currentfile" > ($sessiontimeout/60/24)) && 
	   ($currentfile =~ /^(.*-att\d+)$/)) {
	 $currentfile = $1;
	 unlink ("$maildir/omailtmp/$currentfile");
    }
    
  }
  closedir (OMAILDIR);
}
#################### END DELETEATTACHMENTS #########################




##################### FIRSTTIMEUSER ################################
sub firsttimeuser {
  printheader();
  print '<table border="0" align="center"><tr><td colspan="', ($#available_lang+1) ,'"valign="middle" align="center" bgcolor="#DDDDDD">';
  print '<FONT SIZE="+2"><B>Welcome! Bienvenue! Willkommen!</B></FONT></td></tr>';
  print '<tr><td colspan="', ($#available_lang+1) ,'"valign="middle" align="center" bgcolor="#EEEEEE" width="250">';
  print "Welcome to oMail! Please choose your language:";
  print '</td></tr><tr><td>&nbsp;</td></tr>';
  
  foreach my $tmplangok (@available_lang) {
    print '<tr><td bgcolor="#EEEEEE" valign="middle" align="center">';
    print startform(-action=>"$prefsurl");
    print hidden(-name=>'lang',
		 -default=>$tmplangok,
		 -override=>'1');
    print hidden(-name=>'sessionid',
		 -default=>$thissession,
		 -override=>'1');
    print hidden(-name=>'firsttimeuser',
		 -default=>'yes',
		 -override=>'1');
    print submit($available_lang_names{$tmplangok});
    print end_form(); 
    print '</td></tr>';
  }
  print '</tr></table>';
  printfooter();

  
  my $destdir;
  $destdir = $maildir; 
  
  # config
  if (!(-d "$destdir/config")) { 
    system("/bin/mkdir $destdir/config");
    writelog("firsttime: created directory $destdir/config");
  }
  # trash
  if (!(-d "$destdir/trash")) { 
    system("/bin/mkdir $destdir/trash");
    writelog("firsttime: created directory $destdir/trash");
  }
      # sent
  if (!(-d "$destdir/sent")) { 
    system("/bin/mkdir $destdir/sent");
    writelog("firsttime: created directory $destdir/sent");
  }
  # saved
  if (!(-d "$destdir/saved")) { 
    system("/bin/mkdir $destdir/saved");
    writelog("firsttime: created directory $destdir/saved");
  }
  # omailtmp
  if (!(-d "$destdir/omailtmp")) { 
    system("/bin/mkdir $destdir/omailtmp");
    writelog("firsttime: created directory $destdir/omailtmp");
  }

}
################### END FIRSTTIMEUSER ##############################

###################### READPREFS #########################
sub readprefs {
  my ($key,$value);
  my %prefshash;

  if ( -f "$prefsdir/config" ) {
    open (CONFIG,"$prefsdir/config") or
      omailerror("Can't open your user config file, but it exists!");
    while (<CONFIG>) {
      ($key, $value) = split(/=/, $_);
      chomp($value);
      $prefshash{"$key"} = $value;
    }
    close (CONFIG) or omailerror("Couldn't close your config file!");
  }
  return \%prefshash;
}
##################### END READPREFS ######################


###################### READSIGS #########################
sub readsigs {
  my ($key,$value);
  my %sigshash;
  my $tmps;
  foreach my $sigfile (@sigfiles) {
    if ( -f "$prefsdir/$sigfile" ) {
      $sigshash{$sigfile} = '';
      open (SIGNATURE, "$prefsdir/$sigfile") or
        omailerror("Can't open your signature, but file exists!");
      while ($tmps = <SIGNATURE>) {

	while($tmps =~ /(\n|\r|\r\n|\n\r)$/ ) {	chop($tmps); }
	chomp($tmps);
        $sigshash{$sigfile} .= $tmps . "\\" . "n";
      }
      close (SIGNATURE) or omailerror("Couldn't close your signature file!");
    }
  }
  return \%sigshash;
}
##################### END READSIGS ########################

##################### EMPTY TRASH #########################

sub emptytrash {

  my $tmpmaildir = shift;
  my $tmpfile = shift;
  my $destdir = "$tmpmaildir/trash";
  my $tmpcount = 0;
  my $tmp2 = "";

  # todo: check if dir empty or not
  # fixme: on logout, dir is recreated with owner root!!!!

  if (-d $destdir) {	

    opendir(DIR, $destdir) or omailerror("can't opendir $destdir: $!");
    while (defined($tmpfile = readdir(DIR))) {
	 next if $tmpfile =~ /^\.\.?$/;
	 my $tmp2 = "$destdir/$tmpfile";
	 $tmp2 =~ s/\.\.//; if ($tmp2 =~ /^(.*)$/) { $tmp2 = $1; }  # untaint
	 if (unlink "$tmp2") {
	   $tmpcount++; 
	 } else {
	   writelog("emptytrash: Can't unlink $tmpfile: $!");
	 }
    }
    closedir(DIR);
    if ($tmpcount) {
	 writelog("emptytrash: removed all mails (nb: $tmpcount) set as deleted in $destdir");
    } else {
	 writelog("emptytrash: $destdir empty, no mails in trash to be deleted"); 
    }
  } else {
    writelog("emptytrash: error, directory $destdir not exists");
  }
}

################## END EMPTY TRASH ########################


################ GRABADDR #####################
sub grabaddr {
  printheader("Grab");
  
  my ($bgcolor, $status, $message_size, $searchstring);
  my @headers = @{&getheaders($user, $searchstring)};
  my $base_url = "$scripturl?sessionid=$thissession&sort=$sort";


  print start_form(-method=>"POST",
			    -action=>$prefsurl);
  print hidden(-name=>'action',
			-default=>'addaddresslist',
			-override=>'1');
  print hidden(-name=>'sessionid',
			-default=>$thissession,
			-override=>'1');
  print hidden(-name=>'folder',
			-default=>$folder,
			-override=>'1');
  print hidden(-name=>'sort',
			-default=>$sort,
			-override=>'1');


  # headers table
    
  print '<table border="0" align="center" width="95%" cellpadding="1" cellspacing="1">';
  
  # first row - mbox name / message nb
  
  print '<tr><td colspan="7" bgcolor="#002266" align="center"><table border="0" width="100%" cellpadding="0" cellspacing="0">',
  '<tr><td align="left" width="33%"><font color="#ffffff" face="',$default_variable_font,'" size="3"><b>',uc($TXT{$folder}),'</b>';
  
  print "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; ";  
  print " [ <a href=\"$base_url&action=grabaddr&firstmessage=0&folder=INBOX\"><font color=\"#ffffff\">$TXT{'INBOX'}</font></a> ] " 
    if $folder ne "INBOX";
  print " [ <a href=\"$base_url&action=grabaddr&firstmessage=0&folder=SENT\"><font color=\"#ffffff\">$TXT{'SENT'}</font></a> ] " 
    if $folder ne "SENT";
  print " [ <a href=\"$base_url&action=grabaddr&firstmessage=0&folder=SAVED\"><font color=\"#ffffff\">$TXT{'SAVED'}</font></a> ] " 
    if $folder ne "SAVED";
  print " [ <a href=\"$base_url&action=grabaddr&firstmessage=0&folder=TRASH\"><font color=\"#ffffff\">$TXT{'TRASH'}</font></a> ] "
    if $folder ne "TRASH";
  
  print '  </font></td></tr></table></td></tr>';


  
  # second row : navigatation
  
  print '<tr><td bgcolor="#dcdcdc" colspan=7><table border="0" width="100%" cellpadding="0" cellspacing="0">';
  print '<td align=center><font face="',$default_variable_font,'" size=2>';
  
  
  print '<table border=0 width="100%">';
  print '<tr><td valign="middle" align="left">';
  print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=$firstmessage\">$TXT{'INBOX'}</a> | ";
  print "<a href=\"$base_url&folder=$folder&action=composemessage&firstmessage=$firstmessage\">$TXT{'compose'}</a> | ";
  print "<a href=\"$prefsurl?sessionid=$thissession&folder=$folder&sort=$sort&firstmessage=$firstmessage\">$TXT{'userconfig'}</a> | ";
  print "<a href=\"$prefsurl?action=editaddresses&folder=$folder&sessionid=$thissession&sort=$sort&firstmessage=$firstmessage&lang=$lang\">$TXT{'addressbook'}</a> | ";
  print "<a href=\"$base_url&folder=$folder&action=displayheaders&firstmessage=$firstmessage\">$TXT{'refresh'}</a> | ";

  print "<a href=\"$base_url&folder=$folder&action=logout&firstmessage=$firstmessage\">$TXT{'logout'}</a>";
  

  print '</tr></table>';
  print '</td><td>&nbsp;</td>';
  print '</tr></table></td></tr>';
  
  # row names
  
  print '<tr><td colspan="7" bgcolor="#ffffff">&nbsp;</td></tr>';

  print '<tr><td colspan="5" bgcolor="#ffffff">';

  my %addrhash = ( );
  my $tmp1 = "";
  my $tmp2 = "";

  foreach my $messnum (1 .. 500) {        # fixme
    last if !(defined($headers[$messnum]));
    $tmp1 = ${$headers[$messnum]}{from1};
    $tmp2 = ${$headers[$messnum]}{from2};
    if (length($tmp2) > 5) {
       $addrhash{$tmp2} = $tmp1;
    }
  }

  print '<select name="addresses" size=25 multiple>'; 
  foreach my $key2 (sort(keys(%addrhash))) {
    my $tmp1 = $addrhash{$key2} . "|" . $key2;
    print "<option value=\"$tmp1\">" . $addrhash{$key2} . "  &lt;"  . $key2 . "&gt;</option>\n";
  }
  
  print "</select>";
  
  print '</td>';
  print '<td colspan="5" bgcolor="#ffffff" valign="center">';
  print '<input type="submit" value="' . $TXT{'import_selected_addr'} . '">';
  print '</td></tr>';
  print '</table>';

  print end_form();
  printfooter();
} 
############### END GRABADDR ##################


####################### ONELOGIN ##########################

## check /contrib/onelogin for more infos
sub onelogin {
    
  if (!$allow_one_login) {
    writelog("onelogin access tried, while use not allowed");
    omailerror("onelogin-access not allowed!");
  }
  
  $onelogin_active = 1;

  my $datastring = param("data");
  my $tmp1 = "";
  my $tmp3 = "";
  my $md5 = "";
  my $key = "";
  my $val = "";
  
  $tmp1 = decode_base64($datastring);
  my @atmp2 = split(/\&md5/,$tmp1);
  my $originalstring = $atmp2[0];           # original strign before MD5
  
  #print "data2a : $tmp1\n<br>";
  #print "data2ori : $originalstring\n<br>";
  #$secret_key = "22";

  my $md5result = md5_hex($originalstring . $secret_key);                   # should be the same as in encode_urldata
  
  #print "md1 : $md5result\n<br>";
  
  my @in = split(/[&;]/,$tmp1); 
  
  foreach my $i (0 .. $#in) {
    # Convert plus to space
    $in[$i] =~ s/\+/ /g;
    
    # Split into key and value.  
    ($key, $val) = split(/=/,$in[$i],2); # splits on the first =.
    
    # Convert %XX from hex numbers to alphanumeric
    $key =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;
    $val =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;
    
    # Associate key and value
    param($key, $val);
    #print "<br>$key  <--- $val";
  }
  if ($md5result eq param("md5")) { 
    #print "<br>MD OK!"; 
  } else { 
    #print "<br>MD ERR"; 
    writelog("onelogin: md5 doesn't match!");
    omailerror("md5 security error - contact your server admin");
    exit(0);
  }

  #print "<br>time:  " . param("time") . " --- local: " . time . "";
  if (time < param("time")) { 
    #print "<br>time OK!"; 
  } else { 
    #print "<br>time ERR"; 
    writelog("onelogin: login timeout");
    omailerror("login not valid anymore (timeout) : try again please!"); 
    exit(0);
  }

  writelog("onelogin access for user " . param("userid"));
  param("action","login");
  login();
}

##################### END ONELOGIN ########################



##################### WRITELOG ############################
sub writelog {
  unless ( ($logfile eq 'no') || ( -l "$logfile" ) ) {
    open (LOGFILE,">>$logfile") or omailerror("Couldn't open $logfile!\n");
    my $timestamp = localtime();
    my $logaction = shift;
    my $loggeduser = $user || 'UNKNOWNUSER';
    print LOGFILE "$timestamp - [$$-O] ($ENV{REMOTE_ADDR},$ENV{REMOTE_HOST}) $loggeduser - $action - $logaction\n";
    close (LOGFILE);
  }
}
#################### END WRITELOG #########################


##################### OMAILERROR ##########################

sub omailerror {
  
  printheader();
  print '<br><br><table border="0" align="center"><tr><td valign="middle" align="center" bgcolor="#DDDDDD">';
  print '<FONT SIZE="+2"><B>Problem</B></FONT></td></tr>';
  print '<tr><td valign="middle" align="center" bgcolor="#EEEEEE" width="150" height="150">';
  print shift;
  print '</td></tr>';
  print '<tr><td valign="middle" align="center" bgcolor="#DDDDDD">';
  print '[ <a href="mailto:' . $scriptadmin . '?SUBJECT=oMail_Problem">support</a> ] &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;';

  if ($onelogin_active || $force_login_startpage_on_errors_or_logout) {
    print '[ <a href="' . $one_login_startpage . '">login</a> ]</td></tr>';
  } else {
    print '[ <a href="' . $scripturl . '">login</a> ]</td></tr>';
  }    
 
  print '</table>';

  
  printfooter();
  exit 0;
  
}
################### END OMAILERROR #######################


######################## ABOUT ###########################

sub about {
  
  printheader();

  print '<table border="0" align="center"><tr><td valign="middle" align="center" bgcolor="#DDDDDD">';
  print '<FONT SIZE="+2"><B>About oMail-webmail ',$version,'</B></FONT></td></tr>';
  print '<tr><td valign="middle" bgcolor="#EEEEEE" width="470"><br><br>';

  print <<ABOUT;

<ul>
<li>oMail-webmail $version is a Webmail solution for mail server based on <a href="http://www.qmail.org">qmail</a>
and optionaly <a href="http://www.em.ca/~bruceg/vmailmgr/">vmailmgr</a>
or <a href="http://www.inter7.com/vpopmail/">vpopmail</a>.<br><br></li>


<li>Features:
<ul>
<li>qmail, vmailmgr and vpopmail authentication support (<i>not</i> sendmail)</li>
<li>100% Maildir based</li>
<li>reads the mail directely from the server disk, without need for POP3 or IMAP</li>
<li>multiple language support</li>
<li>session expiration after N minutes for security </li>
<li>multiple signatures und headers support</li>
<li>basic folders support (4 defined folders)</li>
<li>mailbox search engine</li>
<li>featured addressbook</li>
</ul>
<br></li>

<li>This a <a href="http://www.gnu.org/copyleft/gpl.html" target="_new">GPL</a> licensed project, created and currently maintained by
<a href="mailto:omail\@omnis.ch">Olivier M&uuml;ller</a>, Z&uuml;rich,<br> Switzerland. 
If you want to show your appreciation of oMail-webmail, you can consider<br> getting me one of the CDs
from my CD Wishlist at CDNOW :<br> <a href="http://cdnow.com/gift/om\@omnis.ch" target="_new">http://cdnow.com/gift/om\@omnis.ch</A>.
Please put your eMail address into <br>the Message area, so that I can thank you for your present!<br><br></li>
ABOUT
         
  print '<li>Supported languages: <font color=blue>';
  foreach my $tmplangok (@available_lang) {
    print $available_lang_names{$tmplangok}," ";
  }
  print '</font><br><br></li>';

print <<ABOUT5;
<li>oMail-webmail is programmed in <a href="http://www.perl.org">Perl</a>, and is originaly based 
on the  <a href="http://neomail.sourceforge.net">Neomail</a> project, started in 1999 by Ernie Miller.<br><br></li>

<li>If you are interested by this project, you will find more information on following webpages. Subscribing
to the <a href="http://sourceforge.net/mail/?group_id=3658">mailing lists</a> is also
highly recommended.

<table border="0"><tr><td>
<ul>
<li><a href="http://sourceforge.net/project/showfiles.php?group_id=3658">oMail-webmail download page</a></li>
<li><a href="http://cvs.sourceforge.net/cgi-bin/viewcvs.cgi/omail/webmail/">oMail-webmail public CVS tree</a></li>
<li><a href="http://sourceforge.net/mail/?group_id=3658">Mailing lists (subscribe/archives)</a></li>
<li><a href="http://omail.omnis.ch">The oMail-admin project homepage (for vmailmgr users)</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</li>
<li><a href="http://wm.omnis.ch">Live Demo : login: <I>test\@omnis.ch</I>, and password: <I>test</I></a></li>
<li><a href="http://freshmeat.net/projects/omail-webmail/">Freshmeat oMail-webmail entry</a></li>
</ul>
</td></tr></table>
<br></li>

ABOUT5

 print '<li>CVS Version: $Id: omail.pl,v 1.71 2001/08/26 12:35:27 swix Exp $ ';
 print "<br><br></li>";

print <<ABOUT2;

<li>Sponsors:
<ul>
<li><a href="http://www.omnis.ch" target="_new">Omnis Internet Services, Switzerland</a></li>
<li><a href="http://www.insign.ch" target="_new">insign gmbh, Switzerland</a></li>
<li><a href="http://www.webstyle.ch" target="_new">Webstyle gmbh, Switzerland</a></li>
<li><a href="http://www.drakkar.ch" target="_new">Drakkar Communication, Switzerland</a></li>
<li><a href="http://www.bfk.ch" target="_new">BfK, Switzerland</a></li>
<li><a href="http://www.terratec.de" target="_new">TerraTec Electronic GmbH, Germany</a> (look at 
their <a href="http://www.m3po.net/_eng/index_eng.html" target="_new">m3po</a>! :-)</li>
</ul>
<br></li>


<li>
Suggestions, requests and bugfixes welcome in the form below! <br>
(This is <b>NOT</b> intended for end user support: please ask your local sysadmin at 
<a href="mailto:$scriptadmin?SUBJECT=oMail_Webmail_Problem">$scriptadmin</a>, thanks. And the
mailing list is the right place if you need installation support.)
<form action="http://www.8304.ch/cgi-bin/formmail.pl" method="post">
<input type="hidden" name="recipient" value="omail-webmail-feedback\@omnis.ch">
<input type="hidden" name="subject" value="oMail-webmail $version comment form">
<input type="hidden" name="redirect" value="http://$ENV{'HTTP_HOST'}$ENV{'REQUEST_URI'}">
<input type="hidden" name="sender" value="$ENV{'REMOTE_ADDR'}">
<input type="hidden" name="required" value="comment,email">
ABOUT2

 print '<input type="hidden" name="sender" value="Version: $Id: omail.pl,v 1.71 2001/08/26 12:35:27 swix Exp $ ">';

print <<ABOUT3;
<table border="0">
<tr><td align="right">Email: </td><td><small>
<input type="text" size="30" name="email"></small></td></tr>
<tr><td align="right">Comment: </td><td>
<small><textarea name="comment" cols="40" rows="5"></textarea></small>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td></tr>
<tr><td colspan="2" align="center"><small><input type="submit" value="Send"></small></td></tr>
</table>
</form>
</li>
ABOUT3

  print '</ul></td></tr>';
  print '<tr><td valign="middle" align="center" bgcolor="#DDDDDD">';
  print '[ <a href="mailto:' . $scriptadmin . '?SUBJECT=oMail_webmail_Problem">support</a> ] &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;';
  print '[ <a href="' . $scripturl . '">login</a> ]</td></tr>';
 
  print '</table>';

  
  printfooter();
  exit 0;
  
}
##################### END ABOUT #########################

##################### PRINTHEADER #########################
sub printheader {
  unless ($headerprinted) {
    $headerprinted = 1;

    if (!defined($TXT{'charset'})) { $TXT{'charset'} = "iso-8859-1"; }
    if (!defined($TXT{'charset_http'})) { $TXT{'charset_http'} = "iso-8859-1"; }
    
    my $charset_http ='';
    $charset_http = $TXT{'charset_http'} ? $TXT{'charset_http'}:$TXT{'charset'};

    my $tmpuser = $user;
    if ($vmailmgr) { $tmpuser .= "\@$domainname"; }
    my $cookie = "";
    my $action = shift;
    $autorefresh *= 60;
    my $url = "$scripturl?sessionid=$thissession&sort=$sort&action=displayheaders&firstmessage=$firstmessage";

    if ($action eq "displayheaders" && $autorefresh && !$use_cookies) {
      print header(-Refresh=>"$autorefresh; URL=$url",
                   # -type=>"text/html; charset=$TXT{'charset'}",
                   -type=>"text/html; charset=$charset_http",
                   -pragma=>'no-cache');
    } elsif ($action eq "displayheaders" && $autorefresh && $use_cookies) {
      $cookie = cookie(-name=>'omaillogin',
                          -value=>$tmpuser,
                          # -type=>"text/html; charset=$TXT{'charset'}",
                          -type=>"text/html; charset=$charset_http",
                          -expires=>'+1M');      
      print header(-Refresh=>"$autorefresh; URL=$url",
                   -cookie=>$cookie,
                   # -type=>"text/html; charset=$TXT{'charset'}",
                   -type=>"text/html; charset=$charset_http",
                   -pragma=>'no-cache');
    } elsif ($action eq "displayheaders" && !$autorefresh && $use_cookies) {
      $cookie = cookie(-name=>'omaillogin',
                          # -type=>"text/html; charset=$TXT{'charset'}",
                          -type=>"text/html; charset=$charset_http",
                          -value=>$tmpuser,
                          -expires=>'+1M');      
      print header(-cookie=>$cookie,
                   # -type=>"text/html; charset=$TXT{'charset'}",
                   -type=>"text/html; charset=$charset_http",
                   -pragma=>'no-cache');
    } else {
      print header(-pragma=>'no-cache',
                   # -type=>"text/html; charset=$TXT{'charset'}");
                   -type=>"text/html; charset=$charset_http");
    }
    
    print start_html(-"title"=>"oMail-webmail version $version [$lang]",
                     -BGCOLOR=>'#FFFFFF');    
    print <<TEST;

<style type="text/css">
<!--
body {  font-family: $default_variable_font; font-size: 10pt}
th   {  font-family: $default_variable_font; font-size: 10pt; font-weight: bold; background-color: #D3DCE3;}
td   {  font-family: $default_variable_font; font-size: 10pt; }
test.td   {  font-family: $default_variable_font; font-size: 10pt;}
form   {  font-family: $default_variable_font; font-size: 10pt}
h1   {  font-family: $default_variable_font, sans-serif; font-size: 16pt; font-weight: bold}
A:link    {  font-family: $default_variable_font; font-size: 10pt; text-decoration: none; color: blue}
A:visited {  font-family: $default_variable_font; font-size: 10pt; text-decoration: none; color: blue}
A:hover   {  font-family: $default_variable_font; font-size: 10pt; text-decoration: underline; color: red}
A:link.nav {  font-family: $default_variable_font; color: #000000}
A:visited.nav {  font-family: $default_variable_font; color: #000000}
A:hover.nav {  font-family: $default_variable_font; color: red;}
.nav {  font-family: $default_variable_font; color: #000000}
-->
</style>

TEST
  }   

}
################### END PRINTHEADER #######################

################### PRINTFOOTER ##########################
sub printfooter {
  print "<p align=\"center\"><FONT SIZE=\"1\"><BR>
          <a href=\"$scripturl?action=about\" target=\"_new\">
          oMail-webmail</a><br>Version $version<BR>
          </FONT></FONT></p></BODY></HTML>";
}
################# END PRINTFOOTER #########################

